
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program cleanbg
!__________________________________________________________________________________________________________________________________
!
! Clean background by interpolating sources off the background according to their footprints.
!
! This code is part of GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!
       implicit      none

       logical       lfnam1, lfnam2, lunix, larithm, lmedian
                                          
       character*5   cfpptime
       character*6   cfitsversion, lverbose
       character*7   clibname
       character*8   ctime                
       character*10  cdate
       character*21  compda
       character*80  object, bunit, ctype1, ctype2, creator, history
       character*128 header, footer(2)
       character*500 arg1, arg2, arg3, arg4, arg5, arg6, originback, sourcemask, cleanback, worker, outname

       integer       i, j, iorigin, iobjs, iclean, iwork, ndate, iotty, nx, ny, nx1, ny1, nx2, ny2, firstb, lastc, lasts
     &             , ia1, ia2, ia3, ia4, ia5, ia6, iz1, iz2, jz1, jz2, k, l, ij, ji, im1, jm1, ii, iadd, jadd, nave, lb
     &             , nav, ia, ja, iim, iip, jjm, jjp, iim1, iip1, iim2, iip2, jjm1, jjp1, jjm2, jjp2, isp1, ion, icb, irc
     &             , dnx, dny, dnxa, dnxb, dnya, dnyb, nmax, blank, nsort, npts, npts2, nmed, m, n, npmax, mi2, nj2, nr2

       parameter   ( iotty = 6, iadd = 0, jadd = 0 )  !<-- number of additional pixels to offset (unused now)

       real          fitsvers
       real*8        dx, dy, dx1, dy1, dx2, dy2, xr, yr, fint, crpix1, crpix2, crval1, crval2, rx, ry, ra, dec, beam, beam1, beam2
     &             , almostzero, df, f1, f2, fwmin, funmin, funmax, crota1, crota2, epoch, equinox, bzero, bscale, wave, datamin
     &             , datamax, cd11, cd12, cd21, cd22, length1, length2, length3, length4, funw(4), func(4)

       real*8, allocatable :: forigin(:,:), fclean(:,:), fsmask(:,:), fwork(:,:), fwork1(:,:), fclean1(:,:), fclean2(:,:)
     &                      , fclean3(:,:), fclean4(:,:), fwork2(:,:), fwork3(:,:), fwork4(:,:), fun(:), funmed(:,:)

       parameter   ( almostzero = 1.0d-12 )

       external     firstb, lastc, rfits, wfits, when, printerr, sortarr, spline, splint
!__________________________________________________________________________________________________________________________________
!
! Determine operating system

       call osystem ( lunix )

       call ftvers ( fitsvers )

       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       compda = __DATE__//' '//cfpptime

! Get command line parameters (file names).

       call getarg ( 1, arg1 )
       call getarg ( 2, arg2 )
       call getarg ( 3, arg3 )
       call getarg ( 4, arg4 )
       call getarg ( 5, arg5 )
       call getarg ( 6, arg6 )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )
       ia6 = lastc ( arg6 )

       lverbose = '-verb2'
       if (ia6 .gt. 0 .and. arg6(1:ia6-1) .eq. '-verb') lverbose = arg6(1:ia6)
       if (ia5 .gt. 0 .and. arg5(1:ia5-1) .eq. '-verb') lverbose = arg5(1:ia5)
       if (ia4 .gt. 0 .and. arg4(1:ia4-1) .eq. '-verb') lverbose = arg4(1:ia4)
       if (ia3 .gt. 0 .and. arg3(1:ia3-1) .eq. '-verb') lverbose = arg3(1:ia3)

       if (lverbose .eq. '-verb2') write (iotty,'(/a$)') ' CLEANBG = Produce Clean Background = '//compda
       if (arg1(1:ia1) .eq. ':') stop
       if (lverbose .eq. '-verb2') write (iotty,'()')
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Alexander Men`shchikov, SAp IRFU CEA Saclay, France.'
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                                //' by William D Pence.'

       if (ia3 .eq. 0) then
         write (iotty,'(/a)') ' USAGE: cleanbg [+]<nave> <originback> <sourcemask> [-o <outname>] [-verb{0|1|2}]'
         stop 99
       endif
       if (arg1(1:1) .eq. '+') then
         larithm = .true.
         lmedian = .false.
       else
         larithm = .false.
         lmedian = .true.
       endif
       read (arg1(1:ia1),*,err=9) nave
       goto 10
  9    continue       
         write (iotty,'(/a)') '   CLEANBG: ERROR: Trouble reading command-line parameter ''nave''.'
         stop 99
 10    continue
       originback = arg2(1:ia2)
       sourcemask = arg3(1:ia3)
       iorigin = lastc ( originback  )
       iobjs  = lastc ( sourcemask )

! Check if the input file exists and open it.

       inquire ( file=originback(1:iorigin), exist=lfnam1 )
       if (.not.lfnam1) then

! Maybe the file name extension was omitted?

         originback = originback(1:iorigin)//'.fits'
         iorigin = lastc ( originback )
         inquire ( file=originback(1:iorigin), exist=lfnam1 )
       endif

! Check if the input file exists and open it.

       inquire ( file=sourcemask(1:iobjs), exist=lfnam2 )
       if (.not.lfnam2) then

! Maybe the file name extension was omitted?

         sourcemask = sourcemask(1:iobjs)//'.fits'
         iobjs = lastc ( sourcemask )
         inquire ( file=sourcemask(1:iobjs), exist=lfnam2 )
       endif

       isp1 = lasts ( originback ) + 1

       outname = originback(1:max(iorigin-5,1))//'.cb.fits'
       if (arg3(1:2) .eq. '-o') outname = arg4
       if (arg4(1:2) .eq. '-o') outname = arg5
       if (arg5(1:2) .eq. '-o') outname = arg6
       ion = lastc ( outname )
       if (outname(max(ion-4,1):ion) .ne. '.fits') then
         outname = outname(1:ion)//'.fits'
         ion = lastc ( outname )
       endif              
       icb = index ( outname(1:ion), '.cb.' )
       cleanback = outname(1:ion)
       if (icb .eq. 0) then
         if (outname(max(ion-8,1):ion) .ne. '.cb.fits') then
           cleanback = outname(1:max(ion-5,1))//'.cb.fits'
         endif
       endif
       iclean = lastc ( cleanback )
       icb = index ( cleanback(1:iclean), '.cb.' )
       worker = cleanback
       worker(icb:icb+3) = '.wk.'
       iwork = lastc ( worker )

       if (lverbose .eq. '-verb1') write (iotty,'(/a)') ' CLEANBG: '//arg1(1:ia1)//' '//originback(1:iorigin-5)
     &                                                //' '//sourcemask(1:iobjs-5)
       if (.not.lfnam1) then
         write (iotty,'(/a)') '   CLEANBG: ERROR: File '''//originback(1:iorigin-5)//''' not found.'
         stop 1
       endif
       if (.not.lfnam2) then
         write (iotty,'(/a)') '   CLEANBG: ERROR: File '''//sourcemask(1:iobjs-5)//''' not found.'
         stop 2
       endif

! Determine numbers of pixels in the FITS image.

       call getnumpix ( originback(1:iorigin), nx1, ny1 )

       allocate ( forigin(nx1,ny1), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   CLEANBG: ERROR: Trouble allocating memory (5).'
         stop 5
       endif

       if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   Reading image '''//originback(1:iorigin)//'''...'

       call rfits ( nx1, ny1, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, forigin, dx1, dy1, object, ra, dec
     &            , originback(1:iorigin), header, footer, iotty, 0, creator, beam1, funmin, funmax, blank, crota1, crota2
     &            , cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history, lverbose )

       if (lverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax

! Determine numbers of pixels in the FITS image.

       call getnumpix ( sourcemask(1:iobjs), nx2, ny2 )

       allocate ( fsmask(nx2,ny2), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   CLEANBG: ERROR: Trouble allocating memory (6).'
         stop 6
       endif

       if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Reading image '''//sourcemask(1:iobjs)//'''...'

       call rfits ( nx2, ny2, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, fsmask, dx2, dy2, object, ra, dec
     &            , sourcemask(1:iobjs), header, footer, iotty, 0, creator, beam2, funmin, funmax, blank, crota1, crota2
     &            , cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history, lverbose )

       if (lverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax

       if (beam1 .eq. beam2) then
         beam = beam1
       else
         beam = 0.0
       endif

       if (nx1 .ne. nx2 .or. ny1 .ne. ny2 .or. abs ( dx1 - dx2 ) .gt. 0.001d0 * dx2 .or. abs ( dy1 - dy2 ) .gt. 0.001d0 * dy2) then
         write (iotty, '(/a)') '   CLEANBG: ERROR: The images have different parameters.'
         stop 99
       else
         nx = nx1
         ny = ny1
         dx = dx1
         dy = dy1
         rx = crpix1
         ry = crpix2
         xr = crval1
         yr = crval2
       endif
              
       nmax = max ( nx, ny )
       dnx = 0
       dny = 0
       if (nx .gt. ny) dny = nx - ny
       if (ny .gt. nx) dnx = ny - nx
       dnxa = dnx / 2
       dnya = dny / 2
       dnxb = dnxa
       dnyb = dnya
       if (mod ( dnx, 2 ) .gt. 0) dnxa = dnxa + 1
       if (mod ( dny, 2 ) .gt. 0) dnya = dnya + 1

       allocate ( fwork1(nx,ny), fwork2(nx,ny), fwork3(nx,ny), fwork4(nx,ny), fclean(nx,ny), fwork(nx,ny)
     &          , fclean1(nx,ny), fclean2(nx,ny), fclean3(nx,ny), fclean4(nx,ny), fun(nx*ny), funmed(nx,ny), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   CLEANBG: ERROR: Trouble allocating memory (10).'
         stop 10
       endif

       if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Cleaning background by cutting all sources off its face...'

       do j=1,ny
         do i=1,nx
           fwork  (i,j) = 0.0d0
           fwork1 (i,j) = 0.0d0
           fwork2 (i,j) = 0.0d0
           fwork3 (i,j) = 0.0d0
           fwork4 (i,j) = 0.0d0
           fclean (i,j) = forigin(i,j)
           fclean1(i,j) = forigin(i,j)
           fclean2(i,j) = forigin(i,j)
           fclean3(i,j) = forigin(i,j)
           fclean4(i,j) = forigin(i,j)
         enddo
       enddo

! Here the 1-pixel border around the entire mask image is zeroed, aiding in interpolations below.
       
       do i=1,nx
         fsmask(i,1 ) = 0.0d0
         fsmask(i,ny) = 0.0d0
         forigin(i,1 ) = 0.0d0
         forigin(i,ny) = 0.0d0
       enddo
       do j=1,ny
         fsmask(1 ,j) = 0.0d0
         fsmask(nx,j) = 0.0d0
         forigin(1 ,j) = 0.0d0
         forigin(nx,j) = 0.0d0
       enddo
!-------------------------------------------------
! Find background and interpolate in the X-direction.
!-------------------------------------------------
       iz1 = 0
       iz2 = 0
       im1 = 1
       do j=1,ny
         do i=1,nx
           if (fsmask(im1,j) .le. almostzero .and. fsmask(i,j) .gt. almostzero) iz1 = max ( im1 - iadd, 1  )
           if (fsmask(im1,j) .gt. almostzero .and. fsmask(i,j) .le. almostzero) iz2 = min ( i   + iadd, nx )
           if (iz2 .gt. 0) then
             jjm  = max (  j  - nave, 1  )               
             jjp  = min (  j  + nave, ny )
             iim1 = max ( iz1 - nave, 1  )
             iip1 = min ( iz1 + nave, nx )
             iim2 = max ( iz2 - nave, 1  )
             iip2 = min ( iz2 + nave, nx )

! Average start interpolation pixel value 'f1'.

             nav = 0
             f1 = 0.0d0
             do ja=jjm,jjp
               do ia=iim1,iip1
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f1 = f1 + forigin(ia,ja)
                 endif
               enddo
             enddo
             f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

             nav = 0
             f2 = 0.0d0
             do ja=jjm,jjp
               do ia=iim2,iip2
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f2 = f2 + forigin(ia,ja)
                 endif
               enddo
             enddo        
             f2 = f2 / dble ( nav )
             
! Interpolate.

             length1 = dble ( iz2 - iz1 + 1 )
             df = (f2 - f1) / dble ( iz2 - iz1 )
             do k=iz1,iz2                 
               fint = f1 + df * dble ( k - iz1 )
               fwork1 (k,j) = length1
               fclean1(k,j) = fint
             enddo 

! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
! appearance of jumps in values at the edge of the mask.

             do ja=jjm,jjp
               do ia=iim1,iip1
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork1 (ia,ja) = length1
                   fclean1(ia,ja) = f1
                 endif
               enddo
             enddo
             do ja=jjm,jjp
               do ia=iim2,iip2
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork1 (ia,ja) = length1
                   fclean1(ia,ja) = f2
                 endif
               enddo
             enddo        

             fclean2(iz1,j) = f1
             fclean3(iz1,j) = f1
             fclean4(iz1,j) = f1
             fclean2(iz2,j) = f2
             fclean3(iz2,j) = f2
             fclean4(iz2,j) = f2
             iz1 = 0
             iz2 = 0 
           endif
           im1 = i
         enddo
         iz1 = 0
         iz2 = 0
         im1 = 1
       enddo
!-------------------------------------------------
! Find background and interpolate in the Y-direction.
!-------------------------------------------------
       jz1 = 0
       jz2 = 0
       jm1 = 1
       do i=1,nx
         do j=1,ny
           if (fsmask(i,jm1) .le. almostzero .and. fsmask(i,j) .gt. almostzero) jz1 = max ( jm1 - jadd, 1  )
           if (fsmask(i,jm1) .gt. almostzero .and. fsmask(i,j) .le. almostzero) jz2 = min ( j   + jadd, ny )
           if (jz2 .gt. 0) then
             iim  = max (  i  - nave, 1  )               
             iip  = min (  i  + nave, nx )
             jjm1 = max ( jz1 - nave, 1  )
             jjp1 = min ( jz1 + nave, ny )
             jjm2 = max ( jz2 - nave, 1  )
             jjp2 = min ( jz2 + nave, ny )

! Average start interpolation pixel value 'f1'.

             nav = 0
             f1 = 0.0d0
             do ia=iim,iip
               do ja=jjm1,jjp1
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f1 = f1 + forigin(ia,ja)
                 endif
               enddo
             enddo
             f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

             nav = 0
             f2 = 0.0d0
             do ia=iim,iip
               do ja=jjm2,jjp2
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f2 = f2 + forigin(ia,ja)
                 endif
               enddo
             enddo        
             f2 = f2 / dble ( nav )

! Interpolate.

             length2 = dble ( jz2 - jz1 + 1 )
             df = (f2 - f1) / dble ( jz2 - jz1 )
             do k=jz1,jz2
               fint = f1 + df * dble ( k - jz1 )
               fwork2 (i,k) = length2
               fclean2(i,k) = fint
             enddo 

! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
! appearance of jumps in values at the edge of the mask.

             do ia=iim,iip
               do ja=jjm1,jjp1
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork2 (ia,ja) = length2
                   fclean2(ia,ja) = f1
                 endif
               enddo
             enddo
             do ia=iim,iip
               do ja=jjm2,jjp2
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork2 (ia,ja) = length2
                   fclean2(ia,ja) = f2
                 endif
               enddo
             enddo        

             fclean1(i,jz1) = f1
             fclean3(i,jz1) = f1
             fclean4(i,jz1) = f1
             fclean1(i,jz2) = f2
             fclean3(i,jz2) = f2
             fclean4(i,jz2) = f2
             jz1 = 0
             jz2 = 0 
           endif
           jm1 = j
         enddo
         jz1 = 0
         jz2 = 0
         jm1 = 1
       enddo
!-----------------------------------------------------------------------------
! Find background and interpolate in the forward diagonal direction (upper part).
!-----------------------------------------------------------------------------
       iz1 = 0
       iz2 = 0
       jz1 = 0
       jz2 = 0

       im1 = 1

       do j=nmax,(1-dnya),-1

         jm1 = max ( j - 1, 1 )
   
         do i=(1-dnxa),nmax-j+1

           ji = j + i - 1

           if (i .ge. 1 .and. i .le. nx .and. ji .ge. 1 .and. ji .le. ny) then

           if (fsmask(im1,jm1) .le. almostzero .and. fsmask(i,ji) .gt. almostzero) then 
             iz1 = max ( im1 - iadd, 1 )
             jz1 = max ( jm1 - jadd, 1 )
           endif
           if (fsmask(im1,jm1) .gt. almostzero .and. fsmask(i,ji) .le. almostzero) then 
             iz2 = min ( i  + iadd, nx )
             jz2 = min ( ji + jadd, ny )
           endif
           if (jz2 .gt. 0) then
             iim1 = max ( iz1 - nave, 1  )
             iip1 = min ( iz1 + nave, nx )
             iim2 = max ( iz2 - nave, 1  )
             iip2 = min ( iz2 + nave, nx )
             jjm1 = max ( jz1 - nave, 1  )
             jjp1 = min ( jz1 + nave, ny )
             jjm2 = max ( jz2 - nave, 1  )
             jjp2 = min ( jz2 + nave, ny )

! Average start interpolation pixel value 'f1'.

             nav = 0
             f1 = 0.0d0
             do ja=jjm1,jjp1
               do ia=iim1,iip1
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f1 = f1 + forigin(ia,ja)
                 endif
               enddo
             enddo
             f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

             nav = 0
             f2 = 0.0d0
             do ja=jjm2,jjp2
               do ia=iim2,iip2
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f2 = f2 + forigin(ia,ja)
                 endif
               enddo
             enddo        
             f2 = f2 / dble ( nav )

! Interpolate.

             length3 = sqrt ( dble ( iz2 - iz1 + 1 )**2 + dble ( jz2 - jz1 + 1 )**2 )
             df = (f2 - f1) / dble ( jz2 - jz1 )
             l = 0
             do k=jz1,jz2             
               fint = f1 + df * dble ( k - jz1 )
               fwork3 (iz1+l,k) = length3
               fclean3(iz1+l,k) = fint
               l = l + 1
             enddo 

! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
! appearance of jumps in values at the edge of the mask.

             do ja=jjm1,jjp1
               do ia=iim1,iip1
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork3 (ia,ja) = length3
                   fclean3(ia,ja) = f1
                 endif
               enddo
             enddo
             do ja=jjm2,jjp2
               do ia=iim2,iip2
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork3 (ia,ja) = length3
                   fclean3(ia,ja) = f2
                 endif
               enddo
             enddo        

             fclean1(iz1,jz1) = f1
             fclean2(iz1,jz1) = f1
             fclean4(iz1,jz1) = f1
             fclean1(iz2,jz2) = f2
             fclean2(iz2,jz2) = f2
             fclean4(iz2,jz2) = f2
             iz1 = 0
             iz2 = 0 
             jz1 = 0
             jz2 = 0 
           endif
           im1 = i
           jm1 = ji

           endif

         enddo
         iz1 = 0
         iz2 = 0
         jz1 = 0
         jz2 = 0
         im1 = 1
       enddo
!-----------------------------------------------------------------------------
! Find background and interpolate in the forward diagonal direction (lower part).
!-----------------------------------------------------------------------------
       iz1 = 0
       iz2 = 0
       jz1 = 0
       jz2 = 0

       jm1 = 1
       
       do i=nmax,(2-dnxa),-1
         
         im1 = max ( i - 1, 1 )

         do j=(1-dnya),nmax-i+1

           ij = i + j - 1

           if (ij .ge. 1 .and. ij .le. nx .and. j .ge. 1 .and. j .le. ny) then

           if (fsmask(im1,jm1) .le. almostzero .and. fsmask(ij,j) .gt. almostzero) then 
             iz1 = max ( im1 - iadd, 1 )
             jz1 = max ( jm1 - jadd, 1 )
           endif
           if (fsmask(im1,jm1) .gt. almostzero .and. fsmask(ij,j) .le. almostzero) then 
             iz2 = min ( ij + iadd, nx )
             jz2 = min ( j  + jadd, ny )
           endif
           if (iz2 .gt. 0) then
             iim1 = max ( iz1 - nave, 1  )
             iip1 = min ( iz1 + nave, nx )
             iim2 = max ( iz2 - nave, 1  )
             iip2 = min ( iz2 + nave, nx )
             jjm1 = max ( jz1 - nave, 1  )
             jjp1 = min ( jz1 + nave, ny )
             jjm2 = max ( jz2 - nave, 1  )
             jjp2 = min ( jz2 + nave, ny )

! Average start interpolation pixel value 'f1'.

             nav = 0
             f1 = 0.0d0
             do ja=jjm1,jjp1
               do ia=iim1,iip1
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f1 = f1 + forigin(ia,ja)
                 endif
               enddo
             enddo
             f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

             nav = 0
             f2 = 0.0d0
             do ja=jjm2,jjp2
               do ia=iim2,iip2
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f2 = f2 + forigin(ia,ja)
                 endif
               enddo
             enddo        
             f2 = f2 / dble ( nav )

! Interpolate.

             length3 = sqrt ( dble ( iz2 - iz1 + 1 )**2 + dble ( jz2 - jz1 + 1 )**2 )
             df = (f2 - f1) / dble ( iz2 - iz1 )
             l = 0
             do k=iz1,iz2             
               fint = f1 + df * dble ( k - iz1 )
               fwork3 (k,jz1+l) = length3
               fclean3(k,jz1+l) = fint
               l = l + 1
             enddo 

! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
! appearance of jumps in values at the edge of the mask.

             do ja=jjm1,jjp1
               do ia=iim1,iip1
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork3 (ia,ja) = length3
                   fclean3(ia,ja) = f1
                 endif
               enddo
             enddo
             do ja=jjm2,jjp2
               do ia=iim2,iip2
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork3 (ia,ja) = length3
                   fclean3(ia,ja) = f2
                 endif
               enddo
             enddo        

             fclean1(iz1,jz1) = f1
             fclean2(iz1,jz1) = f1
             fclean4(iz1,jz1) = f1
             fclean1(iz2,jz2) = f2
             fclean2(iz2,jz2) = f2
             fclean4(iz2,jz2) = f2
             iz1 = 0
             iz2 = 0 
             jz1 = 0
             jz2 = 0 
           endif
           im1 = ij
           jm1 = j

           endif

         enddo
         iz1 = 0
         iz2 = 0
         jz1 = 0
         jz2 = 0
         jm1 = 1
       enddo
!-----------------------------------------------------------------------------
! Find background and interpolate in the backward diagonal direction (upper part).
!-----------------------------------------------------------------------------
       iz1 = 0
       iz2 = 0
       jz1 = 0
       jz2 = 0
       
       im1 = nx

       do j=nmax,(1-dnya),-1

         jm1 = max ( j - 1, 1 )

         do i=(1-dnxa),nmax-j+1

           ii = nx - i + 1
           ji = j + i - 1

           if (ii .ge. 1 .and. ii .le. nx .and. ji .ge. 1 .and. ji .le. ny) then

           if (fsmask(im1,jm1) .le. almostzero .and. fsmask(ii,ji) .gt. almostzero) then 
             iz1 = min ( im1 + iadd, nx )
             jz1 = max ( jm1 - jadd, 1  )
           endif
           if (fsmask(im1,jm1) .gt. almostzero .and. fsmask(ii,ji) .le. almostzero) then 
             iz2 = max ( ii - iadd, 1  )
             jz2 = min ( ji + jadd, ny )
           endif
           if (jz2 .gt. 0) then
             iim1 = max ( iz1 - nave, 1  )
             iip1 = min ( iz1 + nave, nx )
             iim2 = max ( iz2 - nave, 1  )
             iip2 = min ( iz2 + nave, nx )
             jjm1 = max ( jz1 - nave, 1  )
             jjp1 = min ( jz1 + nave, ny )
             jjm2 = max ( jz2 - nave, 1  )
             jjp2 = min ( jz2 + nave, ny )

! Average start interpolation pixel value 'f1'.

             nav = 0
             f1 = 0.0d0
             do ja=jjm1,jjp1
               do ia=iim1,iip1
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f1 = f1 + forigin(ia,ja)
                 endif
               enddo
             enddo
             f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

             nav = 0
             f2 = 0.0d0
             do ja=jjm2,jjp2
               do ia=iim2,iip2
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f2 = f2 + forigin(ia,ja)
                 endif
               enddo
             enddo        
             f2 = f2 / dble ( nav )

! Interpolate.

             length4 = sqrt ( dble ( iz2 - iz1 + 1 )**2 + dble ( jz2 - jz1 + 1 )**2 )
             df = (f2 - f1) / dble ( jz2 - jz1 )
             l = 0
             do k=jz1,jz2             
               fint = f1 + df * dble ( k - jz1 )
               fwork4 (iz1-l,k) = length4
               fclean4(iz1-l,k) = fint
               l = l + 1
             enddo 

! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
! appearance of jumps in values at the edge of the mask.

             do ja=jjm1,jjp1
               do ia=iim1,iip1
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork4 (ia,ja) = length4
                   fclean4(ia,ja) = f1
                 endif
               enddo
             enddo
             do ja=jjm2,jjp2
               do ia=iim2,iip2
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork4 (ia,ja) = length4
                   fclean4(ia,ja) = f2
                 endif
               enddo
             enddo        

             fclean1(iz1,jz1) = f1
             fclean2(iz1,jz1) = f1
             fclean3(iz1,jz1) = f1
             fclean1(iz2,jz2) = f2
             fclean2(iz2,jz2) = f2
             fclean3(iz2,jz2) = f2
             iz1 = 0
             iz2 = 0 
             jz1 = 0
             jz2 = 0 
           endif
           im1 = ii
           jm1 = ji
       
           endif
       
         enddo
         iz1 = 0
         iz2 = 0
         jz1 = 0
         jz2 = 0
         im1 = nx
       enddo
!-----------------------------------------------------------------------------
! Find background and interpolate in the backward diagonal direction (lower part).
!-----------------------------------------------------------------------------
       iz1 = 0
       iz2 = 0
       jz1 = 0
       jz2 = 0
       
       jm1 = 1

       do i=(1-dnxa),nmax-1
         
         im1 = min ( i + 1, nx )

         do j=(1-dnya),i
           
           ij = i - j + 1

           if (ij .ge. 1 .and. ij .le. nx .and. j .ge. 1 .and. j .le. ny) then

           if (fsmask(im1,jm1) .le. almostzero .and. fsmask(ij,j) .gt. almostzero) then 
             iz1 = min ( im1 + iadd, nx )
             jz1 = max ( jm1 - jadd, 1  )
           endif
           if (fsmask(im1,jm1) .gt. almostzero .and. fsmask(ij,j) .le. almostzero) then 
             iz2 = max ( ij - iadd, 1  )
             jz2 = min ( j  + jadd, ny )
           endif
           if (iz2 .gt. 0) then
             iim1 = max ( iz1 - nave, 1  )
             iip1 = min ( iz1 + nave, nx )
             iim2 = max ( iz2 - nave, 1  )
             iip2 = min ( iz2 + nave, nx )
             jjm1 = max ( jz1 - nave, 1  )
             jjp1 = min ( jz1 + nave, ny )
             jjm2 = max ( jz2 - nave, 1  )
             jjp2 = min ( jz2 + nave, ny )

! Average start interpolation pixel value 'f1'.

             nav = 0
             f1 = 0.0d0
             do ja=jjm1,jjp1
               do ia=iim1,iip1
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f1 = f1 + forigin(ia,ja)
                 endif
               enddo
             enddo
             f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

             nav = 0
             f2 = 0.0d0
             do ja=jjm2,jjp2
               do ia=iim2,iip2
                 if (fsmask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f2 = f2 + forigin(ia,ja)
                 endif
               enddo
             enddo        
             f2 = f2 / dble ( nav )

! Interpolate.

             length4 = sqrt ( dble ( iz2 - iz1 + 1 )**2 + dble ( jz2 - jz1 + 1 )**2 )
             df = (f2 - f1) / dble ( iz2 - iz1 )
             l = 0
             do k=iz1,iz2,-1          
               fint = f1 + df * dble ( k - iz1 )
               fwork4 (k,jz1+l) = length4
               fclean4(k,jz1+l) = fint
               l = l + 1
             enddo 

! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
! appearance of jumps in values at the edge of the mask.

             do ja=jjm1,jjp1
               do ia=iim1,iip1
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork4 (ia,ja) = length4
                   fclean4(ia,ja) = f1
                 endif
               enddo
             enddo
             do ja=jjm2,jjp2
               do ia=iim2,iip2
                 if (fsmask(ia,ja) .le. almostzero) then
                   fwork4 (ia,ja) = length4
                   fclean4(ia,ja) = f2
                 endif
               enddo
             enddo        
             
             fclean1(iz1,jz1) = f1
             fclean2(iz1,jz1) = f1
             fclean3(iz1,jz1) = f1
             fclean1(iz2,jz2) = f2
             fclean2(iz2,jz2) = f2
             fclean3(iz2,jz2) = f2
             iz1 = 0
             iz2 = 0 
             jz1 = 0
             jz2 = 0 
           endif
           im1 = ij
           jm1 = j

           endif

         enddo
         iz1 = 0
         iz2 = 0
         jz1 = 0
         jz2 = 0
         jm1 = 1
       enddo
!-----------------------------------------------------------------------------
! Averaging all the interpolations.
!-----------------------------------------------------------------------------
!!       nmed = 4
!!       if (larithm) then
         nsort = 4

         do j=1,ny
           do i=1,nx
             funw(1) = fwork1(i,j)
             funw(2) = fwork2(i,j)
             funw(3) = fwork3(i,j)
             funw(4) = fwork4(i,j)
             func(1) = fclean1(i,j)
             func(2) = fclean2(i,j)
             func(3) = fclean3(i,j)
             func(4) = fclean4(i,j)

             call sortarr ( nsort, funw, func )
             
             fwork(i,j) = funw(1)
             fclean(i,j) = func(1)
           enddo
         enddo
         
         npts = 10
         npts2 = npts**2

         if (2*npts+1 .lt. 10) then
           if (lverbose .eq. '-verb2') write (iotty,'(a,i1,a)') '   Median filtering in a ', 2*npts+1, '-pixel sliding window...'
         elseif (2*npts+1 .lt. 100) then
           if (lverbose .eq. '-verb2') write (iotty,'(a,i2,a)') '   Median filtering in a ', 2*npts+1, '-pixel sliding window...'
         else  
           if (lverbose .eq. '-verb2') write (iotty,'(a,i3,a)') '   Median filtering in a ', 2*npts+1, '-pixel sliding window...'
         endif

         do j=1,ny
           do i=1,nx
             funmed(i,j) = fclean(i,j)
             nmed = 0
             if (fsmask(i,j) .gt. almostzero) then
               npmax = npts2
               do m=max(i-npts,1),min(i+npts,nx)
                 mi2 = (m - i)**2
                 do n=max(j-npts,1),min(j+npts,ny)
                   nj2 = (n - j)**2
                   nr2 = mi2 + nj2
                   if (fsmask(m,n) .lt. almostzero) then
                     npmax = min ( npmax, nr2 )
                   endif
                 enddo
               enddo
               npmax = nint ( sqrt ( dble ( npmax ) ) )
               do m=max(i-npmax,1),min(i+npmax,nx)
                 do n=max(j-npmax,1),min(j+npmax,ny)
                   if (fsmask(m,n) .gt. almostzero) then
                     nmed = nmed + 1
                     fun(nmed) = fclean(m,n)
                   endif
                 enddo
               enddo
               if (nmed .gt. 1) then
  
                 call sortarr ( nmed, fun, fun )
         
                 funmed(i,j) = fun((nmed+1)/2)
               endif
             endif
           enddo
         enddo

         do j=1,ny
           do i=1,nx
             fclean(i,j) = funmed(i,j)
           enddo
         enddo

       deallocate ( fwork1, fwork2, fwork3, fwork4, fclean1, fclean2, fclean3, fclean4, fun, funmed )

! Special cases when masks touch a border.

       jz1 = 10
       do i=2,nx-1
         if (fsmask(i,jz1) .gt. almostzero) then
           fwmin = 1.0d+99
           do j=jz1,ny-1
             jz2 = j+1
             if (fsmask(i,j) .le. almostzero) then
               exit
             else
               fwmin = fclean(i,j)
             endif
           enddo
           if (jz2 .lt. ny) then
             do j=jz1-1,jz2
               fclean(i,j) = fwmin
             enddo
           endif
         endif
       enddo
       
       jz1 = ny - 9
       do i=2,nx-1
         if (fsmask(i,jz1) .gt. almostzero) then
           fwmin = 1.0d+99
           do j=jz1,2,-1
             jz2 = j-1
             if (fsmask(i,j) .le. almostzero) then
               exit
             else
               fwmin = fclean(i,j)
             endif
           enddo
           if (jz2 .gt. 1) then
             do j=jz1+1,jz2,-1
               fclean(i,j) = fwmin
             enddo
           endif
         endif
       enddo        
       
       iz1 = 10
       do j=2,ny-1
         if (fsmask(iz1,j) .gt. almostzero) then
           fwmin = 1.0d+99
           do i=iz1,nx-1
             iz2 = i+1
             if (fsmask(i,j) .le. almostzero) then
               exit
             else
               fwmin = fclean(i,j)
             endif
           enddo
           if (iz2 .lt. nx) then
             do i=iz1-1,iz2
               fclean(i,j) = fwmin
             enddo
           endif
         endif
       enddo

       iz1 = nx - 9
       do j=2,ny-1
         if (fsmask(iz1,j) .gt. almostzero) then
           fwmin = 1.0d+99
           do i=iz1,2,-1
             iz2 = i-1
             if (fsmask(i,j) .le. almostzero) then
               exit
             else
               fwmin = fclean(i,j)
             endif
           enddo
           if (iz2 .gt. 1) then
             do i=iz1+1,iz2,-1
               fclean(i,j) = fwmin
             enddo
           endif
         endif
       enddo

! Get the current date and time.

       call when ( lunix, ctime, cdate, ndate, 4 )

       if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Writing masked interpolation '''//worker(1:iwork)//'''...'

       header = worker(1:iwork)
       
       call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, fwork, dx, dy, object, crval1
     &            , crval2, worker(1:iwork), header, footer, cdate, ctime, creator, beam, blank, crota1, crota2, cd11, cd12, cd21
     &            , cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history )

       if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Writing cleaned background '''//cleanback(1:iclean)//'''...'

       header = cleanback(1:iclean)
       
       call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, fclean, dx, dy, object, crval1
     &            , crval2, cleanback(1:iclean), header, footer, cdate, ctime, creator, beam, blank, crota1, crota2, cd11, cd12
     &            , cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history )

       deallocate ( forigin, fclean, fsmask, fwork )

       if (lverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, header, footer, iotty
     &            , iolog, creator, beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale
     &            , wave, datamn, datamx, history, lverbose )
!__________________________________________________________________________________________________________________________________
!
! GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay      
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, anynull, myfile
       character*(*) ctype1, ctype2, creator, fname, header, footer(2), bunit, object, history, lverbose
       character*25  cdatetime
       character*80  footer1, footer2, comment
       integer       i, j, nx, ny, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, lastc, iotty, iolog, maxdim
     &             , ilc1, ilc2, rw, nfound, blank, stat01, stat02, stat03, stat04, stat05, stat06, stat07, stat08, stat09
     &             , stat10, stat11, stat12, stat13, stat14, stat15, stat16, stat17, stat18, stat19, stat20, stat21, stat22
     &             , stat23, stat24, stat25, stat26, stat27, stat28, stat29, stat30, stat31, stat32, stat0, nan
       parameter   ( maxdim = 2 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, cd11, cd12, cd21, cd22
       real*8        fun(nx,ny)
       real*8        as2deg, datamx, datamn
       parameter   ( as2deg = 1.0d0 / 3600.0d0 )
       external      lastc, ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat01 = 0
       stat02 = 0
       stat03 = 0
       stat04 = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
       if (stat .eq. 0 .and. (comment(1:22) .eq. 'Alexander Men`shchikov' .or. comment(1:22) .eq. 'Alexander Men''shchikov')) then
         myfile = .true.
       else
         myfile = .false.
         stat = 0
         creator = 'CLEANBG'
       endif

       header = 'FITS Plot Header'

       if (myfile) then
         call ftgkys ( unit, 'HEADER' , header , comment, stat01 )
         call ftgkys ( unit, 'FOOTER1', footer1, comment, stat02 )
         call ftgkys ( unit, 'FOOTER2', footer2, comment, stat03 )
         ilc1 = lastc ( footer1 )
         ilc2 = lastc ( footer2 )
         if (ilc1 .ne. 0 .and. ilc2 .ne. 0) then
           footer(1) = footer1(1:ilc1)//footer2(1:ilc2)
         else
           footer(1) = 'Footer'
         endif
       else
         call ftgkys ( unit, 'OBJECT' , header, comment, stat04 )
         if (stat04 .ne. 0 .or. header(1:10) .eq. '          ') header = fname
         footer(1) = header
       endif

! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       stat08 = 0
       stat0 = stat01 + stat02 + stat03 + stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15
     &       + stat16 + stat17 + stat18 + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28
     &       + stat29 + stat30 + stat31

       if (lverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat01 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword HEADER absent or error'
         if (stat02 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER1 absent or error'
         if (stat03 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER2 absent or error'
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat29 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EPOCH absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = 'No object specified'

! Read the FITS file into the 2D array.

       call ftg2dd ( unit, group, dble ( blank ), naxes(1), naxes(1), naxes(2), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       nx = naxes(1)
       ny = naxes(2)
       funmin = 1.0d30
       funmax =-1.0d30
       nan = 0
       do j=1,ny
         do i=1,nx
           if (isnan ( fun(i,j) )) then  
             nan = nan + 1
             if (lverbose .ne. '-verb0') then
               if (nan .lt. 100) then
                 write (iotty,'(a,i4,a,i4,a)') '   RFITS: Element (', i,',', j, ') contains not a number (NaN)...'
               elseif (nan .eq. 100) then
                 write (iotty,'(a)') '   RFITS: Many NaNs, no more info on the elements.'
               endif
             endif
           else
             funmin = min ( funmin, fun(i,j) )
             funmax = max ( funmax, fun(i,j) )
           endif  
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, header
     &            , footer, cdate, ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale
     &            , wave, datamn, datamx, history )
!__________________________________________________________________________________________________________________________________
!
! GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay      
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend
       character*(*) fname, ctype1, ctype2, object, creator, header, footer(2), ctime, cdate, bunit, cfitsversion, history
       character*19  cdatetime
       integer       i, j, nx, ny, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, naxes(2), iat, iflen
     &             , lastc, blank
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, dfun, rot1, rot2, epoch, equinox, datamn
     &             , cosrota, sinrota, rotangle, datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, pi
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      lastc
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 2-D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 2
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       datamn = 1.0d30
       datamx =-1.0d30
       do j=1,ny
         do i=1,nx
           dfun = fun(i,j)
           if (dfun .lt. datamn ) datamn = dfun
           if (dfun .gt. datamx ) datamx = dfun
         enddo
       enddo

       bitpix = -32     !<-- single precision floating-point values

       cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

       if (cfitsversion .eq. '5.030') then
         call fdelete ( fname, status )
       endif

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Create the new empty FITS file.

       if (cfitsversion .eq. '5.030') then
         call ftinit ( unit, fname, blocksze, status )
       else
         call ftinit ( unit, '!'//fname, blocksze, status )
       endif

! Write the required header keywords.

       call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

       call ftpkys ( unit, 'CREATOR', creator,'Alexander Men`shchikov, SAp IRFU CEA Saclay', status )
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'HEADER' , header, 'title', status )

       iat = 0
       iat = index ( footer(1), ' |F' )
       if (iat .eq. 0) iat = index ( footer(1), ' mm ' )
       iflen = lastc ( footer(1) )

       if (iat .gt. 0) then
         call ftpkys ( unit, 'FOOTER1', footer(1)(1:iat-1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', footer(1)(iat:iflen), 'part 2', status )
       else
         call ftpkys ( unit, 'FOOTER1', footer(1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', ' ', 'part 2', status )
       endif
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )

! Write all parameters.

       dxdeg = dx * as2deg
       dydeg = dy * as2deg

! Correct if the input images had cd11 or cd22 set to zero.

       if (abs ( cd11 ) .lt. almostzero .or. abs ( cd11 ) .gt. dxdeg) then
         cd11 = -dxdeg
         cd12 = 0.0d0
       endif
       if (abs ( cd22 ) .lt. almostzero .or. abs ( cd22 ) .gt. dxdeg) then
         cd21 = 0.0d0
         cd22 = dydeg
       endif
       
       cosrota = - cd11 / dxdeg
       sinrota = - cd12 / dxdeg
       rotangle = acos ( cosrota ) * 180.0d0 / pi
       
       if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = 180.0d0 - rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = 180.0d0 + rotangle
       elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = - rotangle
       endif
       rot2 = rot1

       call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
       call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
       call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
       if  (blank .gt. 0)
     & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
       call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
       call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
       call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
       call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CDELT1' , -dxdeg, 13 , 'coordinate increment along axis', status )
       call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
       call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
       call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
       call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
       call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
       call ftpkyd ( unit, 'EPOCH'  , epoch  , 13, ' ', status )
       call ftpkyd ( unit, 'EQUINOX', equinox, 13, ' ', status )
       call ftphis ( unit, history, status )

       if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'image was convolved with the beam (arcsec)', status )

! Write the array to the FITS file.

       call ftp2dd ( unit, group, nx, nx, ny, fun, status )

! Close the file.

       call ftclos ( unit, status )

! Free the unit number.

       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||