
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program ellipses
!__________________________________________________________________________________________________________________________________
!
! Draw extraction ellipses on top of original FITS images.
!
! This code is part of GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay  
!__________________________________________________________________________________________________________________________________
!
       implicit      none

       logical       lfname1, lfname2, lunix, inellipse, lfootprints
                                            
       character*2   cnwave
       character*5   cattype, cfpptime
       character*6   cfitsversion, lverbose  
       character*7   clibname
       character*8   ctime
       character*10  cdate
       character*21  compda
       character*80  object, creator, ctype1, ctype2, bunit, exmeth, history
       character*128 header, footer(2)
       character*500 arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, argA, argB, argC, argD, argE, argF, filename, outname
     &             , excat
       character*15000 cline, headline(1000)

       integer       i, j, l, k, isp1, fnlen1, ion, ndate, iotty, nx, ny, firstb, lastc, lasts, lhl, irc, lb, iexm, iexc, ia1, ia2
     &             , ia3, ia4, ia5, ia6, ia7, ia8, ia9, iaA, iaB, iaC, iaD, iaE, iaF, blank, nhead, nextr, ncolumns, nxe, nye
     &             , ncolumnx, ncolumny, ncolumna, ncolumnb, ncolumnt, ncolumnfp, ncolumnft, ncolumnr, ncolumnfpe, ncolumnfte
     &             , ncolumnsx, no, m, nxco, nyco, nr, l1, l2, m1, m2, nhcols, ncolumng

       parameter   ( iotty = 6 )

       real          fitsvers
       real*8        dx, dy, wave, crpix1, crpix2, crval1, crval2, ra, dec, beam, funmin, funmax, sq_arcsecs_per_sterad, pi, goodmin
     &             , ellvalfact, crota1, crota2, epoch, equinox, bzero, bscale, datamin, datamax, pio2, pi2, x0o, y0o, dxye, majdiam
     &             , mindiam, angle, radx, rady, obsbeam, thickfact, sigmono, sigmin, fxpbest, fxperro, fxtbest, fxterro, major1
     &             , major2, minor1, minor2, semimajor2, snr1min, snr2min, semiminor2, dbll, dblm, deltax, deltay, sizefact
     &             , goodness, cd11, cd12, cd21, cd22, deltaxdot, deltaydot

       real*8, allocatable :: funa(:,:), funb(:,:)

       parameter   ( sq_arcsecs_per_sterad = 3282.80635d0 * 3600.0d0**2, pi = 3.14159265358979d0, pio2 = pi / 2.0d0
     &             , pi2 = pi * 2.0d0 )

       external      firstb, lastc, rfits, wfits, when, lastbs, getnumpix
!__________________________________________________________________________________________________________________________________
!
! Determine operating system

       call osystem ( lunix )

       call ftvers ( fitsvers )

       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       compda = __DATE__//' '//cfpptime

! Get the current date and time.

       call when ( lunix, ctime, cdate, ndate, 4 )

! Get command line parameters (file names).

       call getarg (  1, arg1 )
       call getarg (  2, arg2 )
       call getarg (  3, arg3 )
       call getarg (  4, arg4 )
       call getarg (  5, arg5 )
       call getarg (  6, arg6 )
       call getarg (  7, arg7 )
       call getarg (  8, arg8 )
       call getarg (  9, arg9 )
       call getarg ( 10, argA )
       call getarg ( 11, argB )
       call getarg ( 12, argC )
       call getarg ( 13, argD )
       call getarg ( 14, argE )
       call getarg ( 15, argF )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )
       ia6 = lastc ( arg6 )
       ia7 = lastc ( arg7 )
       ia8 = lastc ( arg8 )
       ia9 = lastc ( arg9 )
       iaA = lastc ( argA )
       iaB = lastc ( argB )
       iaC = lastc ( argC )
       iaD = lastc ( argD )
       iaE = lastc ( argE )
       iaF = lastc ( argF )
       
       lverbose = '-verb2'
       if (iaF .gt. 0 .and. argF(1:iaF-1) .eq. '-verb') lverbose = argF(1:iaF)
       if (iaE .gt. 0 .and. argE(1:iaE-1) .eq. '-verb') lverbose = argE(1:iaE)
       if (iaD .gt. 0 .and. argD(1:iaD-1) .eq. '-verb') lverbose = argD(1:iaD)
       if (iaC .gt. 0 .and. argC(1:iaC-1) .eq. '-verb') lverbose = argC(1:iaC)
       if (iaB .gt. 0 .and. argB(1:iaB-1) .eq. '-verb') lverbose = argB(1:iaB)
       if (iaA .gt. 0 .and. argA(1:iaA-1) .eq. '-verb') lverbose = argA(1:iaA)
       
       if (lverbose .eq. '-verb2') write (iotty,'(/a$)') ' ELLIPSES = Overlay Extraction Ellipses = '//compda
       if (arg1(1:ia1) .eq. ':') stop
       if (lverbose .eq. '-verb2') write (iotty,'()')
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Alexander Men`shchikov, SAp IRFU CEA Saclay, France.'
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                                //' by William D Pence.'
       if (ia6 .gt. 0) then
         exmeth = arg1(1:ia1)
         iexm = ia1
         cnwave = '  '
         read (arg2(1:ia2),'(a)',err=1) cnwave
         if (cnwave(2:2) .eq. ' ') cnwave = '0'//cnwave(1:1)
         if (arg2(1:ia2) .eq. '0') cnwave = ' '
         goto 2
  1      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting CNWAVE.'
           stop 99
  2      continue
         read (arg3(1:ia3),*,err=3) obsbeam
         goto 4
  3      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting OBSBEAM.'
           stop 99
  4      continue
         read (arg4(1:ia4),*,err=5) goodmin
         goto 6
  5      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting GOODMIN.'
           stop 99
  6      continue
         read (arg5(1:ia5),*,err=7) sigmin
         goto 8
  7      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting SIGMIN.'
           stop 99
  8      continue
         read (arg6(1:ia6),*,err=9) snr1min
         goto 10
  9      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting SNR1MIN.'
           stop 99
 10      continue
         read (arg7(1:ia7),*,err=11) snr2min
         goto 12
 11      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting SNR2MIN.'
           stop 99
 12      continue
         excat = arg8(1:ia8)
         filename = arg9(1:ia9)
       else
         write (iotty,'(/a)') ' USAGE: ellipses <exmeth> <cnwave> <obsbeam> <goodmin> <sigmin> <snr1min> <snr2min> <excat> <filein>'
     &                      //' [<sizefact> <ellvalfact> <thickfact>] [-o <outname>] [-verb{0|1|2}]'
         stop 99
       endif  

       iexc = lastc ( excat )
       fnlen1 = lastc ( filename )

! Check if the input file exists and open it.

       inquire ( file=excat(1:iexc), exist=lfname1 )
       
! Check if the input file exists and open it.

       inquire ( file=filename(1:fnlen1), exist=lfname2 )
       if (.not.lfname2) then

! Maybe the file name extension was omitted?

         filename = filename(1:fnlen1)//'.fits'
         fnlen1 = lastc ( filename )

         inquire ( file=filename(1:fnlen1), exist=lfname2 )
       endif

       isp1 = lasts ( filename ) + 1

       outname = filename(isp1:fnlen1-5)//'.m.fits'
       if (argA(1:2) .eq. '-o') outname = argB
       if (argB(1:2) .eq. '-o') outname = argC
       if (argC(1:2) .eq. '-o') outname = argD
       if (argD(1:2) .eq. '-o') outname = argE
       ion = lastc ( outname )
       if (outname(max(1,ion-4):ion) .ne. '.fits') then
         outname = outname(1:ion)//'.fits'
         ion = lastc ( outname )
       endif

       if (lverbose .eq. '-verb1') write (iotty,'(/a)') ' ELLIPSES: '//arg1(1:ia1)//' '//arg2(1:ia2)//' '//arg3(1:ia3)
     &                                                //' '//arg4(1:ia4)//' '//excat(1:iexc)

       if (.not.lfname1) then
         write (iotty,'(/a)') '   ELLIPSES: ERROR: File '''//excat(1:iexc)//''' not found.'
         stop 1
       endif
       if (.not.lfname2) then
         write (iotty,'(/a)') '   ELLIPSES: ERROR: File '''//filename(1:fnlen1-5)//''' not found.'
         stop 2
       endif

! Determine numbers of pixels in the FITS image.

       call getnumpix ( filename(1:fnlen1), nx, ny )

! Allocate the image arrays.
       
       allocate ( funa(nx,ny), funb(nx,ny), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble allocating memory (12).'
         stop 12
       endif
       
       if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   Reading image '''//filename(1:fnlen1)//'''...'

       call rfits ( nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, funa, dx, dy, object, ra, dec
     &            , filename(1:fnlen1), header, footer, iotty, 0, creator, beam, funmin, funmax, blank, crota1, crota2, cd11, cd12
     &            , cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history, lverbose )

       if (lverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax

       do j=1,ny
         do i=1,nx
           funb(i,j) = funa(i,j)
         enddo
       enddo
               
       dxye = dx
       sizefact = 1.0d0
       if (iaA .gt. 0 .and. argA(1:2) .ne. '-o') then
         read (argA,*,err=13) sizefact
         goto 14
 13      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting SIZEFACT.'
           stop 99
 14      continue
       endif
       ellvalfact = 1.1d0
       if (iaB .gt. 0 .and. argB(1:2) .ne. '-o') then
         read (argB,*,err=15) ellvalfact
         goto 16
 15      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting ELLVALFACT.'
           stop 99
 16      continue
       endif
       thickfact = 3.0d0
       lfootprints = .false.
       if (iaC .gt. 0 .and. argC(1:2) .ne. '-o') then
         read (argC,*,err=17) thickfact
         if (argC(1:1) .eq. '+') then
           lfootprints = .true.
         endif
         if (argC(1:1) .eq. '-') then
           thickfact = -2.0d0
         endif
         goto 18
 17      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting THICKFACT.'
           stop 99
 18      continue
       endif

       if (exmeth(1:iexm) .eq. 'sextractor' ) cattype = 'sextr'
       if (exmeth(1:iexm) .eq. 'getsources' ) cattype = 'getso'
       if (exmeth(1:iexm) .eq. 'clumpfind'  ) cattype = 'clump'
       if (exmeth(1:iexm) .eq. 'gaussclumps') cattype = 'gauss'
       if (exmeth(1:iexm) .eq. 'csar'       ) cattype = 'csar '
       if (exmeth(1:iexm) .eq. 'fellwalker' ) cattype = 'fellw'
       if (exmeth(1:iexm) .eq. 'reinhold'   ) cattype = 'reinh'
       if (exmeth(1:iexm) .eq. 'derivatives') cattype = 'deriv'
       if (exmeth(1:iexm) .eq. 'mre-gcl'    ) cattype = 'mregc'
       if (exmeth(1:iexm) .eq. 'truth'      ) cattype = 'truth'
              
       if (index ( excat(1:iexc), '.smf'       ) .gt. 0) cattype = 'smflx'
       if (index ( excat(1:iexc), '.final'     ) .gt. 0) cattype = 'final'
       if (index ( excat(1:iexc), '.semifinal' ) .gt. 0) cattype = 'semif'

       if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Reading catalog '''//excat(1:iexc)//'''...'

       open ( 15, file=excat(1:iexc), status='old' )

       call skipcomm     ( 15, excat(1:iexc), nhead )
       call linescolumns ( 15, excat(1:iexc), nhead, nextr, ncolumns )

       rewind ( 15 )
       do k=1,nhead
         read (15,'(a)') headline(k)
         lhl = lastc ( headline(k) )
         if (lhl .gt. 5) nhcols = k
       enddo
       lhl = lastc ( headline(nhcols) )

       ncolumnx = 0
       ncolumny = 0
       ncolumng = 0
       ncolumna = 0
       ncolumnb = 0
       ncolumnt = 0
       ncolumnfp = 0
       ncolumnfpe = 0
       ncolumnft = 0
       ncolumnfte = 0
       ncolumnr = 0
       ncolumnsx = 0

       if (cattype .eq. 'truth') then
         if (ncolumnx  .eq. 0) ncolumnx  = index ( headline(nhcols)(1:lhl), ' XCO_PIXEL'  )
         if (ncolumny  .eq. 0) ncolumny  = index ( headline(nhcols)(1:lhl), ' YCO_PIXEL'  )
         if (ncolumnx  .gt. 0 .and. ncolumnx .lt. 5) ncolumnx = 0        
         if (ncolumny  .gt. 0 .and. ncolumny .lt. 5) ncolumny = 0        
         if (ncolumna  .eq. 0) ncolumna  = index ( headline(nhcols)(1:lhl), ' AFWHM_ARCS' )
         if (ncolumnb  .eq. 0) ncolumnb  = index ( headline(nhcols)(1:lhl), ' BFWHM_ARCS' )
         if (ncolumnt  .eq. 0) ncolumnt  = index ( headline(nhcols)(1:lhl), ' FWHM_ARCS'  )
         if (ncolumnfp .eq. 0) ncolumnfp = index ( headline(nhcols)(1:lhl), ' PEAKFLUX'   )
         if (ncolumnft .eq. 0) ncolumnft = index ( headline(nhcols)(1:lhl), ' TOTALFLUX'  )
         if (ncolumnr  .eq. 0) ncolumnr  = index ( headline(nhcols)(1:lhl), ' SN_RATIO'   )
         if (ncolumnx  .eq. 0 .or. ncolumny .eq. 0 .or. ncolumna .eq. 0 .or. ncolumnb .eq. 0 .or. ncolumnt .eq. 0 .or.
     &       ncolumnfp .eq. 0 .or. ncolumnft .eq. 0 .or. ncolumnr .eq. 0) then
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Some columns for the method TRUTH not found in:'
           write (iotty,'(i5,a)') nhcols, '"'//headline(nhcols)(1:lhl)//'"'
           stop 99
         endif
       else
         if (ncolumnx   .eq. 0) ncolumnx   = index ( headline(nhcols)(1:lhl), ' XCO_P' )
         if (ncolumny   .eq. 0) ncolumny   = index ( headline(nhcols)(1:lhl), ' YCO_P' )
         if (ncolumnx   .gt. 0 .and. ncolumnx .lt. 5) ncolumnx = 0        
         if (ncolumny   .gt. 0 .and. ncolumny .lt. 5) ncolumny = 0        
         if (ncolumng   .eq. 0) ncolumng   = index ( headline(nhcols)(1:lhl), ' GOOD' )
         if (lfootprints) then
           if (ncolumna .eq. 0) ncolumna   = index ( headline(nhcols)(1:lhl), ' FOOA'//cnwave )
           if (ncolumnb .eq. 0) ncolumnb   = index ( headline(nhcols)(1:lhl), ' FOOB'//cnwave )
         else
           if (ncolumna .eq. 0) ncolumna   = index ( headline(nhcols)(1:lhl), ' AFWH'//cnwave )
           if (ncolumnb .eq. 0) ncolumnb   = index ( headline(nhcols)(1:lhl), ' BFWH'//cnwave )
         endif
         if (ncolumnt   .eq. 0) ncolumnt   = index ( headline(nhcols)(1:lhl), ' THEP'//cnwave )
         if (ncolumnfp  .eq. 0) ncolumnfp  = index ( headline(nhcols)(1:lhl), ' FXP_BEST'//cnwave )
         if (ncolumnfpe .eq. 0) ncolumnfpe = index ( headline(nhcols)(1:lhl), ' FXP_ERRO'//cnwave )
         if (ncolumnft  .eq. 0) ncolumnft  = index ( headline(nhcols)(1:lhl), ' FXT_BEST'//cnwave )
         if (ncolumnfte .eq. 0) ncolumnfte = index ( headline(nhcols)(1:lhl), ' FXT_ERRO'//cnwave )
         if (ncolumnr   .eq. 0) ncolumnr   = index ( headline(nhcols)(1:lhl), ' SIG_MONO'//cnwave )
         if (ncolumnr   .eq. 0) ncolumnr   = index ( headline(nhcols)(1:lhl), ' SN_RATIO'//cnwave )
         if (ncolumnx  .eq. 0 .or. ncolumny  .eq. 0 .or. ncolumna .eq. 0 .or. ncolumnb .eq. 0 .or. ncolumnt .eq. 0 .or.
     &       ncolumnfp .eq. 0 .or. ncolumnft .eq. 0 .or. ncolumnr .eq. 0 .or. (cattype .eq. 'final' .and. ncolumng .eq. 0)) then
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Some columns for a method not found in:'
           write (iotty,'(i5,a)') nhcols, '"'//headline(nhcols)(1:lhl)//'"'
           stop 99
         endif
       endif
       
       do k=1,nextr

         read (15,'(a)') cline
         
         if (cline(1:1) .eq. '!' .or. cline(1:1) .eq. '#') goto 77
         
         read (cline,*) no
         read (cline(ncolumnx:) ,*) x0o
         read (cline(ncolumny:) ,*) y0o
         if (ncolumng .gt. 0) then
           read (cline(ncolumng:),*) goodness
         else
           goodness = 1.0d99
         endif
         read (cline(ncolumna:) ,*) majdiam
         read (cline(ncolumna:) ,*) majdiam
         read (cline(ncolumnb:) ,*) mindiam
         read (cline(ncolumnt:) ,*) angle
         read (cline(ncolumnfp:),*) fxpbest
         read (cline(ncolumnft:),*) fxtbest
         read (cline(ncolumnr:) ,*) sigmono
         
         if (cattype .eq. 'gauss') angle = - angle - 90.0d0
                
         majdiam = majdiam * sizefact
         mindiam = mindiam * sizefact

         if (cattype .ne. 'truth') then
           read (cline(ncolumnfpe:),*) fxperro
           read (cline(ncolumnfte:),*) fxterro
         else
           fxperro = 0.0d0
           fxterro = 0.0d0
         endif

         if (thickfact .gt. 0.0d0) then
           nxe = int ( max ( majdiam, mindiam ) / dx * 2.0d0 ) + 10 !<-- define number of pixels in the ellipse window in the X-dir
           nye = int ( max ( majdiam, mindiam ) / dy * 2.0d0 ) + 10 !<-- define number of pixels in the ellipse window in the Y-dir
         else
           nxe = int ( obsbeam / dx * 2.0d0 ) + 2
           nye = int ( obsbeam / dy * 2.0d0 ) + 2
         endif
         nxco = nint ( x0o )  !<-- ellipse center's X-coordinate expressed in pixels
         nyco = nint ( y0o )  !<-- ellipse center's Y-coordinate expressed in pixels

         radx = majdiam / 2.0d0  !<-- major radius of the current ellipse
         rady = mindiam / 2.0d0  !<-- minor radius of the current ellipse
         
! Loop over all pixels of the ellipse window.
                                       
         if (abs ( sigmono  ) .ge. sigmin .and. 
     &       abs ( fxpbest  ) .gt. snr1min * fxperro .and.
     &       abs ( fxtbest  ) .gt. snr2min * fxterro .and.
     &       abs ( goodness ) .ge. goodmin) then
           
           deltax = dx * abs ( thickfact )
           deltay = dy * abs ( thickfact )
           deltaxdot = deltax / 2.0d0
           deltaydot = deltay / 2.0d0
           major1 = majdiam - deltax
           minor1 = mindiam - deltay
           major2 = majdiam
           minor2 = mindiam
           semimajor2 = major2 / 2.0d0
           semiminor2 = minor2 / 2.0d0
           
           nr = nint ( semimajor2 / dx )
           l1 = max ( nxco - nr , 1 )
           l2 = min ( nxco + nr , nx )
           m1 = max ( nyco - nr , 1 )
           m2 = min ( nyco + nr , ny )
           do m=m1,m2
             dblm = dble ( m )
             do l=l1,l2
               dbll = dble ( l )
               if (thickfact .gt. 0.0d0) then
                 if (mindiam / 2.0d0 .ge. deltax / 3.0d0) then
                   if (inellipse ( dbll, dblm, dx, dy, x0o, y0o, major2, minor2, angle )) then
                     if (.not.inellipse ( dbll, dblm, dx, dy, x0o, y0o, major1, minor1, angle )) then
                       funb(l,m) = ellvalfact * funmin
                     endif
                   endif
                 else
                   if (inellipse ( dbll, dblm, dx, dy, x0o, y0o, major2, minor2, angle )) then
                     funb(l,m) = ellvalfact * funmin
                   endif
                 endif
               else
                 if (abs ( l - nxco ) * dx .le. deltaxdot .and. abs ( m - nyco ) * dy .le. deltaydot) then
                   funb(l,m) = ellvalfact * funmin
                 endif
               endif
             enddo
           enddo

           major1 = majdiam
           minor1 = mindiam
           major2 = majdiam + deltax
           minor2 = mindiam + deltay
           semimajor2 = major2 / 2.0d0
           semiminor2 = minor2 / 2.0d0
           
           nr = nint ( semimajor2 / dx )
           l1 = max ( nxco - nr , 1 )
           l2 = min ( nxco + nr , nx )
           m1 = max ( nyco - nr , 1 )
           m2 = min ( nyco + nr , ny )
           do m=m1,m2
             dblm = dble ( m )
             do l=l1,l2
               dbll = dble ( l )
               if (thickfact .gt. 0.0d0) then
                 if (inellipse ( dbll, dblm, dx, dy, x0o, y0o, major2, minor2, angle )) then
                   if (.not.inellipse ( dbll, dblm, dx, dy, x0o, y0o, major1, minor1, angle )) then
                     funb(l,m) = ellvalfact * max ( funmax, 1.0d-10 )
                   endif
                 endif
               else
                 if ((abs ( l - nxco ) * dx .le. 3.0d0 * deltaxdot .and. abs ( m - nyco ) * dy .le. 3.0d0 * deltaydot) .and. 
     &               (abs ( l - nxco ) * dx .gt. deltaxdot .or. abs ( m - nyco ) * dy .gt. deltaydot)) then
                   funb(l,m) = ellvalfact * max ( funmax, 1.0d-10 )
                 endif
               endif
             enddo
           enddo
         endif
  77     continue     
       enddo

       close ( 15 )
       
       if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Writing output image '''//outname(1:ion)//'''...'
      
       call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, funb, dx, dy, object, crval1
     &            , crval2, outname(1:ion), header, footer, cdate, ctime, creator, beam, blank, crota1, crota2, cd11, cd12, cd21
     &            , cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history )

       deallocate ( funa, funb )
      
       if (lverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, header, footer, iotty
     &            , iolog, creator, beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale
     &            , wave, datamn, datamx, history, lverbose )
!__________________________________________________________________________________________________________________________________
!
! GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay      
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, anynull, myfile
       character*(*) ctype1, ctype2, creator, fname, header, footer(2), bunit, object, history, lverbose
       character*25  cdatetime
       character*80  footer1, footer2, comment
       integer       i, j, nx, ny, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, lastc, iotty, iolog, maxdim
     &             , ilc1, ilc2, rw, nfound, blank, stat01, stat02, stat03, stat04, stat05, stat06, stat07, stat08, stat09
     &             , stat10, stat11, stat12, stat13, stat14, stat15, stat16, stat17, stat18, stat19, stat20, stat21, stat22
     &             , stat23, stat24, stat25, stat26, stat27, stat28, stat29, stat30, stat31, stat32, stat0, nan
       parameter   ( maxdim = 2 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, cd11, cd12, cd21, cd22
       real*8        fun(nx,ny)
       real*8        as2deg, datamx, datamn
       parameter   ( as2deg = 1.0d0 / 3600.0d0 )
       external      lastc, ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat01 = 0
       stat02 = 0
       stat03 = 0
       stat04 = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
       if (stat .eq. 0 .and. (comment(1:22) .eq. 'Alexander Men`shchikov' .or. comment(1:22) .eq. 'Alexander Men''shchikov')) then
         myfile = .true.
       else
         myfile = .false.
         stat = 0
         creator = 'ELLIPSES'
       endif

       header = 'FITS Plot Header'

       if (myfile) then
         call ftgkys ( unit, 'HEADER' , header , comment, stat01 )
         call ftgkys ( unit, 'FOOTER1', footer1, comment, stat02 )
         call ftgkys ( unit, 'FOOTER2', footer2, comment, stat03 )
         ilc1 = lastc ( footer1 )
         ilc2 = lastc ( footer2 )
         if (ilc1 .ne. 0 .and. ilc2 .ne. 0) then
           footer(1) = footer1(1:ilc1)//footer2(1:ilc2)
         else
           footer(1) = 'Footer'
         endif
       else
         call ftgkys ( unit, 'OBJECT' , header, comment, stat04 )
         if (stat04 .ne. 0 .or. header(1:10) .eq. '          ') header = fname
         footer(1) = header
       endif

! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       stat08 = 0
       stat0 = stat01 + stat02 + stat03 + stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15
     &       + stat16 + stat17 + stat18 + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28
     &       + stat29 + stat30 + stat31

       if (lverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat01 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword HEADER absent or error'
         if (stat02 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER1 absent or error'
         if (stat03 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER2 absent or error'
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat29 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EPOCH absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = 'No object specified'

! Read the FITS file into the 2D array.

       call ftg2dd ( unit, group, dble ( blank ), naxes(1), naxes(1), naxes(2), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       nx = naxes(1)
       ny = naxes(2)
       funmin = 1.0d30
       funmax =-1.0d30
       nan = 0
       do j=1,ny
         do i=1,nx
           if (isnan ( fun(i,j) )) then  
             nan = nan + 1
             if (lverbose .ne. '-verb0') then
               if (nan .lt. 100) then
                 write (iotty,'(a,i4,a,i4,a)') '   RFITS: Element (', i,',', j, ') contains not a number (NaN)...'
               elseif (nan .eq. 100) then
                 write (iotty,'(a)') '   RFITS: Many NaNs, no more info on the elements.'
               endif
             endif
           else
             funmin = min ( funmin, fun(i,j) )
             funmax = max ( funmax, fun(i,j) )
           endif  
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, header
     &            , footer, cdate, ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale
     &            , wave, datamn, datamx, history )
!__________________________________________________________________________________________________________________________________
!
! GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay      
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend
       character*(*) fname, ctype1, ctype2, object, creator, header, footer(2), ctime, cdate, bunit, cfitsversion, history
       character*19  cdatetime
       integer       i, j, nx, ny, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, naxes(2), iat, iflen
     &             , lastc, blank
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, dfun, rot1, rot2, epoch, equinox, datamn
     &             , cosrota, sinrota, rotangle, datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, pi
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      lastc
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 2-D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 2
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       datamn = 1.0d30
       datamx =-1.0d30
       do j=1,ny
         do i=1,nx
           dfun = fun(i,j)
           if (dfun .lt. datamn ) datamn = dfun
           if (dfun .gt. datamx ) datamx = dfun
         enddo
       enddo

       bitpix = -32     !<-- single precision floating-point values

       cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

       if (cfitsversion .eq. '5.030') then
         call fdelete ( fname, status )
       endif

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Create the new empty FITS file.

       if (cfitsversion .eq. '5.030') then
         call ftinit ( unit, fname, blocksze, status )
       else
         call ftinit ( unit, '!'//fname, blocksze, status )
       endif

! Write the required header keywords.

       call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

       call ftpkys ( unit, 'CREATOR', creator,'Alexander Men`shchikov, SAp IRFU CEA Saclay', status )
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'HEADER' , header, 'title', status )

       iat = 0
       iat = index ( footer(1), ' |F' )
       if (iat .eq. 0) iat = index ( footer(1), ' mm ' )
       iflen = lastc ( footer(1) )

       if (iat .gt. 0) then
         call ftpkys ( unit, 'FOOTER1', footer(1)(1:iat-1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', footer(1)(iat:iflen), 'part 2', status )
       else
         call ftpkys ( unit, 'FOOTER1', footer(1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', ' ', 'part 2', status )
       endif
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )

! Write all parameters.

       dxdeg = dx * as2deg
       dydeg = dy * as2deg

! Correct if the input images had cd11 or cd22 set to zero.

       if (abs ( cd11 ) .lt. almostzero .or. abs ( cd11 ) .gt. dxdeg) then
         cd11 = -dxdeg
         cd12 = 0.0d0
       endif
       if (abs ( cd22 ) .lt. almostzero .or. abs ( cd22 ) .gt. dxdeg) then
         cd21 = 0.0d0
         cd22 = dydeg
       endif
       
       cosrota = - cd11 / dxdeg
       sinrota = - cd12 / dxdeg
       rotangle = acos ( cosrota ) * 180.0d0 / pi
       
       if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = 180.0d0 - rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = 180.0d0 + rotangle
       elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = - rotangle
       endif
       rot2 = rot1

       call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
       call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
       call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
       if  (blank .gt. 0)
     & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
       call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
       call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
       call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
       call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CDELT1' , -dxdeg, 13 , 'coordinate increment along axis', status )
       call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
       call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
       call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
       call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
       call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
       call ftpkyd ( unit, 'EPOCH'  , epoch  , 13, ' ', status )
       call ftpkyd ( unit, 'EQUINOX', equinox, 13, ' ', status )
       call ftphis ( unit, history, status )

       if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'image was convolved with the beam (arcsec)', status )

! Write the array to the FITS file.

       call ftp2dd ( unit, group, nx, nx, ny, fun, status )

! Close the file.

       call ftclos ( unit, status )

! Free the unit number.

       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||