
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program fftconv
!__________________________________________________________________________________________________________________________________
!
! Convolution using a Fast Fourier Transform algorithm RLFT3 from Press, Teukolsky, Vetterling, & Flannery (1992):
! "Numerical recipes in FORTRAN. The art of scientific computing", Cambridge University Press, 2nd edition.
!
! This code is part of GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!                 
       implicit      none

       logical       lfname, lunix
                                  
       integer       inmx, insp   
                                  
       character*5   cfpptime
       character*6   cfitsversion, lverbose, cnxo, cnyo, cnx, cny, cnxout, cnyout
       character*7   clibname
       character*8   ctime, cbeam, cwhat
       character*10  cdate
       character*21  compda
       character*80  object, creator, ctype1, ctype2, comm(10), bunit, history
       character*128 header, footer(2)
       character*500 filename, filenamec, outname, arg1, arg2, arg3, arg4, arg5, arg6, arg7

       integer       i, j, fnl, fnlc, irc, ndate, nx, ny, lastc, lasts, ij, nxo, nyo, ia1, ia2, ia3, ia4, ia5, ia6, ia7, isp1, iotty
     &             , iolog, ixo, iyo, nxout, nyout, norm, k, mxo, myo, nx0, ny0, ion, lb, inxo, inyo, inx, iny, inxout, inyout
     &             , blank, fnlen

       real          fitsvers
       real*8        rcore, rstar, distpc, radiau, wave, emergi, freq, view, crpix1, crpix2, crval1, crval2, ra, dec, xout, yout
     &             , xfnyq, yfnyq, bbody, dx, dy, dxf, dyf, eps, pi, totfx1, totfx2, dilute, hwhm, sigm2, rad2, argex, beamx, fnorm
     &             , delx, dely, beam, funmin, funmax, crota1, crota2, epoch, equinox, bzero, bscale, datamin, datamax, almostzero
     &             , beamrfits, constant, expo, correct, cd11, cd12, cd21, cd22

       real*8    , allocatable :: fun(:,:), psf(:,:), funi(:), funipsf(:), funx(:)
       complex*16, allocatable :: speq(:), speqpsf(:), spec(:), specpsf(:)

       parameter   ( eps = 0.01d0, pi = 3.14159265358979d0, almostzero = 1.0d-20, iotty = 6, iolog = 0 )

       external      rfits, wfits, when, fdelete, printerr
!__________________________________________________________________________________________________________________________________
!
! Determine operating system.

       call osystem ( lunix )

! Get the current date and time.

       creator = 'FFT Convolution'

       call when ( lunix, ctime, cdate, ndate, 4 )

       call ftvers ( fitsvers )

       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       compda = __DATE__//' '//cfpptime

! Get command line parameters.

       call getarg ( 1, arg1 )
       call getarg ( 2, arg2 )
       call getarg ( 3, arg3 )
       call getarg ( 4, arg4 )
       call getarg ( 5, arg5 )
       call getarg ( 6, arg6 )
       call getarg ( 7, arg7 )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )
       ia6 = lastc ( arg6 )
       ia7 = lastc ( arg7 )

       lverbose = '-verb2'
       if (ia7 .gt. 0 .and. arg7(1:ia7-1) .eq. '-verb') lverbose = arg7(1:ia7)
       if (ia6 .gt. 0 .and. arg6(1:ia6-1) .eq. '-verb') lverbose = arg6(1:ia6)
       if (ia5 .gt. 0 .and. arg5(1:ia5-1) .eq. '-verb') lverbose = arg5(1:ia5)
       if (ia4 .gt. 0 .and. arg4(1:ia4-1) .eq. '-verb') lverbose = arg4(1:ia4)

       if (lverbose .eq. '-verb2') write (iotty,'(/a$)') ' FFTCONV = FFTransform & Convolve FITS Images = '//compda
       if (arg1(1:ia1) .eq. ':') stop
       if (lverbose .eq. '-verb2') write (iotty,'()')
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Alexander Men`shchikov, SAp IRFU CEA Saclay, France.'
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Using RLFT3 from Numerical Recipes for F77 by William H Press et al.'
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                            //' by William D Pence.'
       if (ia3 .eq. 0) then
         write (iotty,'(/a)') ' USAGE: fftconv <fwhm> <expo> <image> [scalepeak|noscaling|psf] [-o <outname>] [-verb{0|1|2}]'
         stop 99
       endif
       
       if (arg1(1:ia1) .eq. '-0' .and. arg2(1:ia2) .eq. '-0') then
         cwhat = 'fftransf'
       else
         cwhat = 'convolve'
       endif
       read (arg1(1:ia1),*,err=6) beam
       goto 7
   6   continue
         write (iotty,'(/a)') '   FFTCONV: ERROR: Trouble reading BEAM from this string: '''//arg1(1:ia1)//'''.'
         stop 99
   7   continue
       read (arg2(1:ia2),*,err=8) expo
       goto 9
   8   continue
         write (iotty,'(/a)') '   FFTCONV: ERROR: Trouble reading EXPO from this string: '''//arg2(1:ia2)//'''.'
         stop 99
   9   continue
       if (beam .gt. almostzero) then
         if (expo .lt. almostzero) then
           cbeam = 'gaussian'
         else
           cbeam = 'moffatfn'
         endif
       else
         if (beam .lt. -almostzero) then
           cbeam = 'cylinder'
         else
           cbeam = 'exponent'
         endif
       endif
       beam = abs ( beam )

       fnlen = ia3
       filename = arg3(1:ia3)
       fnl = lastc ( filename )

! Check if the input file exists and open it.

       inquire ( file=filename(1:fnlen), exist=lfname )
       if (.not.lfname) then

! Maybe the file name extension was omitted?

         filename = filename(1:fnlen)//'.fits'
         fnl = lastc ( filename )
         inquire ( file=filename(1:fnl), exist=lfname )
         if (.not.lfname) then
           filename = filename(1:fnlen)//'.tab'
           fnl = lastc ( filename )
           inquire ( file=filename(1:fnl), exist=lfname )
         endif
       endif

       isp1 = lasts ( filename ) + 1

       if (cwhat .eq. 'convolve') then
         if (filename(fnl-3:fnl-3) .eq. '.') then
           filenamec = filename(isp1:fnl-4)//'.c'//arg1(1:ia1)//'as.fits'
         else
           filenamec = filename(isp1:fnl-5)//'.c'//arg1(1:ia1)//'as.fits'
         endif
       else
         if (filename(fnl-3:fnl-3) .eq. '.') then
           filenamec = filename(isp1:fnl-4)//'.famp.fits'
         else
           filenamec = filename(isp1:fnl-5)//'.famp.fits'
         endif
       endif
       fnlc = lastc ( filenamec )

       outname = filenamec(1:fnlc)
       if (arg4(1:2) .eq. '-o') outname = arg5
       if (arg5(1:2) .eq. '-o') outname = arg6
       ion = lastc ( outname )
       if (outname(ion-4:ion) .ne. '.fits') then
         outname = outname(1:ion)//'.fits'
         ion = lastc ( outname )
       endif

       if (lverbose .eq. '-verb1') write (iotty,'(/a)') ' FFTCONV: '//arg1(1:ia1)//' '//arg2(1:ia2)//' '//filename(1:fnl-5)

       if (.not.lfname) then
         write (iotty,'(/a)') '   FFTCONV: ERROR: File '''//filename(1:fnlen)//''' not found.'
         stop 1
       endif

       if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   Reading image '''//filename(1:fnl)//'''...'

! Determine numbers of pixels in the FITS image.

       call getnumpix ( filename(1:fnl), nxo, nyo )

! Default parameters.

       xout = 0.0
       yout = 0.0
       norm = 1
       do i=1,10
         comm(i) = 'no comment'
       enddo

! Find power of 2 in X coordinate to use as array dimension for FFT.

       ixo = 0
       do i=1,14
         nx = mod ( 2**i, nxo )
         if (nx .lt. ixo) goto 10
         ixo = nx
       enddo
 10    continue
       nx = 2**i
     
! Find power of 2 in Y coordinate to use as array dimension for FFT.

       iyo = 0
       do j=1,14
         ny = mod ( 2**j, nyo )
         if (ny .lt. iyo) goto 20
         iyo = ny
       enddo
 20    continue
       ny = 2**j
     
       inmx = max ( nx, ny )
       insp = inmx + 1

       allocate ( fun(insp,insp), psf(insp,insp), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a,2i5)') '   FFTCONV: ERROR: Trouble allocating memory (10) for 2 arrays of sizes: ', insp, insp
         stop 10
       endif
              
       do j=1,insp
         do i=1,insp
           fun(i,j) = 0.0d0
           psf(i,j) = 0.0d0
         enddo
       enddo

       if (filename(fnl-3:fnl-3) .eq. '.') then

! Read input image from table file.

         call rimage ( iotty, iolog, insp, nxo, nyo, fun, dx, dy, rcore, distpc, radiau, rstar, bbody, emergi, wave, freq, view
     &               , comm, filename(1:fnl), header, footer, lunix, irc )
       
         if (irc .ne. 0) then
           if (k .eq. 1) then
             write (iotty,'(/a)') '   FFTCONV: ERROR: Trouble reading file '''//filename(1:fnl)//'''.'
             stop 99
           endif
           goto 777
         endif
         creator = 'FFTCONV'
         object  = 'No name'
       else
       
         call rfits ( insp, nxo, nyo, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, fun, dx, dy, object, ra, dec
     &              , filename(1:fnl), header, footer, iotty, 0, creator, beamrfits, funmin, funmax, blank, crota1, crota2
     &              , cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history, lverbose )

         if (lverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax
       endif

       if (dx .le. 1.0d-10 .or. dy .le. 1.0d-10) then
         write (iotty,'(/a,2(1pe15.7))') '   FFTCONV: ERROR: Wrong pixel size(s):', dx, dy
         stop 99
       endif
                                                             
! Compute total flux under the image for rescaling after the convolution. Note that the offsets 'nx0' and 'ny0' below are 
! to avoid border effects in convolution and thus to reduce the error in rescaling of the convolved map.
         
       nx0 = 2.0d0 * beam / dx
       ny0 = 2.0d0 * beam / dy

       funmax =-1.0d+30
       funmin = 1.0d+30
       do j=min(1+ny0,nyo/2-10),max(nyo-ny0,nyo/2+11)
         do i=min(1+nx0,nxo/2-10),max(nxo-nx0,nxo/2+11)
           funmax = max ( funmax, fun(i,j) )
           funmin = min ( funmin, fun(i,j) )
         enddo
       enddo

       if (abs ( funmax - funmin ) .lt. almostzero) cwhat = 'constant'

       if (cwhat .eq. 'constant') then

         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   The original image is constant; no transform will be performed.'

         deallocate ( psf )
       else

! Subtract the minimum value of the input image to make the image entirely positive.
! This is very important for accurate re-normalization of the total flux under the image.

         if (arg4(1:ia4) .ne. 'scalepeak') then
           constant = funmin
           do j=1,nyo
             do i=1,nxo
               fun(i,j) = fun(i,j) - constant
             enddo
           enddo
         else
           constant = 0.0d0
         endif

! Highest (Nyquist) frequency of the 2D sample.

         xfnyq = 1.0d0 / (2.0d0 * dx)
         yfnyq = 1.0d0 / (2.0d0 * dy)

! Find the lowest frequency.

         dxf = 2.0d0 * xfnyq / nx
         dyf = 2.0d0 * yfnyq / ny

! Define the upper limit of frequency point numbers for output.

         nxout = nint ( 2.0d0 * xout / dxf )
         nyout = nint ( 2.0d0 * yout / dyf )
         if (nxout .eq. 0 .or. nxout .gt. nx + 1) nxout = nx + 1
         if (nyout .eq. 0 .or. nyout .gt. ny + 1) nyout = ny + 1

! The number of visibility points should be odd for output.

         if (mod ( nxout, 2 ) .eq. 0) nxout = min ( nxout + 1, nx + 1 )
         if (mod ( nyout, 2 ) .eq. 0) nyout = min ( nyout + 1, ny + 1 )
   
         if (nxo .lt. 10) then
           write (cnxo,'(i1)') nxo
         else if (nxo .lt. 100) then
           write (cnxo,'(i2)') nxo
         else if (nxo .lt. 1000) then
           write (cnxo,'(i3)') nxo
         else if (nxo .lt. 10000) then
           write (cnxo,'(i4)') nxo
         else if (nxo .lt. 100000) then
           write (cnxo,'(i5)') nxo
         endif
         inxo = lastc ( cnxo )
         if (nyo .lt. 10) then
           write (cnyo,'(i1)') nyo
         else if (nyo .lt. 100) then
           write (cnyo,'(i2)') nyo
         else if (nyo .lt. 1000) then
           write (cnyo,'(i3)') nyo
         else if (nyo .lt. 10000) then
           write (cnyo,'(i4)') nyo
         else if (nyo .lt. 100000) then
           write (cnyo,'(i5)') nyo
         endif
         inyo = lastc ( cnyo )
         if (nx .lt. 10) then
           write (cnx,'(i1)') nx
         else if (nx .lt. 100) then
           write (cnx,'(i2)') nx
         else if (nx .lt. 1000) then
           write (cnx,'(i3)') nx
         else if (nx .lt. 10000) then
           write (cnx,'(i4)') nx
         else if (nx .lt. 100000) then
           write (cnx,'(i5)') nx
         endif
         inx = lastc ( cnx )
         if (ny .lt. 10) then
           write (cny,'(i1)') ny
         else if (ny .lt. 100) then
           write (cny,'(i2)') ny
         else if (ny .lt. 1000) then
           write (cny,'(i3)') ny
         else if (ny .lt. 10000) then
           write (cny,'(i4)') ny
         else if (ny .lt. 100000) then
           write (cny,'(i5)') ny
         endif
         iny = lastc ( cny )
         if (nxout .lt. 10) then
           write (cnxout,'(i1)') nxout
         else if (nxout .lt. 100) then
           write (cnxout,'(i2)') nxout
         else if (nxout .lt. 1000) then
           write (cnxout,'(i3)') nxout
         else if (nxout .lt. 10000) then
           write (cnxout,'(i4)') nxout
         else if (nxout .lt. 100000) then
           write (cnxout,'(i5)') nxout
         endif
         inxout = lastc ( cnxout )
         if (nyout .lt. 10) then
           write (cnyout,'(i1)') nyout
         else if (nyout .lt. 100) then
           write (cnyout,'(i2)') nyout
         else if (nyout .lt. 1000) then
           write (cnyout,'(i3)') nyout
         else if (nyout .lt. 10000) then
           write (cnyout,'(i4)') nyout
         else if (nyout .lt. 100000) then
           write (cnyout,'(i5)') nyout
         endif
         inyout = lastc ( cnyout )

         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Original: '//cnxo(1:inxo)//' '//cnyo(1:inyo)
     &                                                 //' Transform: '//cnx(1:inx)//' '//cny(1:iny)
     &                                                 //' Output: '//cnxout(1:inxout)//' '//cnyout(1:inyout)

! Compute total flux under the input image.

         if (arg4(1:ia4) .eq. 'noscaling') then
           totfx1 = 1.0d0
         else
           if (arg4(1:ia4) .ne. 'scalepeak') then
             totfx1 = 0.0d0
             do j=min(1+ny0,nyo/2-10),max(nyo-ny0,nyo/2+11)
               do i=min(1+nx0,nxo/2-10),max(nxo-nx0,nxo/2+11)
                 totfx1 = totfx1 + fun(i,j)
               enddo
             enddo
           else
             totfx1 = 0.0d0
             do j=min(1+ny0,nyo/2-10),max(nyo-ny0,nyo/2+11)
               do i=min(1+nx0,nxo/2-10),max(nxo-nx0,nxo/2+11)
                 totfx1 = max ( totfx1, fun(i,j) )
               enddo
             enddo
           endif
         endif

! Compute a convolution kernel.

         if (cwhat .eq. 'convolve') then
     
           if (lverbose .eq. '-verb2') then
             if (beam .ge. 10000.0d0) then
               if (cbeam .eq. 'moffatfn') write (iotty,'(a,f14.7,a,f5.2,a)') '   Preparing a', beam
     &                                                                     , ' arcsec Moffat beam (exponent=',expo,')...'
               if (cbeam .eq. 'gaussian') write (iotty,'(a,f14.7,a)') '   Preparing a', beam, ' arcsec Gaussian beam...'
               if (cbeam .eq. 'exponent') write (iotty,'(a,f14.7,a)') '   Preparing a', beam, ' arcsec exponential beam...'
               if (cbeam .eq. 'cylinder') write (iotty,'(a,f14.7,a)') '   Preparing a', beam, ' arcsec cylindrical beam...'
             elseif (beam .ge. 1000.0d0) then
               if (cbeam .eq. 'moffatfn') write (iotty,'(a,f13.7,a,f5.2,a)') '   Preparing a', beam
     &                                                                     , ' arcsec Moffat beam (exponent=',expo,')...'
               if (cbeam .eq. 'gaussian') write (iotty,'(a,f13.7,a)') '   Preparing a', beam, ' arcsec Gaussian beam...'
               if (cbeam .eq. 'exponent') write (iotty,'(a,f13.7,a)') '   Preparing a', beam, ' arcsec exponential beam...'
               if (cbeam .eq. 'cylinder') write (iotty,'(a,f13.7,a)') '   Preparing a', beam, ' arcsec cylindrical beam...'
             elseif (beam .ge. 100.0d0) then
               if (cbeam .eq. 'moffatfn') write (iotty,'(a,f12.7,a,f5.2,a)') '   Preparing a', beam
     &                                                                     , ' arcsec Moffat beam (exponent=',expo,')...'
               if (cbeam .eq. 'gaussian') write (iotty,'(a,f12.7,a)') '   Preparing a', beam, ' arcsec Gaussian beam...'
               if (cbeam .eq. 'exponent') write (iotty,'(a,f12.7,a)') '   Preparing a', beam, ' arcsec exponential beam...'
               if (cbeam .eq. 'cylinder') write (iotty,'(a,f12.7,a)') '   Preparing a', beam, ' arcsec cylindrical beam...'
             elseif (beam .ge. 10.0d0) then
               if (cbeam .eq. 'moffatfn') write (iotty,'(a,f11.7,a,f5.2,a)') '   Preparing a', beam
     &                                                                     , ' arcsec Moffat beam (exponent=',expo,')...'
               if (cbeam .eq. 'gaussian') write (iotty,'(a,f11.7,a)') '   Preparing a', beam, ' arcsec Gaussian beam...'
               if (cbeam .eq. 'exponent') write (iotty,'(a,f11.7,a)') '   Preparing a', beam, ' arcsec exponential beam...'
               if (cbeam .eq. 'cylinder') write (iotty,'(a,f11.7,a)') '   Preparing a', beam, ' arcsec cylindrical beam...'
             else
               if (cbeam .eq. 'moffatfn') write (iotty,'(a,f10.7,a,f5.2,a)') '   Preparing a', beam
     &                                                                     , ' arcsec Moffat beam (exponent=',expo,')...'
               if (cbeam .eq. 'gaussian') write (iotty,'(a,f10.7,a)') '   Preparing a', beam, ' arcsec Gaussian beam...'
               if (cbeam .eq. 'exponent') write (iotty,'(a,f10.7,a)') '   Preparing a', beam, ' arcsec exponential beam...'
               if (cbeam .eq. 'cylinder') write (iotty,'(a,f10.7,a)') '   Preparing a', beam, ' arcsec cylindrical beam...'
             endif
           endif

! sigma, parameter of a gaussian, is ~20% smaller (by factor sqrt((log 4)) = 1.177) than its HWHM - HALF width at half maximum.
! if one takes sigma defined in this way, one gets correct value 1/2 (half maximum) of the gaussian at x = HWHM

           hwhm  = beam / 2.0d0
           if (cbeam .eq. 'gaussian') then
             sigm2 = hwhm**2 / log ( 4.0d0 )
           endif
           if (cbeam .eq. 'moffatfn') then
             correct = (2.0d0**(1.0d0 / dble ( expo )) - 1.0d0)
             sigm2 = hwhm**2
           endif
         
           do j=1,nyo
             do i=1,nxo
               delx = dble ( i - (nxo / 2 + 1) ) * dx
               dely = dble ( j - (nyo / 2 + 1) ) * dy
               rad2  = delx**2 + dely**2
               if (cbeam .eq. 'gaussian') then
                 argex = min ( rad2 / (2.0d0 * sigm2), 200.0d0 )
                 beamx = exp ( -argex )
               endif
               if (cbeam .eq. 'moffatfn') then
                 argex = rad2 / sigm2
                 beamx = 1.0d0 / (1.0d0 + correct * argex)**expo
               endif
               if (cbeam .eq. 'exponent') then
                 argex = min ( rad2, 200.0d0 )
                 beamx = exp ( -argex )
               endif
               if (cbeam .eq. 'cylinder') then
                 if (sqrt ( rad2 ) .le. hwhm) then
                   beamx = 1.0d0
                 else
                   beamx = 0.0d0
                 endif
               endif
               psf(i,j) = beamx
             enddo
           enddo
         
           if (arg4(1:ia4) .eq. 'psf') then
             if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Writing the beam ''fftbeam.fits''...'
         
             call wfits ( cfitsversion, insp, nxo, nyo, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, psf, dx, dy, object
     &                  , crval1, crval2, 'fftbeam.fits', header, footer, cdate, ctime, creator, beam, blank, crota1, crota2, cd11
     &                  , cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history )
           endif
         endif

! Initialize FFT arrays with zeros.
               
         allocate ( speq(insp), speqpsf(insp), funi(insp*insp), funipsf(insp*insp), stat=irc )
       
         if (irc .ne. 0) then
           write (iotty,'(/a,2i5)') '   FFTCONV: ERROR: Trouble allocating memory (20) for 4 arrays of sizes: ', insp, insp
           stop 20
         endif
       
         do j=1,insp*insp
           funi(j) = 0.0d0
           funipsf(j) = 0.0d0
         enddo
         fnorm = dble ( nx * ny ) / 2.0d0
         do j=1,ny
           do i=1,nx
             ij = i + (j - 1) * nx
             funi(ij) = fun(i,j)
             funipsf(ij) = psf(i,j)
           enddo
         enddo
       
         deallocate ( fun, psf )

! Fast Fourier Transform to Foruier space.

         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Computing 2D fast Fourier transform of the image...'
       
         call rlft3 ( funi, speq, nx, ny, 1, 1 )

! Reconstruct the power spectrum. Magnitude of the visibility is given by abs ( F ) or sqrt ( real(F)**2 + imag(F)**2 ),
! and phase is given by atan ( imag(F) / real(F) ).

         if (cwhat .eq. 'fftransf') then
     
           allocate ( fun(insp,insp), spec(insp/2*insp), specpsf(insp/2*insp), stat=irc )
     
           if (irc .ne. 0) then
             write (iotty,'(/a,2i5)') '   FFTCONV: ERROR: Trouble allocating memory (21) for 3 arrays of sizes: ', insp/2, insp
             stop 21
           endif
           if (cwhat .eq. 'fftransf') then
             do j=1,insp/2*insp
               spec(j) = ( 0.0d0, 0.0d0 )
               specpsf(j) = ( 0.0d0, 0.0d0 )
             enddo
           endif
     
           call equivalent ( insp/2*insp, funi, spec )
     
           do j=1,ny/2+1
             do i=1,nx/2
               ij = i + (j - 1) * nx / 2
               fun(i+nx/2,j+ny/2) = abs ( spec(ij) )
             enddo
             fun(nx+1,j+ny/2) = abs ( speq(j) )
           enddo
     
           do j=ny/2+1,ny
             do i=1,nx/2
               ij = i + (j - 1) * nx / 2
               fun(i+nx/2,j-ny/2) = abs ( spec(ij) )
             enddo
             fun(nx+1,j-ny/2) = abs ( speq(j) )
           enddo

! Symmetry is used to fill in the other half of the plane in X.

           do j=1,ny/2
             do i=1,nx/2
               fun(nx/2-i+1,ny/2-j+1) = fun(nx/2+i+1,ny/2+j+1)
             enddo
           enddo
       
           do j=ny/2+1,ny+1
             do i=1,nx/2
               fun(nx/2-i+1,j) = fun(nx/2+i+1,ny-j+2)
             enddo
           enddo

! Make the axes correctly centered. Reference pixel is at the center and its value is 0.0

           crpix1 = (nx - 1) / 2 + 1
           crpix2 = (ny - 1) / 2 + 1
           crval1 = 0.0
           crval2 = 0.0
           nxo = nx
           nyo = ny
       
           deallocate ( spec, specpsf )
         else
       
! Fast Fourier Transform to Foruier space.

           if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Computing 2D fast Fourier transform of the beam...'

           call rlft3 ( funipsf, speqpsf, nx, ny, 1, 1 )

! Now convolve (multiply in Fourier space).

           if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Convolving the image by multiplying the transforms...'
       
           allocate ( funx(insp*insp), stat=irc )
       
           if (irc .ne. 0) then
             write (iotty,'(/a,2i5)') '   FFTCONV: ERROR: Trouble allocating memory (25) for array sizes: ', insp, insp
             stop 25
           endif
       
           call equivmult ( insp, funi, funipsf, funx )
       
           do j=1,insp
             speq(j) = speq(j) * speqpsf(j)
           enddo
       
           deallocate ( funi, funipsf )

! Fast Fourier Transform back to image space.

           if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Computing 2D fast Fourier transform back to images...'
       
           call rlft3 ( funx, speq, nx, ny, 1, -1 )
       
           deallocate ( speq, speqpsf )
         
           if (lverbose .eq. '-verb2') write (iotty,'(a,$)') '   Unscrambling transform...'
         
           allocate ( fun(insp,insp), psf(insp,insp), stat=irc )
         
           if (irc .ne. 0) then
             write (iotty,'(/a,2i5)') '   FFTCONV: ERROR: Trouble allocating memory (30) for 2 arrays of sizes: ', insp, insp
             stop 30
           endif
       
           do j=1,ny
             do i=1,nx
               ij = i + (j - 1) * nx
               psf(i,j) = funx(ij) / fnorm
               fun(i,j) = 0.0d0
             enddo
           enddo
       
           deallocate ( funx )

! We need to further unscramble the output.
                    
           mxo = nxo / 2
           myo = nyo / 2
           do j=1,ny-myo
             do i=1,nx-mxo
               fun(i,j) = psf(i+mxo,j+myo)
             enddo
           enddo
           do j=1,nyo-(ny-myo)
             do i=1,nxo-(nx-mxo)
               fun(nx-mxo+i,ny-myo+j) = psf(i,j)
             enddo
           enddo
           do j=1,ny-myo
             do i=1,nxo-(nx-mxo)
               fun(nx-mxo+i,j) = psf(i,j+myo)
             enddo
           enddo
           do j=1,nyo-(ny-myo)
             do i=1,nx-mxo
               fun(i,ny-myo+j) = psf(i+mxo,j)
             enddo
           enddo
       
           deallocate ( psf )

! Compute total flux under the convolved sky image (taking into account both positive and negative values).

           if (arg4(1:ia4) .eq. 'noscaling') then
             totfx2 = 1.0d0
           else
             if (arg4(1:ia4) .ne. 'scalepeak') then
               totfx2 = 0.0d0
               do j=min(1+ny0,nyo/2-10),max(nyo-ny0,nyo/2+11)
                 do i=min(1+nx0,nxo/2-10),max(nxo-nx0,nxo/2+11)
                   totfx2 = totfx2 + fun(i,j)
                 enddo
               enddo
             else
               totfx2 = 0.0d0
               do j=min(1+ny0,nyo/2-10),max(nyo-ny0,nyo/2+11)
                 do i=min(1+nx0,nxo/2-10),max(nxo-nx0,nxo/2+11)
                   totfx2 = max ( totfx2, fun(i,j) )
                 enddo
               enddo
             endif
           endif
           dilute = totfx1 / totfx2
           
           if (lverbose .eq. '-verb2') write (iotty,'(a,1pe11.3)') ' scaling factor:', dilute

! Rescale the convolved image in order to conserve the total flux.

           do j=1,nyo
             do i=1,nxo
               fun(i,j) = fun(i,j) * dilute + constant
             enddo
           enddo
         endif
       endif
              
       if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Writing output image '''//outname(1:ion)//'''...'

       call wfits ( cfitsversion, insp, nxo, nyo, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, fun, dx, dy, object, crval1
     &            , crval2, outname(1:ion), header, footer, cdate, ctime, creator, beam, blank, crota1, crota2, cd11, cd12, cd21
     &            , cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history )

       deallocate ( fun )

       if (cwhat .ne. 'fftransf' .and. cwhat .ne. 'constant') then
         if (dilute .lt. almostzero) then
           write (iotty,'(/a,1pe11.3)') '   FFTCONV: ERROR: Scaling factor is suspicious:', dilute
           stop 99
         endif
       endif
777    continue

       if (lverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'
       
!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine equivmult
       
     &            ( insp, spec, specpsf, funx )
!__________________________________________________________________________________________________________________________________
!
! Replacement for the EQUIVALENCE statement, allowing ALLOCATED arrays.
!__________________________________________________________________________________________________________________________________
!
       integer       j, insp
       complex*16    spec(insp/2*insp), specpsf(insp/2*insp), funx(insp*insp)
!__________________________________________________________________________________________________________________________________
!
       do j=1,insp/2*insp
         funx(j) = spec(j) * specpsf(j)
       enddo
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine equivalent
       
     &            ( ndim, infun, outfun )
!__________________________________________________________________________________________________________________________________
!
! Replacement for the EQUIVALENCE statement, allowing ALLOCATED arrays.
!__________________________________________________________________________________________________________________________________
!
       integer       j, ndim
       complex*16    infun(ndim), outfun(ndim)
!__________________________________________________________________________________________________________________________________
!
       do j=1,ndim
         outfun(j) = infun(j)
       enddo
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rlft3 
       
     &            ( data, speq, nn1, nn2, nn3, isign )
!__________________________________________________________________________________________________________________________________
!
! Given a two- or three- dimensional real array 'data' whose dimensions are nn1, nn2, nn3 (where nn3 is 1 for the case of a
! two-dimensional array), this routine returns (for isign=1) the complex fast fourier transform as two complex arrays: On output,
! 'data' contains the zero and positive frequency values for the first frequency component, while 'speq' contains the Nyquist
! critical frequency values of the first frequency component. Second (and third) frequency components are store for zero, positive,
! and negative frequencies, in standard wrap-around order. For isign=-1, the inverse transform (times nn1*nn2*nn3/2 as a constant
! multiplicative factor) is performed, with output 'data' (viewed as a real array) deriving from input 'data' (viewed as complex
! and 'speq'. The dimensions nn1, nn2, nn3 must always be integer powers of 2.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       isign, nn1, nn2, nn3

! Note that 'data' is dimensioned as complex, its output format.

       complex*16    data(nn1/2,nn2,nn3), speq(nn2,nn3)
       integer       i1, i2, i3, j1, j2, j3, nn(3)
       real*8        theta, wi, wpi, wpr, wr, wtemp
       complex*16    c1, c2, h1, h2, w, wc
       external      fourn
!__________________________________________________________________________________________________________________________________
!
       c1 = cmplx ( 0.5, 0.0 )
       c2 = cmplx ( 0.0, -0.5 * isign )
       theta = 6.28318530717959d0 / dble ( isign * nn1 )
       wpr = -2.0d0 * sin ( 0.5d0 * theta )**2
       wpi = sin ( theta )
       nn(1) = nn1 / 2
       nn(2) = nn2
       nn(3) = nn3

! Case of forward transform.

       if (isign .eq. 1) then

! Here is where most of the computer time is spent.

         call fourn ( data, nn, 3, isign )

! Extend data periodically into 'speq'.

         do 12 i3 = 1,nn3
           do 11 i2 = 1,nn2
             speq(i2,i3) = data(1,i2,i3)
11         continue
12       continue
       endif

       do 15 i3=1,nn3

! Zero frequency is its own reflection; otherwise locate corresponding negative frequency in wrap-around order.

         j3 = 1
         if (i3 .ne. 1) j3 = nn3 - i3 + 2

! Initialize trigonometric recurrence.

         wr = 1.0d0
         wi = 0.0d0

         do 14 i1=1,nn1/4+1

           j1 = nn1 / 2 - i1 + 2

           do 13 i2=1,nn2
             j2 = 1
             if (i2 .ne. 1) j2 = nn2 - i2 + 2

             if (i1 .eq. 1) then            
               wc = conjg ( speq(j2,j3) )
               h1 = c1 * (data(1,i2,i3) + wc)
               h2 = c2 * (data(1,i2,i3) - wc)
               data(1,i2,i3) = h1 + h2
               speq(j2,j3) = conjg ( h1 - h2 )
             else                            
               wc =conjg ( data(j1,j2,j3) )
               h1 = c1 * (data(i1,i2,i3) + wc)
               h2 = c2 * (data(i1,i2,i3) - wc)
               data(i1,i2,i3) = h1 + w * h2
               data(j1,j2,j3) = conjg ( h1 - w * h2 )
             endif
13         continue

! Do the recurrence.

           wtemp = wr
           wr = wr * wpr - wi * wpi + wr
           wi = wi * wpr + wtemp * wpi + wi
           w = cmplx ( sngl ( wr ), sngl ( wi ) )
14       continue
15     continue
                         
! Case of reverse transform.

       if (isign .eq. -1) then

         call fourn ( data, nn, 3, isign )

       endif

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine fourn 
       
     &            ( data, nn, ndim, isign )
!__________________________________________________________________________________________________________________________________
!
! -132 -O3 -fp-model precise -fp-speculation off -warn all -watch source -traceback -assume noold_unit_star -heap-arrays -parallel -module modules
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       isign, ndim, nn(ndim)
       real*8        data(*)
       integer       i1, i2, i2rev, i3, i3rev, ibit, idim, ifp1, ifp2, ip1, ip2, ip3, k1, k2, n, nprev, nrem, ntot
       real*8        tempi, tempr, theta, wi, wpi, wpr, wr, wtemp
!__________________________________________________________________________________________________________________________________
!
       ntot = 1
       do 11 idim=1,ndim
         ntot = ntot * nn(idim)
11     continue
       nprev = 1

       do 18 idim=1,ndim
         n     = nn(idim)
         nrem  = ntot / (n * nprev)
         ip1   = 2 * nprev
         ip2   = ip1 * n
         ip3   = ip2 * nrem
         i2rev = 1

         do 14 i2=1,ip2,ip1
                
           if (i2 .lt. i2rev) then
             
             do 13 i1=i2,i2+ip1-2,2
               do 12 i3=i1,ip3,ip2

                 i3rev = i2rev + i3 - i2
                 tempr = data(i3  )
                 tempi = data(i3+1)

                 data(i3       ) = data(i3rev    )
                 data(i3 + 1   ) = data(i3rev + 1)
                 data(i3rev    ) = tempr
                 data(i3rev + 1) = tempi
12             continue
13           continue
           endif
           
           ibit = ip2 / 2

1          if ((ibit .ge. ip1) .and. (i2rev .gt. ibit)) then
             i2rev = i2rev - ibit
             ibit  = ibit / 2
             goto 1
           endif
           i2rev = i2rev + ibit
14       continue
         
         ifp1 = ip1

2        if (ifp1 .lt. ip2) then

           ifp2  = 2 * ifp1
           theta = isign * 6.28318530717959d0 / (ifp2 / ip1)
           wpr   = -2.d0 * sin ( 0.5d0 * theta )**2
           wpi   = sin ( theta )
           wr    = 1.d0
           wi    = 0.d0

           do 17 i3=1,ifp1,ip1
             do 16 i1=i3,i3+ip1-2,2
               do 15 i2=i1,ip3,ifp2

                 k1 = i2
                 k2 = k1 + ifp1
                 tempr = sngl ( wr ) * data(k2  ) - sngl ( wi ) * data(k2+1)
                 tempi = sngl ( wr ) * data(k2+1) + sngl ( wi ) * data(k2  )

                 data(k2  ) = data(k1  ) - tempr
                 data(k2+1) = data(k1+1) - tempi
                 data(k1  ) = data(k1  ) + tempr
                 data(k1+1) = data(k1+1) + tempi
15             continue
16           continue
             wtemp = wr
             wr    = wr * wpr - wi * wpi + wr
             wi    = wi * wpr + wtemp * wpi + wi
17         continue
           ifp1 = ifp2
           goto 2
         endif
         nprev = n * nprev
18     continue

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine icompare 
       
     &            ( arr1, arr2, len, eps, icmp )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       iprnt, icmp, lastc, iotty
       parameter   ( iprnt = 10, iotty = 6 )
       integer       j, len
       real*8        arr1(len), arr2(len), eps, abserr, relerr
       external      lastc
!__________________________________________________________________________________________________________________________________
!
       icmp = 0

       do 11 j=1,len

         abserr = abs ( arr1(j) - arr2(j) )
         relerr = abs ( abserr / (arr2(j) + 1.0d-30) )

         if (abs ( arr2(j) ) .gt. 1.0d-14 .and. relerr .gt. eps) then

           icmp = icmp + 1

           if (icmp .eq. 1) call mess ( ' ', ' ' )
           if (icmp .le. iprnt) then
             write (iotty,'(i9,1p2e15.7,0pf12.8,1p2e15.7)')  j, arr1(j), arr2(j), eps, abserr, relerr
           endif
         endif
11     continue

       if (icmp .ge. 1) write (iotty,'()')

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rimage 
       
     &            ( iotty, iolog, inmx, iend, jend, fin, dx1, dx2, rcore, distpc, radiau, rstar, bbody, emergi, wave, freq, view
     &            , comm, fname, header, footer, lunix, irc )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       lfname, lunix
       integer       inmx, iotty, iolog, irc
       character*(*) fname, header, footer(2), comm(10)
       character*80  filen
       character*15  d1
       character*4   d2
       integer       i, j, iend, jend, fnl
       real*8        dx1, dx2, rcore, distpc, radiau, rstar, bbody, emergi, wave, freq, view, q1
       real*8        fin(inmx,inmx)
       external      skipcomm
!__________________________________________________________________________________________________________________________________
!
       irc = 0
       filen = fname
       fnl = len ( fname )

! Check if the input file exists and open it.

       inquire ( file=fname, exist=lfname )

       if (.not.lfname) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' Image file '''//filen(1:fnl)//''' not found.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' Image file '''//filen(1:fnl)//''' not found.'
         irc = 1
         return
       endif

! Open the input data file.

       if (.not.lunix) open ( 1, file=fname, recl=100000, status='unknown' )
       if (     lunix) open ( 1, file=fname, status='unknown' )

       call skipcomm ( 1, fname(1:fnl) )

! Read the input intensity image file (with header/footer).

       read (1,'(a)') header
       read (1,'(a)') footer (1)
       read (1,*) iend, jend, dx1, dx2, rcore

       comm(1) = '! embedded core radius [arcsec]'

       if (iend .gt. inmx .or. jend .gt. inmx) then
         if (iotty .gt. 0) write (iotty,'(/a,2i5)') ' RIMAGE: ERROR: Too large dimension(s):', iend, jend
         if (iolog .gt. 0) write (iolog,'(/a,2i5)') ' RIMAGE: ERROR: Too large dimension(s):', iend, jend
         stop 99
       endif

       read (1,*)
       read (1,'(a14,10001(1pe10.3))' ) d1

       do i=1,iend
         read (1,'(a4,10001(1pe10.3))') d2, q1,(fin(i,j), j=1,jend)
       enddo

       d1 = d2

       read (1,'()',err=10)
       read (1,'(1pd10.3,a)',err=10) distpc, comm(2)
       read (1,'(1pd10.3,a)',err=10) radiau, comm(3)
       read (1,'(1pd10.3,a)',err=10) rstar , comm(4)
       read (1,'(1pd10.3,a)',err=10) bbody , comm(5)
       read (1,'(1pd10.3,a)',err=10) emergi, comm(6)
       read (1,'(1pd10.3,a)',err=10) wave  , comm(7)
       read (1,'(1pd10.3,a)',err=10) freq  , comm(8)
       read (1,'(1pd10.3,a)',err=10) view  , comm(9)
  10   continue

       close ( 1 )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( inmx, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, header, footer
     &            , iotty, iolog, creator, beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero
     &            , bscale, wave, datamn, datamx, history, lverbose )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, anynull, myfile
       character*(*) ctype1, ctype2, creator, fname, header, footer(2), bunit, object, history, lverbose
       character*25  cdatetime
       character*80  footer1, footer2, comment
       integer       inmx, i, j, nx, ny, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, lastc, iotty, iolog, maxdim
     &             , ilc1, ilc2, rw, nfound, blank, stat01, stat02, stat03, stat04, stat05, stat06, stat07, stat08, stat09
     &             , stat10, stat11, stat12, stat13, stat14, stat15, stat16, stat17, stat18, stat19, stat20, stat21, stat22
     &             , stat23, stat24, stat25, stat26, stat27, stat28, stat29, stat30, stat31, stat32, stat0, nan
       parameter   ( maxdim = 2 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, cd11, cd12, cd21, cd22
       real*8        fun(inmx,inmx)
       real*8        as2deg, datamx, datamn
       parameter   ( as2deg = 1.0d0 / 3600.0d0 )
       external      lastc, ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat01 = 0
       stat02 = 0
       stat03 = 0
       stat04 = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
       if (stat .eq. 0 .and. (comment(1:22) .eq. 'Alexander Men`shchikov' .or. comment(1:22) .eq. 'Alexander Men''shchikov')) then
         myfile = .true.
       else
         myfile = .false.
         stat = 0
         creator = 'MODFITS'
       endif

       header = 'FITS Plot Header'

       if (myfile) then
         call ftgkys ( unit, 'HEADER' , header , comment, stat01 )
         call ftgkys ( unit, 'FOOTER1', footer1, comment, stat02 )
         call ftgkys ( unit, 'FOOTER2', footer2, comment, stat03 )
         ilc1 = lastc ( footer1 )
         ilc2 = lastc ( footer2 )
         if (ilc1 .ne. 0 .and. ilc2 .ne. 0) then
           footer(1) = footer1(1:ilc1)//footer2(1:ilc2)
         else
           footer(1) = 'Footer'
         endif
       else
         call ftgkys ( unit, 'OBJECT' , header, comment, stat04 )
         if (stat04 .ne. 0 .or. header(1:10) .eq. '          ') header = fname
         footer(1) = header
       endif

! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       stat08 = 0
       stat0 = stat01 + stat02 + stat03 + stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15
     &       + stat16 + stat17 + stat18 + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28
     &       + stat29 + stat30 + stat31

       if (lverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat01 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword HEADER absent or error'
         if (stat02 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER1 absent or error'
         if (stat03 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER2 absent or error'
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat29 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EPOCH absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = 'No object specified'

! Read the FITS file into the 2D array.

       call ftg2dd ( unit, group, dble ( blank ), inmx, naxes(1), naxes(2), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       nx = naxes(1)
       ny = naxes(2)
       funmin = 1.0d30
       funmax =-1.0d30
       nan = 0
       do j=1,ny
         do i=1,nx
           if (isnan ( fun(i,j) )) then  
             nan = nan + 1
             if (lverbose .ne. '-verb0') then
               if (nan .lt. 100) then
                 write (iotty,'(a,i4,a,i4,a)') '   RFITS: Element (', i,',', j, ') contains not a number (NaN)...'
               elseif (nan .eq. 100) then
                 write (iotty,'(a)') '   RFITS: Many NaNs, no more info on the elements.'
               endif
             endif
           else
             funmin = min ( funmin, fun(i,j) )
             funmax = max ( funmax, fun(i,j) )
           endif  
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, inmx, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname
     &            , header, footer, cdate, ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero
     &            , bscale, wave, datamn, datamx, history )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend
       character*(*) fname, ctype1, ctype2, object, creator, header, footer(2), ctime, cdate, bunit, cfitsversion, history
       character*19  cdatetime
       integer       inmx, i, j, nx, ny, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, naxes(2), iat, iflen
     &             , lastc, blank
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, dfun, rot1, rot2, epoch, equinox, datamn
     &             , cosrota, sinrota, rotangle, datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, pi
       real*8        fun(inmx,inmx)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      lastc
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 2-D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 2
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       datamn = 1.0d30
       datamx =-1.0d30
       do j=1,ny
         do i=1,nx
           dfun = fun(i,j)
           if (dfun .lt. datamn ) datamn = dfun
           if (dfun .gt. datamx ) datamx = dfun
         enddo
       enddo

       bitpix = -32     !<-- single precision floating-point values

       cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

       if (cfitsversion .eq. '5.030') then
         call fdelete ( fname, status )
       endif

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Create the new empty FITS file.

       if (cfitsversion .eq. '5.030') then
         call ftinit ( unit, fname, blocksze, status )
       else
         call ftinit ( unit, '!'//fname, blocksze, status )
       endif

! Write the required header keywords.

       call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

       call ftpkys ( unit, 'CREATOR', creator,'Alexander Men`shchikov, SAp IRFU CEA Saclay', status )
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'HEADER' , header, 'title', status )

       iat = 0
       iat = index ( footer(1), ' |F' )
       if (iat .eq. 0) iat = index ( footer(1), ' mm ' )
       iflen = lastc ( footer(1) )

       if (iat .gt. 0) then
         call ftpkys ( unit, 'FOOTER1', footer(1)(1:iat-1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', footer(1)(iat:iflen), 'part 2', status )
       else
         call ftpkys ( unit, 'FOOTER1', footer(1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', ' ', 'part 2', status )
       endif
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )

! Write all parameters.

       dxdeg = dx * as2deg
       dydeg = dy * as2deg
       
! Correct if the input images had cd11 or cd22 set to zero.

       if (abs ( cd11 ) .lt. almostzero .or. abs ( cd11 ) .gt. dxdeg) then
         cd11 = -dxdeg
         cd12 = 0.0d0
       endif
       if (abs ( cd22 ) .lt. almostzero .or. abs ( cd22 ) .gt. dxdeg) then
         cd21 = 0.0d0
         cd22 = dydeg
       endif
       
       cosrota = - cd11 / dxdeg
       sinrota = - cd12 / dxdeg
       rotangle = acos ( cosrota ) * 180.0d0 / pi
       
       if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = 180.0d0 - rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = 180.0d0 + rotangle
       elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = - rotangle
       endif
       rot2 = rot1

       call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
       call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
       call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
       if  (blank .gt. 0)
     & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
       call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
       call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
       call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
       call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CDELT1' , -dxdeg, 13 , 'coordinate increment along axis', status )
       call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
       call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
       call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
       call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
       call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
       call ftpkyd ( unit, 'EPOCH'  , epoch  , 13, ' ', status )
       call ftpkyd ( unit, 'EQUINOX', equinox, 13, ' ', status )
       call ftphis ( unit, history, status )

       if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'image was convolved with the beam (arcsec)', status )

! Write the array to the FITS file.

       call ftp2dd ( unit, group, inmx, nx, ny, fun, status )

! Close the file.

       call ftclos ( unit, status )

! Free the unit number.

       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||