
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program imgstat
!__________________________________________________________________________________________________________________________________
!
! Compute various statistical quantities for a FITS image.
!
! This code is part of GETSOURCES • Multi-Scale Multi-Wavelength Source Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!
       implicit      none

       logical       lfnam1, lfnam2, lunix, loutsave, linsave, lsaveinmask, lsaveoutmask, lstd, lmedimage, lmeaimage, lhilog, llolog
     &             , lmedian, lpdf, lpdflin, lpdflog, lpdfpos, ltiming, lmedapprox, inellipse
                                                                                                                        
       integer       infu, iotty, iolog, nfu

       parameter   ( infu = 2000, iotty = 6, iolog = 0, nfu = 10 )
                                                                                                                        
       character*3   cwavenum, dot
       character*4   cpmx, camx, cx, cy
       character*5   cfpptime
       character*6   cfitsversion, lverbose                                                                           
       character*7   clibname                                                                                           
       character*8   ctime
       character*9   ccpu, cwal, cccpu, ccwal
       character*10  cdate
       character*21  compda
       character*80  object, creator, ctype1, ctype2, coption, bunit, history
       character*128 header, footer(2)
       character*500 filename1, filename2, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, imagename, maskname, outname, outfits

       integer       i, j, k, l, ndate, firstb, lastc, ia1, ia2, ia3, ia4, ia5, ia6, ia7, ia8, blank, nx1, ny1, nx2, ny2, iof, nbins
     &             , nxy, fnlen1, fnlen2, ion, nto, m, n, npx, irc, ico, nedge, im1, jm1, ip1, jp1, lb, nm, kk, namx, npp1, lenimg
     &             , lenmsk, npoints, nbinslo, nbinshi, ncx, ncy, imax2, jmax2, ii, jj, nn, npmx, n1, n2, m1, m2, wchours, wcmins
     &             , wcsecs, cpuhours, cpumins, cpusecs, ndc, ndw

       real          fitsvers
       real*8        wave, ra, dec, beam, sigma1, funmin1, funmax1, funmin2, funmax2, fmean2, sigma2, dx1, dy1, dx2, dy2, funlo
     &             , fmean, sigmas, variance, rnpx, totsum2, totsumedge, maskedgemean, crpix1, crpix2, crval1, crval2, sigma, funhi
     &             , crota1, crota2, epoch, equinox, bzero, bscale, datamin, datamax, mom2, mom3, mom4, skewness2, kurtosis2, xau
     &             , deltaf, fmedian2, delta, cd11, cd12, cd21, cd22, pdfbin, pdfmin, sq_arcsecs_per_sterad, convftot, pdflo
     &             , funmin2nozero, almostzero, funmin, funmax, pc, amu, muH2, Msun, bin, distpc, convmass, fwhm, funhalf, dfh
     &             , isumpct, numtotal, numpart, rw2, rx2, ry2, cpu, wal, cpu_sum, wal_sum, timer, cputot, wctot, selectk, xpoints
     &             , stdmean, stdstdm, medapprox, afwhm, bfwhm, atheta, equivrad, elongation, mxco, myco, fhmax, dbli, dblj

       integer, allocatable :: cimask(:,:)
       real*8, allocatable :: fun1(:,:), fun2(:,:), funstd(:,:), funm(:,:), funpdf(:,:), funone(:), funint(:), plane(:), funx(:)
     &                      , funbin(:,:)

       parameter   ( sq_arcsecs_per_sterad = 3282.80635d0 * 3600.0d0**2, pc = 3.085678d18, amu = 1.6605402D-24, muH2 = 2.8d0
     &             , xau = 1.4959787e+13, distpc = 140.0d0, Msun = 1.9891d33, almostzero = 1.0d-20, ltiming = .true., dot = '•' )

       external      firstb, lastc, rfits, wfits, when
!__________________________________________________________________________________________________________________________________
!
!!!       call when ( lunix, ctime, cdate, ndate, 4 )

! Determine operating system.

       call osystem ( lunix )

       call ftvers ( fitsvers )

       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       compda = __DATE__//' '//cfpptime

! Get the current date and time.

       call when ( lunix, ctime, cdate, ndate, 4 )


! Get command line parameters (file names).

       call getarg ( 1, arg1 )
       call getarg ( 2, arg2 )
       call getarg ( 3, arg3 )
       call getarg ( 4, arg4 )
       call getarg ( 5, arg5 )
       call getarg ( 6, arg6 )
       call getarg ( 7, arg7 )
       call getarg ( 8, arg8 )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )
       ia6 = lastc ( arg6 )
       ia7 = lastc ( arg7 )
       ia8 = lastc ( arg8 )

       lverbose = '-verb2'
       cwavenum = ''
       if (ia8 .gt. 0 .and. arg8(1:5) .eq. '-verb') then
         lverbose = arg8(1:6)
         if (ia8 .eq. 9) cwavenum = arg8(7:9)
       endif
       if (ia7 .gt. 0 .and. arg7(1:5) .eq. '-verb') then
         lverbose = arg7(1:6)
         if (ia7 .eq. 9) cwavenum = arg7(7:9)
       endif
       if (ia6 .gt. 0 .and. arg6(1:5) .eq. '-verb') then
         lverbose = arg6(1:6)
         if (ia6 .eq. 9) cwavenum = arg6(7:9)
       endif
       if (ia5 .gt. 0 .and. arg5(1:5) .eq. '-verb') then
         lverbose = arg5(1:6)
         if (ia5 .eq. 9) cwavenum = arg5(7:9)
       endif
       if (ia4 .gt. 0 .and. arg4(1:5) .eq. '-verb') then
         lverbose = arg4(1:6)
         if (ia4 .eq. 9) cwavenum = arg4(7:9)
       endif
       if (ia3 .gt. 0 .and. arg3(1:5) .eq. '-verb') then
         lverbose = arg3(1:6)
         if (ia3 .eq. 9) cwavenum = arg3(7:9)
       endif

       if (lverbose .eq. '-verb2') then
         write (iotty,'( )')
         if (arg1(1:ia1) .eq. ':') write (iotty,'(a$)') '  '
         write (iotty,'(a$)') ' IMGSTAT '//dot//' Compute Image Statistics '//dot//' '//compda
       endif
       if (arg1(1:ia1) .eq. ':') stop 2 !<-- '2' to indicate that IMGSTAT belongs to GETSOURCES v2
       if (lverbose .eq. '-verb2') write (iotty,'()')
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Alexander Men’shchikov, DAp IRFU CEA Saclay, France.'
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                                //' by William D Pence.'
       if (ia1 .eq. 0) then
         write (iotty,'( )')
         write (iotty,'(a)') ' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ USAGE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
         write (iotty,'( )')
         write (iotty,'(a)') ' imgstat [<action>] <imagein> [<maskimage>] [-o <imageout>] [-verb{0|1|2}]' 
         write (iotty,'( )')   
         write (iotty,'(a)') ' This utility computes various statistical quantities in a FITS image;'
         write (iotty,'(a)') ' <imagein> and <imageout> are the names of the input and output images. If the'
         write (iotty,'(a)') ' optional mask is supplied in <maskimage>, then the quantities can be computed'
         write (iotty,'(a)') ' inside or outside the masked area. If the mask image name is not specified,'
         write (iotty,'(a)') ' then all pixels of the input image are used.'
         write (iotty,'( )')   
         write (iotty,'(a)') ' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PARAMETERS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
         write (iotty,'( )')   
         write (iotty,'(a)') ' <action>'
         write (iotty,'( )')   
         write (iotty,'(a)') ' save    [-nomed] ........... compute statistics (inside a mask) and save'
         write (iotty,'(a)') ' savein  [-nomed] ........... compute statistics inside a mask and save'
         write (iotty,'(a)') ' saveout [-nomed] ........... compute statistics outside a mask and save'
         write (iotty,'(a)') ' std[ev]  <npts> <sigmas> ... stdev*sigmas image using a window with R=<npts>'
         write (iotty,'(a)') ' med[ian|bin] <npts> ........ median-filter image using a window with R=<npts>'
         write (iotty,'(a)') ' mea[n]   <npts> ............ mean-filter image using a window with R=<npts>'
         write (iotty,'(a)') ' pdf [+|-]<pdfbin> .......... compute pixel distribution function (histogram)'
         write (iotty,'( )')   
         write (iotty,'(a)') ' bin size <pdfbin> for creating PDFs is logarithmic if > 0 or linear if < 0.'
         write (iotty,'(a)')
         write (iotty,'(a)') ' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
         stop 99
       endif

       fnlen1 = 0
       fnlen2 = 0
       lstd = .false.
       lmedimage = .false.
       lmeaimage = .false.
       lpdf = .false.
       linsave = .false.
       loutsave = .false.
       lsaveinmask = .false.
       lsaveoutmask = .false.
       lmedian = .true.
       lmedapprox = .false.
       if (arg1(1:4) .eq. 'save') then
         linsave = .true.
         loutsave = .false.
       endif
       if (arg1(1:6) .eq. 'savein') then
         lsaveinmask = .true.
         linsave = .true.
         loutsave = .false.
       endif
       if (arg1(1:7) .eq. 'saveout') then 
         lsaveoutmask = .true.
         linsave = .false.
         loutsave = .true.
       endif
       if (arg2(1:ia2) .eq. '-nomed') then
         lmedian = .false.     
       endif
       if (arg1(1:3) .eq. 'std' .or. arg1(1:3) .eq. 'rms') then
         lstd = .true.
       endif
       if (arg1(1:3) .eq. 'med') then
         lmedimage = .true.
         if (arg1(1:ia1) .eq. 'medbin') then
           lmedapprox = .true.
         endif
       endif
       if (arg1(1:3) .eq. 'mea') then
         lmeaimage = .true.
       endif
       if (arg1(1:3) .eq. 'pdf') then
         lpdf = .true.
       endif
       coption = arg1(1:ia1)
       ico = lastc ( coption )
       if (lstd) then       
         ia3 = lastc ( arg3 )
         if (ia3 .eq. 0) then
           write (iotty,'(/a)') '   IMGSTAT: ERROR: Too few command line parameters.'
           stop 99
         endif
         coption = coption(1:ico)//' '//arg2(1:ia2)//' '//arg3(1:ia3)
         ico = lastc ( coption )
         read (arg2,*) xpoints
         npoints = nint ( xpoints )
         read (arg3,*,err=1122) sigmas
         goto 1133
 1122    write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble reading SIGMAS, verify it was specified after NPTS.'
         stop 99
 1133    arg1 = arg4
         arg2 = arg5
         arg3 = arg6
         arg4 = arg7
         arg5 = arg8
       elseif (lmedimage .or. lmeaimage) then       
         ia3 = lastc ( arg3 )
         if (ia3 .eq. 0) then
           write (iotty,'(/a)') '   IMGSTAT: ERROR: Too few command line parameters.'
           stop 99
         endif
         coption = coption(1:ico)//' '//arg2(1:ia2)
         ico = lastc ( coption )
         read (arg2,*) xpoints
         npoints = nint ( xpoints )
         arg1 = arg3
         arg2 = arg4
         arg3 = arg5
         arg4 = arg6
         arg5 = arg7
       elseif (lpdf) then       
         ia3 = lastc ( arg3 )
         if (ia3 .eq. 0) then
           write (iotty,'(/a)') '   IMGSTAT: ERROR: Too few command line parameters.'
           stop 99
         endif
         coption = coption(1:ico)//' '//arg2(1:ia2)
         ico = lastc ( coption )
         read (arg2,*) pdfbin
         lpdflin = .false.
         lpdflog = .false.
         lpdfpos = .false.
         if (arg2(1:1) .eq. '-') then
           lpdflin = .true.
         else
           lpdflog = .true.
           if (arg2(1:1) .eq. '+') lpdfpos = .true.
         endif
         pdfbin = abs ( pdfbin )
         arg1 = arg3
         arg2 = arg4
         arg3 = arg5
         arg4 = arg6
         arg5 = arg7
       else
         if (linsave .or. loutsave .or. lpdf) then
           if (lmedian) then
             arg1 = arg2
             arg2 = arg3
             arg3 = arg4
             arg4 = arg5
             arg5 = arg6
           else
             arg1 = arg3
             arg2 = arg4
             arg3 = arg5
             arg4 = arg6
             arg5 = arg7
           endif
         endif
       endif
       ia1 = lastc ( arg1 )

       filename1 = arg1(1:ia1)
       fnlen1 = lastc ( filename1 )

! Check if the input file exists and open it.

       inquire ( file=filename1(1:fnlen1), exist=lfnam1 )                      
       if (.not.lfnam1) then

! Maybe the file name extension was omitted?

         filename1 = filename1(1:fnlen1)//'.fits'
         fnlen1 = lastc ( filename1 )
         inquire ( file=filename1(1:fnlen1), exist=lfnam1 )
       endif
       imagename = filename1(1:fnlen1)
       lenimg = fnlen1

       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )
       lfnam2 = .false.
       
       if (ia2 .gt. 0 .and. arg2(1:5) .ne. '-verb') then
         filename2 = arg2(1:ia2)
         fnlen2 = lastc ( filename2 )

! Check if the input file exists and open it.

         inquire ( file=filename2(1:fnlen2), exist=lfnam2 )
         if (.not.lfnam2) then

! Maybe the file name extension was omitted?
  
           filename2 = filename2(1:fnlen2)//'.fits'
           fnlen2 = lastc ( filename2 )
           inquire ( file=filename2(1:fnlen2), exist=lfnam2 )
         endif
       endif
       if (lfnam2) then
         maskname = filename2(1:fnlen2)
         lenmsk = fnlen2
       else
         maskname = ' '
         lenmsk = 6
       endif
       
       outname = ''
       if (arg2(1:2) .eq. '-o') then
         outname = arg3
         ia2 = 0
         ia3 = 0
       endif
       if (arg3(1:2) .eq. '-o') then 
         outname = arg4
         ia3 = 0
       endif
       if (arg4(1:2) .eq. '-o') then 
         outname = arg5
       endif
       ion = lastc ( outname )

       if (lverbose .eq. '-verb1') 
     &     write (iotty,'(/a)') ' IMGSTAT: '//coption(1:ico)//' '//filename1(1:fnlen1-5)//' '//filename2(1:fnlen2-5)

       if (.not.lfnam1) then
         write (iotty,'(/a)') '   IMGSTAT: ERROR: FITS file '''//filename1(1:fnlen1-5)//''' not found.'
         stop 1
       endif

! Determine numbers of pixels in the FITS image.

       call getpixels ( imagename(1:lenimg), nx1, ny1, dx1, dy1 )
       
       write (cx,'(i4)') nx1
       write (cy,'(i4)') ny1
       ncx = int ( log10 ( dble ( nx1 ) ) ) + 1
       ncy = int ( log10 ( dble ( ny1 ) ) ) + 1

       if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   Reading '''//imagename(1:lenimg)//'''... '
     &                                                //cx(4-ncx+1:4)//' x '//cy(4-ncy+1:4)

       allocate ( fun1(nx1,ny1), stat=irc )
   
       if (irc .ne. 0) then
         write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (10).'
         stop 10
       endif

       call rfits ( nx1, ny1, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, fun1, dx1, dy1, object, ra, dec
     &            , imagename(1:lenimg), header, footer, iotty, 0, creator, beam, funmin1, funmax1, blank, crota1, crota2, cd11
     &            , cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history, lverbose )

       if (lverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))'  ) '   Minmax values:', funmin1, funmax1
       if (lverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7),a)') '     Pixel sizes:', dx1, dy1, ' arcsec'

! Second file (mask image) if exists.       

       if (lfnam2) then

! Determine numbers of pixels in the FITS image.

         call getpixels ( maskname(1:lenmsk), nx2, ny2, dx2, dy2 )

         write (cx,'(i4)') nx2
         write (cy,'(i4)') ny2
         ncx = int ( log10 ( dble ( nx2 ) ) ) + 1
         ncy = int ( log10 ( dble ( ny2 ) ) ) + 1
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Reading '''//maskname(1:lenmsk)//'''... '
     &                                                //cx(4-ncx+1:4)//' x '//cy(4-ncy+1:4)

         allocate ( fun2(nx2,ny2), stat=irc )
       
         if (irc .ne. 0) then
           write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (20).'
           stop 20
         endif

         call rfits ( nx2, ny2, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, fun2, dx2, dy2, object, ra, dec
     &              , maskname(1:lenmsk), header, footer, iotty, 0, creator, beam, funmin2, funmax2, blank, crota1, crota2, cd11
     &              , cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history, lverbose )
       
         if (lverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))'  ) '   Minmax values:', funmin2, funmax2
         if (lverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7),a)') '     Pixel sizes:', dx2, dy2, ' arcsec'

         if (nx1 .ne. nx2 .or. ny1 .ne. ny2 .or. abs ( dx1 - dx2 ) .gt. 1.0d-3 * dx1 .or. abs ( dy1 - dy2 ) .gt. 1.0d-3 * dy1) then
           write (iotty,'(/a)') '   IMGSTAT: ERROR: Images have different parameters (look in headers):'
           write (iotty,*) nx1, nx2, ny1, ny2, abs ( dx1 - dx2 ), abs ( dy1 - dy2 )
           stop 99
         endif
       else

         allocate ( fun2(nx1,ny1), stat=irc )
       
         if (irc .ne. 0) then
           write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (30).'
           stop 30
         endif

         do j=1,ny1
           do i=1,nx1
             fun2(i,j) = 1.0d0
           enddo
         enddo
       endif

       convftot = 1.0e6 * dx1 * dy1 / sq_arcsecs_per_sterad  !<-- intensity in MJy/sr to flux in Jy
       convmass = dx1 * dy1 * (distpc * xau)**2 * amu * muH2 / Msun
       
! Compute statistics only if production of stdev or median images was not requested.

       if (.not.lstd .and. .not.lmedimage .and. .not.lmeaimage) then
         nxy = 0
         totsum2 = 0.0d0
         funmin2 = 1.0d+30
         funmax2 =-1.0d+30
         funmin2nozero = 1.0d+30
         fmean2 = 0.0d0
         imax2 = 0
         jmax2 = 0
         do j=1,ny1
           do i=1,nx1
             if (fun2(i,j) .gt. almostzero .or. lsaveoutmask) then
               nxy = nxy + 1
               totsum2 = totsum2 + fun1(i,j)
               funmin2 = min ( fun1(i,j), funmin2 )
               if (fun1(i,j) .ge. funmax2) then
                 imax2 = i
                 jmax2 = j
                 funmax2 = fun1(i,j)
               endif
               if (fun1(i,j) .gt. almostzero) funmin2nozero = min ( fun1(i,j), funmin2nozero )
             endif
           enddo
         enddo
         if (nxy .eq. 0) then
           funmin2 = 0.0d0
           funmax2 = 0.0d0
           funmin2nozero = 0.0d0
         else
           fmean2 = totsum2 / dble ( nxy )
         endif

         nedge = 0
         totsumedge = 0.0d0
         maskedgemean = 0.0d0
!!         if (lsaveinmask) then
         if (lfnam2) then
           do j=1,ny1
             do i=1,nx1
               if (fun2(i,j) .gt. almostzero) then
                 im1 = max ( i - 1, 1 )
                 jm1 = max ( j - 1, 1 )
                 ip1 = min ( i + 1, nx1 )
                 jp1 = min ( j + 1, ny1 )
                 if (fun2(im1,j  ) .lt. almostzero .or. fun2(ip1,j  ) .lt. almostzero .or.
     &               fun2(i  ,jm1) .lt. almostzero .or. fun2(i  ,jp1) .lt. almostzero .or.
     &               fun2(im1,jm1) .lt. almostzero .or. fun2(im1,jp1) .lt. almostzero .or.
     &               fun2(ip1,jm1) .lt. almostzero .or. fun2(ip1,jp1) .lt. almostzero) then
                   nedge = nedge + 1
                   totsumedge = totsumedge + fun1(i,j)
                 endif
               endif
             enddo
           enddo
           if (nedge .gt. 0) maskedgemean = totsumedge / dble ( nedge )
  
!!           totsum2 = totsum2 - maskedgemean * dble ( nxy )  !<-- Crude background subtraction in the presence of mask image
         endif

! Find the peak sizes in two ways.

         nxy = 0
         sigma2 = 0.0d0
         fwhm = 0.0d0
         nedge = 0
         nn = 0
         funhalf = funmax2 / 2.0d0
         delta = 0.0d0
         if (imax2 .gt. 0 .and. jmax2 .gt. 0) then
           do j=1,ny1
             jj = j - jmax2
             do i=1,nx1
               ii = i - imax2
               if (fun2(i,j) .gt. almostzero .or. lsaveoutmask) then
                 nxy = nxy + 1
                 sigma2 = sigma2 + (fun1(i,j) - fmean2)**2
                 if (fun1(i,j) .ge. funhalf) then
                   im1 = max ( i - 1, 1 )
                   jm1 = max ( j - 1, 1 )
                   ip1 = min ( i + 1, nx1 )
                   jp1 = min ( j + 1, ny1 )
                   if (fun1(im1,j  ) .lt. funhalf .or. fun1(ip1,j  ) .lt. funhalf .or.
     &                 fun1(i  ,jm1) .lt. funhalf .or. fun1(i  ,jp1) .lt. funhalf .or.
     &                 fun1(im1,jm1) .lt. funhalf .or. fun1(im1,jp1) .lt. funhalf .or.
     &                 fun1(ip1,jm1) .lt. funhalf .or. fun1(ip1,jp1) .lt. funhalf) then
                     nedge = nedge + 1
                     fwhm = fwhm + sqrt ( dble ( ii )**2 + dble ( jj )**2 )
                     dfh = fun1(i,j) - funhalf
                     if (fun1(im1,j) .lt. funhalf) then
                       nn = nn + 1
                       delta = delta + dfh / (fun1(i,j) - fun1(im1,j))
                     endif
                     if (fun1(ip1,j) .lt. funhalf) then
                       nn = nn + 1
                       delta = delta + dfh / (fun1(i,j) - fun1(ip1,j))
                     endif
                     if (fun1(i,jm1) .lt. funhalf) then
                       nn = nn + 1
                       delta = delta + dfh / (fun1(i,j) - fun1(i,jm1))
                     endif
                     if (fun1(i,jp1) .lt. funhalf) then
                       nn = nn + 1
                       delta = delta + dfh / (fun1(i,j) - fun1(i,jp1))
                     endif
                   endif
                 endif
               endif
             enddo
           enddo
           if (nxy .gt. 1) sigma2 = sqrt ( sigma2 / dble ( nxy - 1 ) )
           if (nn .gt. 0) delta = delta / dble ( nn )
           if (nedge .gt. 0) then
             fwhm = fwhm / dble ( nedge )               !!* 1.020107739d0  !<-- empirical factor...
             fwhm = 2.0d0 * (fwhm + delta) * dx1
           endif
         endif

         call sizemeasure ( iotty, 1, 1, nx1, 1, ny1, nx1, ny1, fun2, fun1, dx1, dy1, mxco, myco, afwhm, bfwhm, atheta, equivrad
     &                    , elongation, almostzero )

         fhmax = -1.0d99
         do j=1,ny1
           dblj = dble ( j )
           do i=1,nx1
             dbli = dble ( i )
             if (fun2(i,j) .gt. almostzero .or. lsaveoutmask) then
               if (.not.inellipse ( dbli, dblj, dx1, dy1, mxco, myco, afwhm, bfwhm, atheta )) then
                 fhmax = max ( fhmax, fun1(i,j) )
               endif
             endif
           enddo
         enddo
       
! Calculation of the median value; do it only if requested explicitly, to save time.

         if (lmedian) then
           allocate ( funx(nx1*ny1), stat=irc )
           if (irc .ne. 0) then
             write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (130).'
             stop 130
           endif
           npx = 0
           do j=1,ny1
             do i=1,nx1
               if (fun2(i,j) .gt. almostzero .or. lsaveoutmask) then
                 npx = npx + 1
                 funx(npx) = fun1(i,j)
               endif
             enddo
           enddo
           if (npx .ne. 0) then
             fmedian2 = selectk ( (npx + 1) / 2, npx, funx )
!!             fmedian2 = medapprox ( npx, funx, funmin2, funmax2 )
           endif
           deallocate ( funx )

           if (lpdf) then
             lhilog = .false.
             llolog = .false.
             do k=1,nbins
               funint(k) = 0.0d0
               funone(k) = 0.0d0
             enddo
             pdflo = funmin2
             if (funmin2 .lt. -1.5d0 * pdfbin) then
               pdflo = -0.5d0 * pdfbin
             elseif (funmin2 .lt. 0.5d0 * pdfbin) then
               pdflo = funmin2nozero
             endif
             
             if (lpdflin) then
               if (pdflo .lt. 0.0d0) then
                 nbinslo = max ( nint ( (abs ( funmin2 ) - abs ( pdflo ) ) / pdfbin ) + 1, 1 )
                 nbinshi = max ( nint ( (abs ( funmax2 ) - abs ( pdflo ) ) / pdfbin ) + 1, 1 )
                 nbins = nbinslo + nbinshi
               else
                 nbins = max ( nint ( (funmax2 - funmin2) / pdfbin ) + 1, 2 )
               endif
             else
               if (pdflo .lt. 0.0d0) then
                 nbinslo = max ( nint ( (log10 ( abs ( funmin2 ) ) - log10 ( abs ( pdflo ) )) / pdfbin ) + 1, 1 )
                 nbinshi = max ( nint ( (log10 ( funmax2 ) - log10 ( abs ( pdflo ) )) / pdfbin ) + 1, 1 )
                 nbins = nbinslo + nbinshi
               else
                 nbins = max ( nint ( (log10 ( funmax2 ) - log10 ( pdflo ) ) / pdfbin ) + 1, 2 )
               endif
             endif

             allocate ( funone(nbins+1), funint(nbins+1), plane(nbins+1), stat=irc )
       
             if (irc .ne. 0) then
               write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (10).'
               stop 11
             endif

             if (lpdflin) then
               do k=nbinslo,1,-1
                 plane(k) = -(abs ( pdflo ) + dble ( nbinslo - k ) * pdfbin)
               enddo
               do k=nbinslo+1,nbins
                 plane(k) = abs ( pdflo ) + dble ( k - nbinslo - 1 ) * pdfbin
               enddo
             else
               if (pdflo .lt. 0.0d0) then
                 do k=nbinslo,1,-1
                   plane(k) = -10.0d0**( log10 ( abs ( pdflo ) ) + dble ( nbinslo - k ) * pdfbin )
                 enddo
                 do k=nbinslo+1,nbins
                   plane(k) = 10.0d0**( log10 ( abs ( pdflo ) ) + dble ( k - nbinslo - 1 ) * pdfbin )
                 enddo
               else
                 do k=1,nbins
                   plane(k) = 10.0d0**( log10 ( pdflo ) + dble ( k - 1 ) * pdfbin )
                 enddo
               endif
             endif

             allocate ( funpdf(nx1,ny1), stat=irc )
             if (irc .ne. 0) then
               write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (55).'
               stop 55
             endif

             if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Computing pixel distribution (bin = '//coption(5:ico)//')...'

             open (111,file=imagename(1:lenimg)//'.pdf.dat',status='unknown')
             write (111,'(a)') '# IMGSTAT = Compute Image Statistics = '//compda
             write (111,'(a)') '# Alexander Men’shchikov, DAp IRFU CEA Saclay, France.'
             write (111,'(a)') '# Using '//clibname(lb:7)//' library version'//cfitsversion//' by William D Pence.'
             write (111,'(a)') '#'
             write (111,'(a)') '# PIXEL DISTRIBUTION FROM THE IMAGE: '//imagename(1:lenimg)
             write (111,'(a)') '# FOR A BIN SIZE: '//coption(5:ico)
             write (111,'(a)') '#'
             write (111,'(a)') '# TABULATED QUANTITIES:'
             write (111,'(a)') '#'
             write (111,'(a)') '#   N ......... Bin number'
             write (111,'(a)') '#   BINLOW .... Lower edge of the bin'
             write (111,'(a)') '#   BINCOO .... Bin-centered coordinate'
             write (111,'(a)') '#   NPIXELS ... Number of pixels per bin'
             write (111,'(a)') '#   NPIXBIN ... Product NPIXELS x BINCOO'
             write (111,'(a)') '#   FMEAN ..... Mean value within the bin'
             write (111,'(a)') '#   SIGMA ..... Standard deviation within the bin'
             write (111,'(a)') '#'
             write (111,'(a)') '#      N     BINLOW      BINCOO     NPIXELS     NPIXBIN     FMEAN       SIGMA'
             write (111,'(a)') '#'

             npx = 0
             do k=1,nbins-1
               funint(k) = (plane(k) + plane(k+1)) / 2.0d0
               do j=1,ny1
                 do i=1,nx1
                   if (fun2(i,j) .gt. almostzero .or. lsaveoutmask) then
                     if (fun1(i,j) .gt. plane(k)) then
                       if (fun1(i,j) .le. plane(k+1)) then
                         npx = npx + 1
                         funone(k) = funone(k) + 1.0d0
                       endif
                     endif
                   endif
                 enddo
               enddo
             enddo

             if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Saving PDF data in '''//imagename(1:lenimg)//'.pdf.dat''...'

             funmax = 0.0d0
             funmin = 1.0d99
             do k=1,nbins-1
               if (lpdflin .or. funone(k) .gt. almostzero) then
                 if (funone(k) .lt. funmin) funmin = funone(k)
                 if (funone(k) .gt. funmax) funmax = funone(k)
               endif
             enddo
             pdfmin = funint(1)
             do k=1,nbins-1
               if (lpdflin .or. .not.lpdfpos .or. funint(k) .gt. almostzero) then
                 if (funone(k) .gt. almostzero) pdfmin = funint(k)
                 exit
               endif
             enddo
             
             do j=1,ny1
               do i=1,nx1
                 funpdf(i,j) = 0.0d0
               enddo
             enddo
             kk = 0
             do k=1,nbins-1
               funone(k) = max ( funone(k), funmin )
               nxy = 0
               fmean = 0.0d0
               do j=1,ny1
                 do i=1,nx1
                   if (fun2(i,j) .gt. almostzero .or. lsaveoutmask) then
                     if (fun1(i,j) .gt. plane(k)) then
                       if (fun1(i,j) .le. plane(k+1)) then
                         funpdf(i,j) = funone(k)
                         nxy = nxy + 1
                         fmean = fmean + fun1(i,j)
                       endif
                     endif
                   endif
                 enddo
               enddo
               if (nxy .gt. 0) fmean = fmean / dble ( nxy )
               nxy = 0
               sigma = 0.0d0
               do j=1,ny1
                 do i=1,nx1
                   if (fun2(i,j) .gt. almostzero .or. lsaveoutmask) then
                     if (fun1(i,j) .gt. plane(k)) then
                       if (fun1(i,j) .le. plane(k+1)) then
                         nxy = nxy + 1
                         sigma = sigma + (fun1(i,j) - fmean)**2
                       endif
                     endif
                   endif
                 enddo
               enddo
               if (nxy .gt. 1) sigma = sqrt ( sigma / dble ( nxy - 1 ) )
               if (funint(k) .ge. pdfmin .and. funone(k) .gt. 3.0d0) then
                 kk = kk + 1
                 write (111,'(i8,6(1pe12.4))') kk, plane(k), funint(k), funone(k), abs ( funint(k) ) * funone(k), fmean, sigma
               endif
             enddo

             close (111)
             
             outfits = imagename(1:lenimg)//'.pdf.fits'
             iof = lastc ( outfits )

             if (lverbose .eq. '-verb2') 
     &         write (iotty,'(a,f4.1,a)') '   Writing PDF image '''//outfits(1:iof)//'''...'

             call wfits ( cfitsversion, nx1, ny1, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, funpdf, dx1, dy1, object
     &                  , crval1, crval2, outfits(1:iof), header, footer, cdate, ctime, creator, beam, blank, crota1, crota2, cd11
     &                  , cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history )

             deallocate ( funpdf )
             deallocate ( funone, funint, plane )
           endif
         endif
       
! Higher-order statistical moments.       
       
         mom2 = 0.0d0
         mom3 = 0.0d0
         mom4 = 0.0d0
         nxy = 0
         
         do j=1,ny1
           do i=1,nx1
             if (fun2(i,j) .gt. almostzero .or. lsaveoutmask) then
               nxy = nxy + 1
               deltaf = fun1(i,j) - fmean2
               mom2 = mom2 + deltaf**2
               mom3 = mom3 + deltaf**3
               mom4 = mom4 + deltaf**4
             endif
           enddo
         enddo
         
         if (nxy .gt. 0) then
           rnpx = dble ( nxy )
           mom2 = mom2 / rnpx
           mom3 = mom3 / rnpx
           mom4 = mom4 / rnpx
         endif
         
         skewness2 = 0.0d0
         kurtosis2 = 0.0d0
         if (mom2 .gt. 0.0d0) then
           skewness2 = mom3 / sqrt ( mom2**3 )
           kurtosis2 = mom4 / mom2**2 - 3.0d0
         endif
         
         if (lverbose .eq. '-verb2') write (iotty,'(3(a,1pe15.7))') '   Maximum:', funmax2  , '      Mean:', fmean2 
         if (lmedian) then
           if (lverbose .eq. '-verb2') write (iotty,'(2(a,1pe15.7))') '   Minimum:', funmin2  , '    Median:', fmedian2
         else
           if (lverbose .eq. '-verb2') write (iotty,'(a,1pe15.7,a )') '   Minimum:', funmin2  , '    Median:  not requested'
         endif
         if (lverbose .eq. '-verb2') write (iotty,'(2(a,1pe15.7))') '    TotSum:', totsum2  , '     StDev:', sigma2                         
         if (lverbose .eq. '-verb2') write (iotty,'(2(a,1pe15.7))') '  Skewness:', skewness2, '  Kurtosis:', kurtosis2
         
         if (lfnam2) then
           if (lverbose .eq. '-verb2') write (iotty,'(2(a,1pe15.7))') '     EMean:', maskedgemean
           
           if ((lsaveinmask .or. lsaveoutmask) .and. nxy .lt. 9) then
             if (lverbose .eq. '-verb2' ) then
               write (iotty,'(/a,i3)') '   IMGSTAT: WARNING: Number of pixels in the mask is too small:', nxy
             endif
           endif
         endif

         if (lverbose .eq. '-verb2') then
           write (iotty,'(a,1pe15.7,a)') '   Minimum:', funmin2nozero, ' over positive non-zero pixels'
           write (iotty,'(2(a,1pe15.7,a))') '   TotFlux:', totsum2 * convftot, ' Jy', '   Mass:', totsum2 * convmass, ' Msun'
           write (iotty,'(a,1pe14.7,a)') '   (mass of a column density image at distance of 140 pc, convmass:',convmass,')'
           write (cx,'(i4)') imax2
           write (cy,'(i4)') jmax2
           ncx = int ( log10 ( dble ( imax2 ) ) ) + 1
           ncy = int ( log10 ( dble ( jmax2 ) ) ) + 1
           write (iotty,'(a,1pe14.7,a)') '   Pos of maximum: '//cx(4-ncx+1:4)//' '//cy(4-ncy+1:4)//' afwhm:', fwhm, ' arcsec'
           write (cx,'(i4)') nint ( mxco )
           write (cy,'(i4)') nint ( myco )
           ncx = int ( log10 ( dble ( mxco ) ) ) + 1
           ncy = int ( log10 ( dble ( myco ) ) ) + 1
           write (iotty,'(a,2(1pe14.7),a,e14.7,a)') '   Centroid: '//cx(4-ncx+1:4)//' '//cy(4-ncy+1:4)//'  full width:', afwhm
     &                                            , bfwhm, ' arcsec', atheta, ' deg'
           funhalf = fun1(nint(mxco),nint(myco)) / 2.0d0
           write (iotty,'(2(a,1pe14.7),a)') '   Half-maximum:', funhalf, '  from moments:', fhmax, ' (at full width)'
         endif
 10      continue
          
         if (linsave .or. loutsave) then
           if (.not.lfnam2) then
             if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Saving image statistics in ''.+imgstat1''...'
             open  (11,file='.+imgstat1'//cwavenum,status='unknown')     
             write (11,'(9(f56.19,1x))') funmin2, funmax2, fmean2, fmedian2, totsum2, sigma2, skewness2, kurtosis2, funmin2nozero
             close (11)
           else
             if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Saving image statistics in ''.+imgstat2''...'
             open  (22,file='.+imgstat2'//cwavenum,status='unknown')     
             write (22,'(10(f56.19,1x))') funmin2, funmax2, fmean2, fmedian2, totsum2, sigma2, skewness2, kurtosis2, maskedgemean
     &                                  , funmin2nozero
             close (22)
           endif

           if (lverbose .eq. '-verb2') write (iotty, '(f56.19)') funmin2
           if (lverbose .eq. '-verb2') write (iotty, '(f56.19)') funmax2
           if (lverbose .eq. '-verb2') write (iotty, '(f56.19)') fmean2
           if (lverbose .eq. '-verb2') write (iotty, '(f56.19)') fmedian2
           if (lverbose .eq. '-verb2') write (iotty, '(f56.19)') totsum2
           if (lverbose .eq. '-verb2') write (iotty, '(f56.19)') sigma2
           if (lverbose .eq. '-verb2') write (iotty, '(f56.19)') skewness2
           if (lverbose .eq. '-verb2') write (iotty, '(f56.19)') kurtosis2
           if (lfnam2) then
             if (lverbose .eq. '-verb2') write (iotty, '(f56.19)') maskedgemean
           endif
           if (lverbose .eq. '-verb2') write (iotty, '(f56.19)') funmin2nozero
         endif
       endif
                        
! Computing and writing stdev image.

       if (lstd) then

         if (.not.lfnam2) then 
           if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Mask image for STDEV image calculation not supplied.'
         endif

         allocate ( funstd(nx1,ny1), stat=irc )
         if (irc .ne. 0) then
           write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (50).'
           stop 50
         endif

         if (.not.allocated ( fun2 )) then
           allocate ( fun2(nx1,ny1), stat=irc )
           if (irc .ne. 0) then
             write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (70).'
             stop 70
           endif
           do j=1,ny1
             do i=1,nx1
               fun2(i,j) = 1.0d0    
             enddo
           enddo
         endif

         write (cpmx,'(i4.4)') 2 * npoints + 1
         npmx = int ( log10 ( dble ( 2 * npoints + 1 ) ) ) + 1
         write (camx,'(i4.4)') int ( dble ( 2 * npoints + 1 ) * dx1 )
         namx = int ( log10 ( dble ( 2 * npoints + 1 ) * dx1 ) ) + 1

         if (lverbose .eq. '-verb2') then
           write (iotty,'(a)') '   Computing stdev in a circular window of D = '//cpmx(4-npmx+1:4)//' pixels = '
     &                          //camx(4-namx+1:4)//' arcsec...'
         endif

         if (xpoints .le. 2.0d0) then
           rw2 = (xpoints + 0.5d0)**2
         else
           rw2 = (xpoints)**2
         endif
         npp1 = npoints + 1

         allocate ( cimask(-npoints:npoints,-npoints:npoints), stat=irc )
         if (irc .ne. 0) then
           write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (160).'
           stop 160
         endif

! The circular mask.
         
!!         if (lverbose .eq. '-verb2') write (iotty,'()')
         do l=-npoints,npoints
           ry2 = (dble ( l ))**2
           do k=-npoints,npoints
             rx2 = (dble ( k ))**2
             if (rx2 + ry2 .le. rw2) then
               cimask(k,l) = 1 
             else
               cimask(k,l) = 0
             endif
           enddo
!!           if (lverbose .eq. '-verb2') write (iotty,'(1000i2)') (cimask(k,l),k=-npoints,min(-npoints+77,npoints))
         enddo
         
         if (ltiming .and. lverbose .eq. '-verb2') then
           cpu = 0.0d0
           wal = 0.0d0
           cpu_sum = 0.0d0
           wal_sum = 0.0d0
!!           if (iotty .gt. 0) write (iotty,'(/a)') ' STARTED: '//cdate//' '//ctime
!!           if (iolog .gt. 0) write (iolog,'(/a)') ' STARTED: '//cdate//' '//ctime
         endif
         if (ltiming) then
           cpu = timer ( 'cpu', 0.0d0 )
           wal = timer ( 'wal', 0.0d0 )
         endif

         isumpct = 0.0d0
         numtotal = dble ( ny1 )
         numpart = 0.0d0
       
         if (lverbose .eq. '-verb2') then
           if (iotty .gt. 0) then
             write (iotty,'(11x,a)') '__________________________________________________'
             write (iotty,'(a)', advance='no') ' Progress: '
             if (iotty .gt. 0) endfile   ( iotty, err=666 )
 666         continue 
             if (iotty .gt. 0) backspace ( iotty )
           endif       
           if (iolog .gt. 0) then
             write (iolog,'(11x,a)') '__________________________________________________'
             write (iolog,'(a)', advance='no') ' Progress: '
           endif
         endif

         do j=1,ny1
           n1 = max ( j - npoints, 1 )
           n2 = min ( j + npoints, ny1 )
           if (lverbose .eq. '-verb2') then
             numpart = dble ( j )
             call showprogress ( iotty, iolog, lunix, numpart, numtotal, isumpct )
           endif

           do i=1,nx1
             m1 = max ( i - npoints, 1 )
             m2 = min ( i + npoints, nx1 )
             funstd(i,j) = 0.0d0
             
! First compute the mean value.

             nto = 0
             fmean = 0.0d0
             do m=m1,m2
               do n=n1,n2
                 if (fun2(m,n) .gt. almostzero) then
                   if (cimask(m-i,n-j) .ge. 1) then
                     nto = nto + 1
                     fmean = fmean + fun1(m,n)
                   endif
                 endif
               enddo
             enddo
             if (nto .ge. 1) then
               fmean = fmean / dble ( nto )
             endif
             
! Next compute the standard deviation.

             variance = 0.0d0
             do m=m1,m2
               do n=n1,n2
                 if (fun2(m,n) .gt. almostzero) then
                   if (cimask(m-i,n-j) .ge. 1) then
                     variance = variance + (fun1(m,n) - fmean)**2
                   endif
                 endif
               enddo
             enddo
             if (nto .ge. 2) then
               sigma1 = sqrt ( variance / dble ( nto - 1 ) )
             else
               sigma1 = 0.0d0
             endif
             if (fun2(i,j) .gt. almostzero) then
               funstd(i,j) = sigmas * sigma1
             endif
           enddo
         enddo

         outfits = outname(1:ion)
         if (ion .eq. 0) then
           outfits = imagename(1:lenimg)
           iof = lastc ( outfits )
           if (outfits(max(iof-4,1):iof) .ne. '.fits') then
             outfits = outfits(1:iof)//'.fits'
           endif                            
           iof = lastc ( outfits )
           if (outfits(max(iof-3,1):iof) .eq. '.std') then
             outfits = outfits(1:iof)//'.fits'
           elseif (outfits(max(iof-8,1):iof) .eq. '.std'//cpmx//'.fits') then
             outfits = outfits(1:iof)
           else
             outfits = outfits(1:max(iof-5,1))//'.std'//cpmx//'.fits'
           endif
         endif
         iof = lastc ( outfits )
         if (outfits(max(iof-4,1):iof) .ne. '.fits') then
           outfits = outfits(1:iof)//'.fits'
         else
           outfits = outfits(1:iof-5)//'.fits'
         endif                            
         iof = lastc ( outfits )

         if (lverbose .eq. '-verb2') 
     &     write (iotty,'(/a,f4.1,a)') '   Writing', sigmas, '*sigma stdev image '''//outfits(1:iof)//'''...'

         call wfits ( cfitsversion, nx1, ny1, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, funstd, dx1, dy1, object
     &              , crval1, crval2, outfits(1:iof), header, footer, cdate, ctime, creator, beam, blank, crota1, crota2, cd11
     &              , cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history )

         if (lverbose .eq. '-verb2') then
           funhi = 0.0d0
           do j=1,ny1
             do i=1,nx1
               funhi = max ( funhi, fun1(i,j) )
             enddo
           enddo
           funlo = funhi / 1.0d4
           bin = 7.0d-2
           nbins = max ( nint ( (log10 ( funhi ) - log10 ( funlo ) ) / bin ) + 1, 2 )

           allocate ( plane(nbins+1), funbin(nx1,ny1), stat=irc )
       
           if (irc .ne. 0) then
             write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (12).'
             stop 12
           endif
           do k=1,nbins+1
             plane(k) = 10.0d0**( log10 ( funlo ) + dble ( k - 1 ) * bin )
           enddo
           write (iotty,'(/a)') '   Computing stdev distribution within intensity bins...'
  
           open (111,file=outfits(1:iof)//'.bin.dat',status='unknown')
           write (111,'(a)') '# IMGSTAT = Compute Image Statistics = '//compda
           write (111,'(a)') '# Alexander Men’shchikov, DAp IRFU CEA Saclay, France.'
           write (111,'(a)') '# Using '//clibname(lb:7)//' library version'//cfitsversion//' by William D Pence.'
           write (111,'(a)') '#'
           write (111,'(a)') '# STDEV DISTRIBUTION FROM IMAGE: '//imagename(1:lenimg)
           write (111,'(a)') '#'
           write (111,'(a,1pe11.4)') '# FUNHI:', funhi
           write (111,'(a,1pe11.4)') '# FUNLO:', funlo
           write (111,'(a,1pe11.4)') '# LGBIN:', bin
           write (111,'(a,i5)') '# NBINS:', nbins
           write (111,'(a)') '#'
           write (111,'(a)') '# TABULATED QUANTITIES:'
           write (111,'(a)') '#'
           write (111,'(a)') '#   N ......... Bin number'
           write (111,'(a)') '#   BINLOW .... Lower edge of the bin'
           write (111,'(a)') '#   BINCOO .... Bin-centered coordinate'
           write (111,'(a)') '#   NPIXELS ... Number of pixels per bin'
           write (111,'(a)') '#   STDMEAN ... Mean stdev value within the bin'
           write (111,'(a)') '#   STDSTDM ... STDEV of STDMEAN within the bin'
           write (111,'(a)') '#'
           write (111,'(a)') '#      N    BINLOW      BINCOO      NPIXELS     STDMEAN    STDSTDM'
           write (111,'(a)') '#'
           kk = 0
           do k=1,nbins
             nto = 0
             stdmean = 0.0d0
             do j=1,ny1
               do i=1,nx1
                 if (fun2(i,j) .gt. almostzero) then
                   if (fun1(i,j) .gt. plane(k)) then
                     if (fun1(i,j) .le. plane(k+1)) then
                       nto = nto + 1
                       stdmean = stdmean + funstd(i,j)
                     endif
                   endif
                 endif
               enddo
             enddo
             if (nto .gt. 1) stdmean = stdmean / dble ( nto )
             variance = 0.0d0
             do j=1,ny1
               do i=1,nx1
                 if (fun2(i,j) .gt. almostzero) then
                   if (fun1(i,j) .gt. plane(k)) then
                     if (fun1(i,j) .le. plane(k+1)) then
                       nto = nto + 1
                       variance = variance + (funstd(i,j) - stdmean)**2
                     endif
                   endif
                 endif
               enddo
             enddo
             if (nto .gt. 1) then
               stdstdm = sqrt ( variance / dble ( nto - 1 ) )
             else
               stdstdm = 0.0d0
             endif
             if (nto .gt. 1) then
               kk = kk + 1
               write (111,'(i8,6(1pe12.4))') kk, plane(k), (plane(k) + plane(k+1)) / 2.0d0, dble ( nto ), stdmean, stdstdm
             endif
           enddo
           close (111)

           do j=1,ny1
             do i=1,nx1
               funbin(i,j) = 0.0d0
             enddo
           enddo
           do k=1,nbins
             do j=1,ny1
               do i=1,nx1
                 if (fun2(i,j) .gt. almostzero) then
                   if (fun1(i,j) .gt. plane(k)) then
                     if (fun1(i,j) .le. plane(k+1)) then
                       funbin(i,j) = plane(k)
                     endif
                   endif
                 endif
               enddo
             enddo
           enddo
           outfits = outfits(1:iof-5)//'.bin.fits'
           iof = lastc ( outfits )
           
           if (lverbose .eq. '-verb2') 
     &       write (iotty,'(/a,f4.1,a)') '   Writing binned intensity image '''//outfits(1:iof)//'''...'

           call wfits ( cfitsversion, nx1, ny1, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, funbin, dx1, dy1, object
     &                , crval1, crval2, outfits(1:iof), header, footer, cdate, ctime, creator, beam, blank, crota1, crota2, cd11
     &                , cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history )

           deallocate ( plane, funbin )
         endif
         deallocate ( funstd, cimask )
       endif
       
! Computing and writing median- or mean- filtered image.

       if (lmedimage .or. lmeaimage) then

         if (.not.lfnam2) then 
           if (lmedimage .and. lverbose .eq. '-verb2') write (iotty,'(a)') '   Mask image for median filtering not supplied.'
           if (lmeaimage .and. lverbose .eq. '-verb2') write (iotty,'(a)') '   Mask image for mean filtering not supplied.'
         endif

         allocate ( funm(nx1,ny1), funx(nx1*ny1), stat=irc )
         if (irc .ne. 0) then
           write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (150).'
           stop 150
         endif
         do j=1,ny1
           do i=1,nx1
             funm(i,j) = fun1(i,j)  
           enddo
         enddo
         if (.not.allocated ( fun2 )) then
           allocate ( fun2(nx1,ny1), stat=irc )
           if (irc .ne. 0) then
             write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (170).'
             stop 170
           endif
           do j=1,ny1
             do i=1,nx1
               fun2(i,j) = 1.0d0    
             enddo
           enddo
         endif
         do i=1,nx1*ny1
           funx(i) = 0.0d0    
         enddo

         write (cpmx,'(i4.4)') 2 * npoints + 1
         npmx = int ( log10 ( dble ( 2 * npoints + 1 ) ) ) + 1
         write (camx,'(i4.4)') int ( dble ( 2 * npoints + 1 ) * dx1 )
         namx = int ( log10 ( dble ( 2 * npoints + 1 ) * dx1 ) ) + 1

         if (lverbose .eq. '-verb2') then
           if (lmedimage) write (iotty,'(a)') '   Median filtering in a circular window of D = '//cpmx(4-npmx+1:4)//' pixels = '
     &                                      //camx(4-namx+1:4)//' arcsec...'
           if (lmeaimage) write (iotty,'(a)') '   Mean filtering in a circular window of D = '//cpmx(4-npmx+1:4)//' pixels = '
     &                                      //camx(4-namx+1:4)//' arcsec...'
         endif

         if (xpoints .le. 2.0d0) then
           rw2 = (xpoints + 0.5d0)**2
         else
           rw2 = (xpoints)**2
         endif
         npp1 = npoints + 1

         allocate ( cimask(-npoints:npoints,-npoints:npoints), stat=irc )
         if (irc .ne. 0) then
           write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (160).'
           stop 160
         endif

! The circular mask could also serve as weights for weighted median.
         
!!         if (lverbose .eq. '-verb2') write (iotty,'()')
         do l=-npoints,npoints
           ry2 = (dble ( l ))**2
           do k=-npoints,npoints
             rx2 = (dble ( k ))**2
             if (rx2 + ry2 .le. rw2) then
!!               cimask(k,l) = npoints + 1 - nint ( sqrt ( rx2 + ry2 ) )  !<-- arbitrary (linear circular) weights
               cimask(k,l) = 1 
             else
               cimask(k,l) = 0
             endif
           enddo
!!           if (lverbose .eq. '-verb2') write (iotty,'(1000i2)') (cimask(k,l),k=-npoints,min(-npoints+77,npoints))
         enddo

         if (ltiming .and. lverbose .eq. '-verb2') then
           cpu = 0.0d0
           wal = 0.0d0
           cpu_sum = 0.0d0
           wal_sum = 0.0d0
!!           if (iotty .gt. 0) write (iotty,'(/a)') ' STARTED: '//cdate//' '//ctime
!!           if (iolog .gt. 0) write (iolog,'(/a)') ' STARTED: '//cdate//' '//ctime
         endif
         if (ltiming) then
           cpu = timer ( 'cpu', 0.0d0 )
           wal = timer ( 'wal', 0.0d0 )
         endif

         isumpct = 0.0d0
         numtotal = dble ( ny1 )
         numpart = 0.0d0
       
         if (lverbose .eq. '-verb2') then
           if (iotty .gt. 0) then
             write (iotty,'(11x,a)') '__________________________________________________'
             write (iotty,'(a)', advance='no') ' Progress: '
             if (iotty .gt. 0) endfile   ( iotty, err=888 )
 888         continue 
             if (iotty .gt. 0) backspace ( iotty )
           endif       
           if (iolog .gt. 0) then
             write (iolog,'(11x,a)') '__________________________________________________'
             write (iolog,'(a)', advance='no') ' Progress: '
           endif
         endif

! Median filtering using an approximate binning method.

         if (lmedimage .and. lmedapprox) then
           do j=1,ny1
             n1 = max ( j - npoints, 1 )
             n2 = min ( j + npoints, ny1 )
             if (lverbose .eq. '-verb2') then
               numpart = dble ( j )
               call showprogress ( iotty, iolog, lunix, numpart, numtotal, isumpct )
             endif
             do i=1,nx1
               if (fun2(i,j) .gt. almostzero) then
                 m1 = max ( i - npoints, 1 )
                 m2 = min ( i + npoints, nx1 )
                 nm = 0
                 funmin = 1.0d99
                 funmax = -1.0d99
                 do m=m1,m2
                   do n=n1,n2
                     if (fun2(m,n) .gt. almostzero .and. cimask(m-i,n-j) .ge. 1) then
                       nm = nm + 1
                       funx(nm) = fun1(m,n)
                       funmin = min ( funmin, funx(nm) )
                       funmax = max ( funmax, funx(nm) )
                     endif
                   enddo
                 enddo
                 if (nm .ne. 0) then
                   funm(i,j) = medapprox ( nm, funx, funmin, funmax )
                 endif
               endif
             enddo
           enddo
         endif
         
! Median filtering using an accurate sorting method.

         if (lmedimage .and. .not.lmedapprox) then
           do j=1,ny1
             n1 = max ( j - npoints, 1 )
             n2 = min ( j + npoints, ny1 )
             if (lverbose .eq. '-verb2') then
               numpart = dble ( j )
               call showprogress ( iotty, iolog, lunix, numpart, numtotal, isumpct )
             endif
             do i=1,nx1
               if (fun2(i,j) .gt. almostzero) then
                 m1 = max ( i - npoints, 1 )
                 m2 = min ( i + npoints, nx1 )
                 nm = 0
                 do m=m1,m2
                   do n=n1,n2
                     if (fun2(m,n) .gt. almostzero .and. cimask(m-i,n-j) .ge. 1) then
                       nm = nm + 1
                       funx(nm) = fun1(m,n)
                     endif
                   enddo
                 enddo
                 if (nm .ne. 0) then
                   funm(i,j) = selectk ( (nm + 1) / 2, nm, funx )    !<-- SELECTK is slower by a factor of 1.5 than MEDAPPROX
                 endif
               endif
             enddo
           enddo
         endif

! Mean filtering.

         if (lmeaimage) then
           do j=1,ny1
             n1 = max ( j - npoints, 1 )
             n2 = min ( j + npoints, ny1 )
             if (lverbose .eq. '-verb2') then
               numpart = dble ( j )
               call showprogress ( iotty, iolog, lunix, numpart, numtotal, isumpct )
             endif
             do i=1,nx1
               if (fun2(i,j) .gt. almostzero) then
                 m1 = max ( i - npoints, 1 )
                 m2 = min ( i + npoints, nx1 )
                 nm = 0
                 funx(1) = 0.0d0
                 do m=m1,m2
                   do n=n1,n2
                     if (fun2(m,n) .gt. almostzero .and. cimask(m-i,n-j) .ge. 1) then
                       nm = nm + 1
                       funx(1) = funx(1) + fun1(m,n)
                     endif
                   enddo
                 enddo
                 if (nm .ne. 0) then
                   funm(i,j) = funx(1) / dble ( nm )
                 endif
               endif
             enddo
           enddo
         endif

         outfits = outname(1:ion)
         if (ion .eq. 0) then
           outfits = imagename(1:lenimg)
           iof = lastc ( outfits )
           if (outfits(max(iof-4,1):iof) .ne. '.fits') then
             outfits = outfits(1:iof)//'.fits'
           endif                            
           iof = lastc ( outfits )
           if (lmedimage) then
             if (outfits(max(iof-3,1):iof) .eq. '.med') then
               outfits = outfits(1:iof)//'.fits'
             elseif (outfits(max(iof-8,1):iof) .eq. '.med'//cpmx//'.fits') then
               outfits = outfits(1:iof)
             else
               outfits = outfits(1:max(iof-5,1))//'.med'//cpmx//'.fits'
             endif
           endif
           if (lmeaimage) then
             if (outfits(max(iof-3,1):iof) .eq. '.mea') then
               outfits = outfits(1:iof)//'.fits'
             elseif (outfits(max(iof-8,1):iof) .eq. '.mea'//cpmx//'.fits') then
               outfits = outfits(1:iof)
             else
               outfits = outfits(1:max(iof-5,1))//'.mea'//cpmx//'.fits'
             endif
           endif
         endif
         iof = lastc ( outfits )
         if (outfits(max(iof-4,1):iof) .ne. '.fits') then
           outfits = outfits(1:iof)//'.fits'
         else
           outfits = outfits(1:iof-5)//'.fits'
         endif                            
         iof = lastc ( outfits )

         if (lmedimage .and. lverbose .eq. '-verb2') write (iotty,'(/a)') '   Writing median-filtered '''//outfits(1:iof)//'''...'
         if (lmeaimage .and. lverbose .eq. '-verb2') write (iotty,'(/a)') '   Writing mean-filtered '''//outfits(1:iof)//'''...'

         call wfits ( cfitsversion, nx1, ny1, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, funm, dx1, dy1, object
     &              , crval1, crval2, outfits(1:iof), header, footer, cdate, ctime, creator, beam, blank, crota1, crota2, cd11
     &              , cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history )

         deallocate ( funm, funx, cimask )
       endif

       deallocate ( fun1 )
  
       if ((lmedimage .or. lmeaimage .or. lstd) .and. ltiming .and. lverbose .eq. '-verb2') then
         cpu_sum = cpu_sum + timer ( 'cpu', cpu )
         wal_sum = wal_sum + timer ( 'wal', wal )

! Total cpu usage, expressed in hours, minutes, and seconds.

         cputot = cpu_sum
         cpuhours = int ( cputot / 3600.0d0 )
         cpumins  = int ( cputot / 60.0d0 ) -  60 * cpuhours
         cpusecs  = int ( cputot + 0.5d0 ) - 3600 * cpuhours - 60 * cpumins

! Final wall clock time, expressed in hours, minutes, and seconds.

         wctot = wal_sum
         wchours = int ( wctot / 3600.0d0 )
         wcmins  = int ( wctot / 60.0d0 ) -  60 * wchours
         wcsecs  = int ( wctot + 0.5d0 ) - 3600 * wchours - 60 * wcmins
    
         call when ( lunix, ctime, cdate, ndate, 4 )
    
         write (ccpu,'(1pe9.2)') cpu_sum
         write (cwal,'(1pe9.2)') wal_sum
         write (cccpu,'(i3.2,a,i2.2,a,i2.2)') cpuhours,':',cpumins,':',cpusecs
         write (ccwal,'(i3.2,a,i2.2,a,i2.2)') wchours,':',wcmins,':',wcsecs
         ndc = 1
         ndw = 1
         if (cccpu(1:1) .eq. ' ') ndc = 2
         if (ccwal(1:1) .eq. ' ') ndw = 2
    
!!         if (iotty .gt. 0) write (iotty,'(/a)') ' ENDED: '//cdate//' '//ctime//' CPU:'//ccpu//' s '//cccpu(ndc:)
!!     &                                      //' WALL:'//cwal//' s '//ccwal(ndw:)
!!         if (iolog .gt. 0) write (iolog,'(/a)') ' ENDED: '//cdate//' '//ctime//' CPU:'//ccpu//' s '//cccpu(ndw:)
!!     &                                      //' WALL:'//cwal//' s '//ccwal(ndw:)
!!         if (iolog .gt. 0) close ( iolog )
       endif

 999   if (lverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'
       
!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine sizemeasure 
       
     &            ( iotty, nextr, nxmin, nxmax, nymin, nymax, nx, ny, mask, image, dx, dy, momxco, momyco, afwhm, bfwhm, atheta
     &            , equivrad, elongation, almostzero )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none

       integer       n, nx, ny, l, m, nextr, irc, iotty
       integer       nxmin(*), nxmax(*), nymin(*), nymax(*)
 
       real*8        dx, dy, pi, sqpixel, sourcearea, dblm, dbll, convtodegrees, sqreightlogtwo, pixel
     &             , intsxx, intsyy, intsxy, xcorsc, ycorsc, momp2, momm2, sqrmo, asig, bsig, intsd, ints0m, ints0
     &             , almostzero, intx, inty, int0, int0m, intd, nonzero
     
       real*8        image(nx,ny), mask(nx,ny), momxco(*), momyco(*), afwhm(*), bfwhm(*), atheta(*), equivrad(*), elongation(*)

       real*8 , allocatable :: momxx(:), momyy(:), momxy(:)

       parameter   ( pi = 3.14159265358979d0, convtodegrees = 180.0d0 / pi, sqreightlogtwo = sqrt ( 8.0d0 * log ( 2.0d0 ) ) )
!__________________________________________________________________________________________________________________________________
!                  
       pixel = dx
       sqpixel = dx * dy

       allocate ( momxx(nextr), momyy(nextr), momxy(nextr), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   IMGSTAT: ERROR: Trouble allocating memory (76).'
         stop 76
       endif

       do n=1,nextr
         intx = 0.0d0                        
         inty = 0.0d0                        
         int0 = 0.0d0
         nonzero = 0.0d0
         do m=nymin(n),nymax(n)              
           dblm = dble ( m )                 
           do l=nxmin(n),nxmax(n)            
             dbll = dble ( l )               
             if (mask(l,m) .gt. 0.9d0) then
               if (image(l,m) .gt. almostzero) then
                 intd = image(l,m) * sqpixel
                 int0 = int0 + intd
                 intx = intx + dbll * intd
                 inty = inty + dblm * intd
               endif
               nonzero = nonzero + 1.0d0
             endif
           enddo
         enddo
         int0m = int0 + almostzero
         momxco(n) = intx / int0m
         momyco(n) = inty / int0m
       enddo
       
       do n=1,nextr
         ints0 = 0.0d0
         intsxx = 0.0d0
         intsyy = 0.0d0
         intsxy = 0.0d0           
         do m=nymin(n),nymax(n)
           dblm = dble ( m )
           do l=nxmin(n),nxmax(n)
             dbll = dble ( l )       
             if (mask(l,m) .gt. 0.9d0) then

! ONLY FOR TESTING! 
! Limit the pixel values to ensure accurate FWHM for power-law sources; a correction factor is required (see below).

               if (image(l,m) .gt. 0.4d0*image(nint(momxco(n)),nint(momyco(n))) .and. 
     &             image(l,m) .le. 0.6d0*image(nint(momxco(n)),nint(momyco(n)))) then
               intsd = image(l,m) * sqpixel
               ints0 = ints0 + intsd
               xcorsc = dbll - momxco(n)
               ycorsc = dblm - momyco(n)
               intsxx = intsxx + xcorsc * xcorsc * intsd
               intsyy = intsyy + ycorsc * ycorsc * intsd
               intsxy = intsxy + xcorsc * ycorsc * intsd
               endif
             endif
           enddo
         enddo
         ints0m = ints0 + almostzero
         momxx(n) = intsxx / ints0m
         momyy(n) = intsyy / ints0m
         momxy(n) = intsxy / ints0m
       enddo

       do n=1,nextr
         momp2 = (momxx(n) + momyy(n)) / 2.0d0
         momm2 = (momxx(n) - momyy(n)) / 2.0d0
         
         if (abs ( momxy(n) ) .lt. almostzero) then
           if (momxx(n) .gt. momyy(n)) then
             momp2 = momxx(n)
             momm2 = momyy(n)
             atheta(n) = 90.0d0
           else
             momp2 = momyy(n)
             momm2 = momxx(n)
             atheta(n) = 0.0d0
           endif
           sqrmo = sqrt ( momm2**2 + momxy(n)**2 )
           asig  = sqrt ( momp2 + sqrmo ) * pixel
           if (momp2 .gt. sqrmo) then 
             bsig = sqrt ( momp2 - sqrmo ) * pixel
           else
             asig = 0.0d0
             bsig = 0.0d0
           endif
         else       
           sqrmo = sqrt ( momm2**2 + momxy(n)**2 )
           if (momp2 + sqrmo .ge. 0.0d0) then 
             asig  = sqrt ( momp2 + sqrmo ) * pixel
           endif
           if (momp2 - sqrmo .ge. 0.0d0) then 
             bsig  = sqrt ( momp2 - sqrmo ) * pixel
           endif
           if (momp2 + sqrmo .lt. 0.0d0 .or. momp2 - sqrmo .lt. 0.0d0) then 
             asig = 0.0d0
             bsig = 0.0d0
           endif
           atheta(n) = 180.0d0 - atan2 ( sqrmo + momm2, momxy(n) ) * convtodegrees
         endif                                        

! ONLY FOR TESTING!
! The empirical factor 1.1955 is here to correct for the fact that the pixel values used in the intensity moments (see above) 
! were limited to the { 0.4 Ipeak, 0.6 Ipeak } range, to get accurate FHWM sizes of both Gaussian and power-law sources.
         
         afwhm(n) = sqreightlogtwo * max ( asig, pixel )   * 1.1955d0
         bfwhm(n) = sqreightlogtwo * max ( bsig, pixel )   * 1.1955d0
         elongation(n) = afwhm(n) / bfwhm(n)
         
         sourcearea = nonzero * sqpixel
         equivrad(n) = sqrt ( sourcearea / pi )
       enddo
       
       deallocate ( momxx, momyy, momxy )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function medapprox
      
     &          ( npts, xdata, leftend, rightend )
!__________________________________________________________________________________________________________________________________
!
! Adapted by A. Men'shchikov from a code for the binapprox algorithm by Ryan J. Tibshirani (2008). 
! Webpage: http://www.stat.cmu.edu/~ryantibs/median/ Paper: http://www.stat.cmu.edu/~ryantibs/papers/median.pdf
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       nbins
       parameter   ( nbins = 1000 )
       integer       i, j, nb, kmed, npts, leftcount, ntotcount, ncounts(0:nbins)
       real*8        medapprox, scalefac, leftend, rightend, xdata(npts)           !! , xmean, sigma, dnpts
!__________________________________________________________________________________________________________________________________
!
!!       dnpts = dble ( npts )
!!       xmean = sum ( xdata ) / dnpts
!!       sigma = sqrt ( dot_product ( xdata - xmean, xdata - xmean ) / dnpts )
!!       leftend  = xmean - sigma
!!       rightend = xmean + sigma

       if (leftend .eq. rightend) then
         medapprox = leftend
         return
       endif

       scalefac = dble ( nbins ) / (rightend - leftend)
       leftcount = 0
       do i=0,nbins
         ncounts(i) = 0
       enddo

       do i=1,npts
         if (xdata(i) .lt. leftend) then
           leftcount = leftcount + 1
         else if (xdata(i) .lt. rightend) then
           nb = int ( (xdata(i) - leftend ) * scalefac )
           ncounts(nb) = ncounts(nb) + 1
         endif
       enddo

       if (mod ( npts, 2 ) .eq. 1) then
       
! Odd number 'npts' of data set elements, find the bin that contains the median.
       
         kmed = (npts + 1) / 2
         ntotcount = leftcount
         do i=0,nbins
           ntotcount = ntotcount + ncounts(i)
           if (ntotcount .ge. kmed) then
             medapprox = dble ( 2 * i + 1 ) / (2.0d0 * scalefac) + leftend
             exit
           endif
         enddo
       else
       
! Even number 'npts' of data set elements, find the bins that contain the medians.
       
         kmed = npts / 2
         ntotcount = leftcount
         do i=0,nbins
           ntotcount = ntotcount + ncounts(i)
           if (ntotcount .ge. kmed) then
             j = i
   1         continue
             if (ntotcount .eq. kmed) then
               j = j + 1
               ntotcount = ntotcount + ncounts(j)
               goto 1
             endif
             medapprox = dble ( i + j + 1 ) / (2.0d0 * scalefac) + leftend
             exit
           endif
         enddo
       endif

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine showprogress ( iotty, iolog, lunix, icount, ntotal, isumpct )
!__________________________________________________________________________________________________________________________________
!
! Display progress bar in percentage of current work; each point means that 2 % of work has been done (50 points in total)
!__________________________________________________________________________________________________________________________________
!
       logical       lshow, lunix
       character*3   dot
       integer       i, ishow, iotty, iolog
       real*8        icount, ntotal, isumpct, ipcent
       parameter   ( dot = '•' )
!__________________________________________________________________________________________________________________________________
!
       ipcent = icount * 100.0d0 / ntotal
       ishow = int ( ipcent / 2.0d0 - isumpct )
       lshow = mod ( int ( ipcent ), 2 ) .eq. 0
     
       if (lshow) then
         do i=1,ishow
           if (iotty .gt. 0) write (iotty,'(a)', advance='no') dot
           if (iolog .gt. 0) write (iolog,'(a)', advance='no') dot
           if (lunix) then
             if (iotty .gt. 0) endfile   ( iotty, err=888 )
 888         continue 
             if (iotty .gt. 0) backspace ( iotty )
           endif
           isumpct = isumpct + 1.0d0
         enddo
       endif
     
       if (abs ( icount - ntotal ) .lt. ntotal * 1.0d-14) then
          call sleep ( 1 )
          if (iotty .gt. 0) write (iotty,'(a)', advance='no') ' done'
          if (iolog .gt. 0) write (iolog,'(a)', advance='no') ' done'
          if (lunix) then
            if (iotty .gt. 0) endfile   ( iotty, err=999 )
 999        continue 
            if (iotty .gt. 0) backspace ( iotty )
          endif
          call sleep ( 1 )
          if (iotty .gt. 0) write (iotty,'()')
          if (iolog .gt. 0) write (iolog,'()')
       endif
     
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, header, footer, iotty
     &            , iolog, creator, beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale
     &            , wave, datamn, datamx, history, lverbose )
!__________________________________________________________________________________________________________________________________
!
! GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, DAp IRFU CEA Saclay      
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, anynull, myfile
       character*(*) ctype1, ctype2, creator, fname, header, footer(2), bunit, object, history, lverbose
       character*25  cdatetime
       character*80  footer1, footer2, comment
       integer       i, j, nx, ny, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, lastc, iotty, iolog, maxdim
     &             , ilc1, ilc2, rw, nfound, blank, stat01, stat02, stat03, stat04, stat05, stat06, stat07, stat08, stat09
     &             , stat10, stat11, stat12, stat13, stat14, stat15, stat16, stat17, stat18, stat19, stat20, stat21, stat22
     &             , stat23, stat24, stat25, stat26, stat27, stat28, stat29, stat30, stat31, stat32, stat0
       parameter   ( maxdim = 2 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, cd11, cd12, cd21, cd22, almostzero
       real*8        fun(nx,ny)
       real*8        as2deg, datamx, datamn
       parameter   ( as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      lastc, ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat01 = 0
       stat02 = 0
       stat03 = 0
       stat04 = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

!!       call ftgphx ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, bscale, bzero, blank, nblank, stat )

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
       if (stat .eq. 0 .and. (comment(1:22) .eq. 'Alexander Men’shchikov' .or. comment(1:22) .eq. 'Alexander Men''shchikov')) then
         myfile = .true.
       else
         myfile = .false.
         stat = 0
         creator = 'IMGSTAT'
       endif

       header = 'FITS Plot Header'

       if (myfile) then
         call ftgkys ( unit, 'HEADER' , header , comment, stat01 )
         call ftgkys ( unit, 'FOOTER1', footer1, comment, stat02 )
         call ftgkys ( unit, 'FOOTER2', footer2, comment, stat03 )
         ilc1 = lastc ( footer1 )
         ilc2 = lastc ( footer2 )
         if (ilc1 .ne. 0 .and. ilc2 .ne. 0) then
           footer(1) = footer1(1:ilc1)//footer2(1:ilc2)
         else
           footer(1) = 'Footer'
         endif
       else
         call ftgkys ( unit, 'OBJECT' , header, comment, stat04 )
         if (stat04 .ne. 0 .or. header(1:10) .eq. '          ') header = fname
         footer(1) = header
       endif

! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       stat08 = 0
       stat0 = stat01 + stat02 + stat03 + stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15
     &       + stat16 + stat17 + stat18 + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28
     &       + stat29 + stat30 + stat31

       if (lverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat01 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword HEADER absent or error'
         if (stat02 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER1 absent or error'
         if (stat03 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER2 absent or error'
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
!!!         if (stat08 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BLANK absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat29 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EPOCH absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
         !!!         if (stat0  .gt. 0) write (iotty,'()')
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = 'No object specified'

! Read the FITS file into the 2D array.

       call ftg2dd ( unit, group, dble ( blank ), naxes(1), naxes(1), naxes(2), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

! Correct if the input images have no CDELT* keywords, but do have CD*_* keywords.

       if (stat24 .ne. 0 .and. abs ( cd11 ) .gt. almostzero .and. abs ( cd12 ) .lt. almostzero) then
         dxdeg = cd11
       endif
       if (stat25 .ne. 0 .and. abs ( cd22 ) .gt. almostzero .and. abs ( cd21 ) .lt. almostzero) then
         dydeg = cd22
       endif

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       nx = naxes(1)
       ny = naxes(2)
       funmin = 1.0d30
       funmax =-1.0d30
       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j) )) then  
             funmin = min ( funmin, fun(i,j) )
             funmax = max ( funmax, fun(i,j) )
           endif  
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, header
     &            , footer, cdate, ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale
     &            , wave, datamn, datamx, history )
!__________________________________________________________________________________________________________________________________
!
! GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, DAp IRFU CEA Saclay      
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend
       character*(*) fname, ctype1, ctype2, object, creator, header, footer(2), ctime, cdate, bunit, cfitsversion, history
       character*19  cdatetime
       integer       i, j, nx, ny, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, naxes(2), iat, iflen
     &             , lastc, blank
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, rot1, rot2, epoch, equinox, datamn
     &             , cosrota, sinrota, rotangle, datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, pi
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      lastc
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 2-D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 2
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       datamn = 1.0d30
       datamx =-1.0d30
       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j) )) then
             if (fun(i,j) .lt. datamn) datamn = fun(i,j)
             if (fun(i,j) .gt. datamx) datamx = fun(i,j)
           endif
         enddo
       enddo

! When packing numbers using the formulas below, I got values of the function different in the 5-th digit from DATAMIN and
!  DATAMAX placed in the header when writing the data. From now on I will use bzero = 0 bscale = 1 and bitpix = -32, as these
! seem to cure the problem.

cc!      bitpix = 16
cc!      bzero  = 0.5d0 * (funmin + funmax)
cc!      bscale = dabs ( funmax - bzero ) / 32767.0d0

       bitpix = -32     !<-- single precision floating-point values
!!       bzero  = 0.0d0
!!       bscale = 1.0d0

       cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

       if (cfitsversion .eq. '5.030') then
         call fdelete ( fname, status )
       endif

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Create the new empty FITS file.

       if (cfitsversion .eq. '5.030') then
         call ftinit ( unit, fname, blocksze, status )
       else
         call ftinit ( unit, '!'//fname, blocksze, status )
       endif

! Write the required header keywords.

       call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

       call ftpkys ( unit, 'CREATOR', creator,'Alexander Men’shchikov, DAp IRFU CEA Saclay', status )
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'HEADER' , header, 'title', status )

       iat = 0
       iat = index ( footer(1), ' |F' )
       if (iat .eq. 0) iat = index ( footer(1), ' mm ' )
       iflen = lastc ( footer(1) )

       if (iat .gt. 0) then
         call ftpkys ( unit, 'FOOTER1', footer(1)(1:iat-1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', footer(1)(iat:iflen), 'part 2', status )
       else
         call ftpkys ( unit, 'FOOTER1', footer(1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', ' ', 'part 2', status )
       endif
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )

! Write all parameters.

       dxdeg = dx * as2deg
       dydeg = dy * as2deg

! Correct if the input images had cd11 or cd22 set to zero.

       if (abs ( cd11 ) .lt. almostzero .or. abs ( cd11 ) .gt. dxdeg) then
         cd11 = -dxdeg
         cd12 = 0.0d0
       endif
       if (abs ( cd22 ) .lt. almostzero .or. abs ( cd22 ) .gt. dxdeg) then
         cd21 = 0.0d0
         cd22 = dydeg
       endif
       
       cosrota = - cd11 / dxdeg
       sinrota = - cd12 / dxdeg
       rotangle = acos ( cosrota ) * 180.0d0 / pi
       
       if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = 180.0d0 - rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = 180.0d0 + rotangle
       elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = - rotangle
       endif
       rot2 = rot1

       call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
       call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
       call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
       if  (blank .gt. 0)
     & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
       call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
       call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
       call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
       call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CDELT1' , -dxdeg, 13 , 'coordinate increment along axis', status )
       call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
       call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
       call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
       call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
       call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
       call ftpkyd ( unit, 'EPOCH'  , epoch  , 13, ' ', status )
       call ftpkyd ( unit, 'EQUINOX', equinox, 13, ' ', status )
       call ftphis ( unit, history, status )

       if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'image was convolved with the beam (arcsec)', status )

! Write the array to the FITS file.

       call ftp2dd ( unit, group, nx, nx, ny, fun, status )

! Close the file.

       call ftclos ( unit, status )

! Free the unit number.

       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end
                                                                                                                                   
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||