
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program operate
!__________________________________________________________________________________________________________________________________
!
! Operate on two FITS images.
!
! This code is part of GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay  
!__________________________________________________________________________________________________________________________________
!                 
       implicit      none

       integer       iotty
       parameter   ( iotty = 6 )

       logical       lfnama, lfnamb, lunix
                                          
       character*5   cfpptime
       character*6   cfitsversion, lverbose         
       character*7   clibname
       character*8   ctime
       character*10  cdate
       character*21  compda
       character*80  object, bunit, bunit1, bunit2, ctype1, ctype2, ctype11, ctype12, ctype21, ctype22, object1, object2, creator
     &             , history, history1, history2
       character*128 header, footer(2), header2, footer2(2)
       character*500 filenama, filenamb, outname, arg1, arg2, arg3, arg4, arg5, arg6

       integer       i, j, ion, fnlen1, fnlen2, ndate, nx, ny, nx1, ny1, nx2, ny2, firstb, lastc, lasts, ia1
     &             , ia2, ia3, ia4, ia5, ia6, isp1, isp2, blank, blank1, blank2, irc, lb

       real          fitsvers
       real*8        dx, dy, dx1, dy1, dx2, dy2, ra, dec, beam
     &             , beam1, beam2, epoch, epoch2, equinox, equinox1, equinox2, epoch1, ra1, ra2, dec1, dec2, bzero, bscale
     &             , bzero1, bzero2, bscale1, bscale2, funmin, funmax, crpix11, crpix12, crval11, crval12, datamin1, datamax1
     &             , crpix1, crpix2, crval1, crval2, crota11, crota12, crota1, crota2, datamin2, datamax2, datamin, datamax
     &             , crpix21, crpix22, crval21, crval22, crota21, crota22, cd11, cd12, cd21, cd22, wave, wave1, wave2
     &             , cd111, cd112, cd121, cd122, cd211, cd212, cd221, cd222

       real*8, allocatable :: funa(:,:), funb(:,:)

       external      firstb, lastc, rfits, wfits, when
!__________________________________________________________________________________________________________________________________
!
! Determine operating system

       call osystem ( lunix )

       call ftvers ( fitsvers )

       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       compda = __DATE__//' '//cfpptime

! Get command line parameters (file names).

       call getarg ( 1, arg1 )
       call getarg ( 2, arg2 )
       call getarg ( 3, arg3 )
       call getarg ( 4, arg4 )
       call getarg ( 5, arg5 )
       call getarg ( 6, arg6 )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )
       ia6 = lastc ( arg6 )

       lverbose = '-verb2'
       if (ia6 .gt. 0 .and. arg6(1:ia6-1) .eq. '-verb') lverbose = arg6(1:ia6)
       if (ia5 .gt. 0 .and. arg5(1:ia5-1) .eq. '-verb') lverbose = arg5(1:ia5)
       if (ia4 .gt. 0 .and. arg4(1:ia4-1) .eq. '-verb') lverbose = arg4(1:ia4)
       if (ia3 .gt. 0 .and. arg3(1:ia3-1) .eq. '-verb') lverbose = arg3(1:ia3)

       if (lverbose .eq. '-verb2') write (iotty,'(/a$)') ' OPERATE = Operate on Two FITS Images = '//compda
       if (arg1(1:ia1) .eq. ':') stop
       if (lverbose .eq. '-verb2') write (iotty,'()')
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Alexander Men`shchikov, SAp IRFU CEA Saclay, France.'
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                                //' by William D Pence.'
       if (ia3 .eq. 0 .or. 
     &     (arg2(1:ia2) .ne. '+' .and. arg2(1:ia2) .ne. '-' .and. arg2(1:ia2) .ne. 'x' .and. arg2(1:ia2) .ne. '%' .and.
     &      arg2(1:ia2) .ne. '=' .and. arg2(1:ia2) .ne. 'o' .and. arg2(1:ia2) .ne. 'l' .and. arg2(1:ia2) .ne. 'g' .and. 
     &      arg2(1:ia2) .ne. 'r' .and. arg2(1:ia2) .ne. 'h')) then
         write (iotty,'(/a)') ' USAGE: operate <image1> {+|-|x|%|=|o|l|g|r|h} <image2> [-o <outname>] [-verb{0|1|2}]'
         write (iotty,'(  )')   
         write (iotty,'(a )') ' This is a versatile utility that allows various modifications of two FITS'  
         write (iotty,'(a )') ' images; <image1>, <image2>, and <outname> are the input and output images.'
         write (iotty,'(a )') ' Between <image1> and <image2> goes an operator (one of the above symbols)'
         write (iotty,'(a )') ' that specifies one of several binary operations on the two input images.'
         write (iotty,'(a )') ' The following briefly describes the operations:'  
         write (iotty,'(  )')   
         write (iotty,'(a )') ' + ..... addition: outname = image1 + image2'
         write (iotty,'(a )') ' - ..... subtraction: outname = image1 - image2'
         write (iotty,'(a )') ' x ..... multiplication: outname = image1 * image2'
         write (iotty,'(a )') ' % ..... division: outname = image1 / image2'
         write (iotty,'(a )') ' = ..... relative difference: outname = (image1 - image2) / image2'
         write (iotty,'(a )') ' o ..... zeroing: outname = 0 where image1 < image2, otherwise outname = image1'
         write (iotty,'(a )') ' l ..... minimization: outname = min ( image1, image2)'
         write (iotty,'(a )') ' g ..... maximization: outname = max ( image1, image2)'
         write (iotty,'(a )') ' r ..... replacement of non-zero pixels: outname = image1 -> image2'
         write (iotty,'(a )') ' h ..... replacement of the file header: outname = image2 (header1)'
         if (ia2 .gt. 0 .and. ia3 .gt. 0) then
           write (iotty,'(/a)') '   OPERATE: ERROR: Unknown operator: '''//arg2(1:ia2)//''''
         endif
         stop 99
       endif

       filenama = arg1(1:ia1)
       fnlen1 = lastc ( filenama )
       
       filenamb = arg3(1:ia3)
       fnlen2 = lastc ( filenamb )

! Check if the input file exists and open it.

       inquire ( file=filenama(1:fnlen1), exist=lfnama )
       if (.not.lfnama) then

! Maybe the file name extension was omitted?

         filenama = filenama(1:fnlen1)//'.fits'
         fnlen1 = lastc ( filenama )
         inquire ( file=filenama(1:fnlen1), exist=lfnama )
       endif

! Check if the input file exists and open it.

       inquire ( file=filenamb(1:fnlen2), exist=lfnamb )
       if (.not.lfnamb) then

! Maybe the file name extension was omitted?

         filenamb = filenamb(1:fnlen2)//'.fits'
         fnlen2 = lastc ( filenamb )
         inquire ( file=filenamb(1:fnlen2), exist=lfnamb )
       endif
                             
       isp1 = lasts ( filenama ) + 1
       isp2 = lasts ( filenamb ) + 1
       
       outname = filenama(isp1:fnlen1-5)//'.'//arg2(1:ia2)//'.'//filenamb(isp2:fnlen2)
       if (arg2(1:2) .eq. '-o') outname = arg3
       if (arg3(1:2) .eq. '-o') outname = arg4
       if (arg4(1:2) .eq. '-o') outname = arg5
       ion = lastc ( outname )
       if (outname(ion-4:ion) .ne. '.fits') then
         outname = outname(1:ion)//'.fits'
         ion = lastc ( outname )
       endif

       if (lverbose .eq. '-verb1') write (iotty,'(/a)') ' OPERATE: '//filenama(isp1:fnlen1-5)//' '//arg2(1:ia2)
     &                                                     //' '//filenamb(isp2:fnlen2-5)
       if (.not.lfnama) then
         write (iotty,'(/a)') '   OPERATE: ERROR: File '''//filenama(1:fnlen1-5)//''' not found.'
         stop 1
       endif
       if (.not.lfnamb) then
         write (iotty,'(/a)') '   OPERATE: ERROR: File '''//filenamb(1:fnlen2-5)//''' not found.'
         stop 2
       endif

! Determine numbers of pixels in the FITS image.

       call getnumpix ( filenama(1:fnlen1), nx1, ny1 )

! Allocate the image arrays.
       
       allocate ( funa(nx1,ny1), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   OPERATE: ERROR: Trouble allocating memory (10).'
         stop 10
       endif
       
       if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   Reading 1st image '''//filenama(1:fnlen1)//'''...'

       call rfits ( nx1, ny1, bunit1, ctype11, ctype12, crpix11, crpix12, crval11, crval12, funa, dx1, dy1, object1, ra1
     &            , dec1, filenama(1:fnlen1), header, footer, iotty, 0, creator, beam1, funmin, funmax, blank1, crota11, crota12
     &            , cd111, cd112, cd121, cd122, epoch1, equinox1, bzero1, bscale1, wave1, datamin1, datamax1, history1, lverbose )

       if (lverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax

! Determine numbers of pixels in the FITS image.

       call getnumpix ( filenamb(1:fnlen2), nx2, ny2 )

       allocate ( funb(nx2,ny2), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   OPERATE: ERROR: Trouble allocating memory (20).'
         stop 20
       endif

       if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Reading 2nd image '''//filenamb(1:fnlen2)//'''...'

       call rfits ( nx2, ny2, bunit2, ctype21, ctype22, crpix21, crpix22, crval21, crval22, funb, dx2, dy2, object2, ra2
     &            , dec2, filenamb(1:fnlen2), header2, footer2, iotty, 0, creator, beam2, funmin, funmax, blank2, crota21, crota22
     &            , cd211, cd212, cd221, cd222, epoch2, equinox2, bzero2, bscale2, wave2, datamin2, datamax2, history2, lverbose )

       if (lverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax

       if (beam1 .eq. beam2) then
         beam = beam1
       else
         beam = 0.0d0
       endif

       if ((nx1 .ne. nx2 .or. ny1 .ne. ny2) .and. arg2(1:ia2) .ne. 'h') then
         write (iotty, '(/a)') '   OPERATE: ERROR: The images have different numbers of pixels:'
         write (iotty, '(a )') '                   '//filenama(1:fnlen1)
         write (iotty, '(a )') '                   '//filenamb(1:fnlen2)
         stop 3
       else
         if (arg2(1:ia2) .ne. 'h') then
           nx = nx1
           ny = ny1
         else
           nx = nx2
           ny = ny2
         endif
         bunit = bunit1
         ctype1 = ctype11
         ctype2 = ctype12
         dx = dx1
         dy = dy1
         crpix1 = crpix11
         crpix2 = crpix12
         crval1 = crval11
         crval2 = crval12
         blank = blank1
         crota1 = crota11
         crota2 = crota12
         cd11 = cd111
         cd12 = cd112
         cd21 = cd121
         cd22 = cd122
         epoch = epoch1
         equinox = equinox1
         object = object1
         ra = ra1
         dec = dec1
         bzero = bzero1
         bscale = bscale1
         wave = wave1
         history = history1
       endif

       if (arg2(1:ia2) .eq. '+') then
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Adding images: image1 + image2...'
         do j=1,ny
           do i=1,nx
             funb(i,j) = funa(i,j) + funb(i,j)
           enddo
         enddo
       endif
       if (arg2(1:ia2) .eq. '-') then
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Subtracting images: image1 - image2...' 
         do j=1,ny
           do i=1,nx
             funb(i,j) = funa(i,j) - funb(i,j)
           enddo
         enddo
       endif
       if (arg2(1:ia2) .eq. 'x') then
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Multiplying images: image1 * image2...'
         do j=1,ny
           do i=1,nx
              funb(i,j) = funa(i,j) * funb(i,j)
           enddo
         enddo
       endif
       if (arg2(1:ia2) .eq. '%') then 
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Dividing images: image1 / image2...'
         do j=1,ny
           do i=1,nx
             if (abs ( funb(i,j) ) .gt. 1.0d-30) funb(i,j) = funa(i,j) / (funb(i,j) + 1.0d-30)
           enddo
         enddo
       endif
       if (arg2(1:ia2) .eq. '=') then
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Differencing images: (image1 - image2) / image2...'
         do j=1,ny
           do i=1,nx
             if (abs ( funa(i,j) / (funb(i,j) + 1.0d-30) ) .le. 1.0d10) funb(i,j) = funa(i,j) / (funb(i,j) + 1.0d-30) - 1.0d0
           enddo
         enddo
       endif
       if (arg2(1:ia2) .eq. 'o') then
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Zeroing image1 where image1 < image2...'
         do j=1,ny
           do i=1,nx
             if(funa(i,j) .gt. funb(i,j)) then
               funb(i,j) = funa(i,j)
             else  
               funb(i,j) = 0.0d0
             endif
           enddo
         enddo
       endif
       if (arg2(1:ia2) .eq. 'l') then
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Minimizing images: min ( image1, image2 )...'
         do j=1,ny
           do i=1,nx
             funb(i,j) = min ( funa(i,j), funb(i,j) )
           enddo
         enddo
       endif
       if (arg2(1:ia2) .eq. 'g') then
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Maximizing images: max ( image1, image2 )...'
         do j=1,ny
           do i=1,nx
             funb(i,j) = max ( funa(i,j), funb(i,j) )
           enddo
         enddo
       endif
       if (arg2(1:ia2) .eq. 'r') then
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Replacing non-zero pixels: image1 -> image2...'
         do j=1,ny
           do i=1,nx
             if(funa(i,j) .gt. 1.0d-30) then
               funb(i,j) = funa(i,j)
             endif
           enddo
         enddo
       endif
       if (arg2(1:ia2) .eq. 'h') then
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Replacing header: image1 -> image2...'
         header  = header2
         footer  = footer2
         wave    = wave2
         object  = object2
         epoch   = epoch2
         equinox = equinox2
         history = history2
       endif

! Get the current date and time.

       call when ( lunix, ctime, cdate, ndate, 4 )

       if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Writing output image '''//outname(1:ion)//'''...'

       header = outname(1:ion)
       
       call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, funb, dx, dy, object, crval1
     &            , crval2, outname(1:ion), header, footer, cdate, ctime, creator, beam, blank, crota1, crota2, cd11, cd12, cd21
     &            , cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history )

       deallocate ( funa, funb )
       
       if (lverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, header, footer, iotty
     &            , iolog, creator, beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale
     &            , wave, datamn, datamx, history, lverbose )
!__________________________________________________________________________________________________________________________________
!
! GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay      
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, anynull, myfile
       character*(*) ctype1, ctype2, creator, fname, header, footer(2), bunit, object, history, lverbose
       character*25  cdatetime
       character*80  footer1, footer2, comment
       integer       i, j, nx, ny, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, lastc, iotty, iolog, maxdim
     &             , ilc1, ilc2, rw, nfound, blank, stat01, stat02, stat03, stat04, stat05, stat06, stat07, stat08, stat09
     &             , stat10, stat11, stat12, stat13, stat14, stat15, stat16, stat17, stat18, stat19, stat20, stat21, stat22
     &             , stat23, stat24, stat25, stat26, stat27, stat28, stat29, stat30, stat31, stat32, stat0, nan
       parameter   ( maxdim = 2 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, cd11, cd12, cd21, cd22
       real*8        fun(nx,ny)
       real*8        as2deg, datamx, datamn
       parameter   ( as2deg = 1.0d0 / 3600.0d0 )
       external      lastc, ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat01 = 0
       stat02 = 0
       stat03 = 0
       stat04 = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
       if (stat .eq. 0 .and. (comment(1:22) .eq. 'Alexander Men`shchikov' .or. comment(1:22) .eq. 'Alexander Men''shchikov')) then
         myfile = .true.
       else
         myfile = .false.
         stat = 0
         creator = 'OPERATE'
       endif

       header = 'FITS Plot Header'

       if (myfile) then
         call ftgkys ( unit, 'HEADER' , header , comment, stat01 )
         call ftgkys ( unit, 'FOOTER1', footer1, comment, stat02 )
         call ftgkys ( unit, 'FOOTER2', footer2, comment, stat03 )
         ilc1 = lastc ( footer1 )
         ilc2 = lastc ( footer2 )
         if (ilc1 .ne. 0 .and. ilc2 .ne. 0) then
           footer(1) = footer1(1:ilc1)//footer2(1:ilc2)
         else
           footer(1) = 'Footer'
         endif
       else
         call ftgkys ( unit, 'OBJECT' , header, comment, stat04 )
         if (stat04 .ne. 0 .or. header(1:10) .eq. '          ') header = fname
         footer(1) = header
       endif

! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       stat08 = 0
       stat0 = stat01 + stat02 + stat03 + stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15
     &       + stat16 + stat17 + stat18 + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28
     &       + stat29 + stat30 + stat31

       if (lverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat01 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword HEADER absent or error'
         if (stat02 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER1 absent or error'
         if (stat03 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER2 absent or error'
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat29 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EPOCH absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = 'No object specified'

! Read the FITS file into the 2D array.

       call ftg2dd ( unit, group, dble ( blank ), naxes(1), naxes(1), naxes(2), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       nx = naxes(1)
       ny = naxes(2)
       funmin = 1.0d30
       funmax =-1.0d30
       nan = 0
       do j=1,ny
         do i=1,nx
           if (isnan ( fun(i,j) )) then  
             nan = nan + 1
             if (lverbose .ne. '-verb0') then
               if (nan .lt. 100) then
                 write (iotty,'(a,i4,a,i4,a)') '   RFITS: Element (', i,',', j, ') contains not a number (NaN)...'
               elseif (nan .eq. 100) then
                 write (iotty,'(a)') '   RFITS: Many NaNs, no more info on the elements.'
               endif
             endif
           else
             funmin = min ( funmin, fun(i,j) )
             funmax = max ( funmax, fun(i,j) )
           endif  
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, header
     &            , footer, cdate, ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale
     &            , wave, datamn, datamx, history )
!__________________________________________________________________________________________________________________________________
!
! GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay      
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend
       character*(*) fname, ctype1, ctype2, object, creator, header, footer(2), ctime, cdate, bunit, cfitsversion, history
       character*19  cdatetime
       integer       i, j, nx, ny, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, naxes(2), iat, iflen
     &             , lastc, blank
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, dfun, rot1, rot2, epoch, equinox, datamn
     &             , cosrota, sinrota, rotangle, datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, pi
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      lastc
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 2-D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 2
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       datamn = 1.0d30
       datamx =-1.0d30
       do j=1,ny
         do i=1,nx
           dfun = fun(i,j)
           if (dfun .lt. datamn ) datamn = dfun
           if (dfun .gt. datamx ) datamx = dfun
         enddo
       enddo

       bitpix = -32     !<-- single precision floating-point values

       cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

       if (cfitsversion .eq. '5.030') then
         call fdelete ( fname, status )
       endif

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Create the new empty FITS file.

       if (cfitsversion .eq. '5.030') then
         call ftinit ( unit, fname, blocksze, status )
       else
         call ftinit ( unit, '!'//fname, blocksze, status )
       endif

! Write the required header keywords.

       call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

       call ftpkys ( unit, 'CREATOR', creator,'Alexander Men`shchikov, SAp IRFU CEA Saclay', status )
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'HEADER' , header, 'title', status )

       iat = 0
       iat = index ( footer(1), ' |F' )
       if (iat .eq. 0) iat = index ( footer(1), ' mm ' )
       iflen = lastc ( footer(1) )

       if (iat .gt. 0) then
         call ftpkys ( unit, 'FOOTER1', footer(1)(1:iat-1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', footer(1)(iat:iflen), 'part 2', status )
       else
         call ftpkys ( unit, 'FOOTER1', footer(1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', ' ', 'part 2', status )
       endif
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )

! Write all parameters.

       dxdeg = dx * as2deg
       dydeg = dy * as2deg

! Correct if the input images had cd11 or cd22 set to zero.

       if (abs ( cd11 ) .lt. almostzero .or. abs ( cd11 ) .gt. dxdeg) then
         cd11 = -dxdeg
         cd12 = 0.0d0
       endif
       if (abs ( cd22 ) .lt. almostzero .or. abs ( cd22 ) .gt. dxdeg) then
         cd21 = 0.0d0
         cd22 = dydeg
       endif
       
       cosrota = - cd11 / dxdeg
       sinrota = - cd12 / dxdeg
       rotangle = acos ( cosrota ) * 180.0d0 / pi
       
       if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = 180.0d0 - rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = 180.0d0 + rotangle
       elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = - rotangle
       endif
       rot2 = rot1

       call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
       call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
       call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
       if  (blank .gt. 0)
     & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
       call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
       call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
       call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
       call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CDELT1' , -dxdeg, 13 , 'coordinate increment along axis', status )
       call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
       call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
       call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
       call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
       call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
       call ftpkyd ( unit, 'EPOCH'  , epoch  , 13, ' ', status )
       call ftpkyd ( unit, 'EQUINOX', equinox, 13, ' ', status )
       call ftphis ( unit, history, status )

       if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'image was convolved with the beam (arcsec)', status )

! Write the array to the FITS file.

       call ftp2dd ( unit, group, nx, nx, ny, fun, status )

! Close the file.

       call ftclos ( unit, status )

! Free the unit number.

       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||