
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program readhead
!__________________________________________________________________________________________________________________________________
!
! Read and display headers of a FITS image, optionally save some keywords of its header.
!
! May contain fragments from "The FITSIO Cookbook" v1.1 (April 1995) by William D Pence (HEASARC).
!
! This code is part of GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!                 
       implicit      none

       logical       lfname, lunix, lsave
                                         
       character*5   cfpptime
       character*6   cfitsversion, lverbose                    
       character*8   keyword1, keyword2, clibname
       character*9   action
       character*21  compda
       character*500 filename, arg1, arg2, arg3, arg4, arg5

       integer       firstb, lastc, ia1, ia2, ia3, ia4, ia5, iac, fnlen, lb, iotty

       real          fitsvers
       
       parameter   ( iotty = 6 )

       external      firstb, lastc, rphead
!__________________________________________________________________________________________________________________________________
!
! Determine operating system

       call osystem ( lunix )

       call ftvers ( fitsvers )

       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       compda = __DATE__//' '//cfpptime

! Get command line parameter (file name).

       call getarg ( 1, arg1 )
       call getarg ( 2, arg2 )
       call getarg ( 3, arg3 )
       call getarg ( 4, arg4 )
       call getarg ( 5, arg5 )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )

       lverbose = '-verb2'
       if (ia5 .gt. 0 .and. arg5(1:ia5-1) .eq. '-verb') lverbose = arg5(1:ia5)
       if (ia4 .gt. 0 .and. arg4(1:ia4-1) .eq. '-verb') lverbose = arg4(1:ia4)
       if (ia3 .gt. 0 .and. arg3(1:ia3-1) .eq. '-verb') lverbose = arg3(1:ia3)

       if (lverbose .eq. '-verb2') write (iotty,'(/a$)') ' READHEAD = Display FITS Headers = '//compda
       if (arg1(1:ia1) .eq. ':') stop
       if (lverbose .eq. '-verb2') write (iotty,'()')
       if (lverbose .eq. '-verb2') write (iotty,'(a)') ' Alexander Men`shchikov, SAp IRFU CEA Saclay, France.'
       if (lverbose .eq. '-verb2') write (iotty,'(a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                               //' by William D Pence.'

       if (ia1 .eq. 0 .or. (arg1(1:4) .eq. 'save' .and. (ia2 .eq. 0 .or. ia3 .eq. 0))) then
         write (iotty,'(/a)') ' USAGE: readhead [save[+show] <keyword1> [<keyword2>]] <image> [-verb{0|1|2}]'
         stop 99
       endif
       
! Get the file name.
                   
       keyword1 = 'xxxxxxx'
       keyword2 = 'yyyyyyy'
       lsave = .false.
       if (ia4 .gt. 0 .and. arg4(1:ia4-1) .ne. '-verb') then 
         if (arg1(1:4) .eq. 'save') lsave = .true.
         action   = arg1
         keyword1 = arg2
         keyword2 = arg3
         filename = arg4
       else
         if (ia3 .gt. 0 .and. arg3(1:ia3-1) .ne. '-verb') then
           if (arg1(1:4) .eq. 'save') lsave = .true.
           action   = arg1
           keyword1 = arg2
           filename = arg3
         else
           filename = arg1
         endif
       endif
       iac = lastc ( action )
       ia1 = lastc ( keyword1 )
       
       if (keyword2 .eq. 'yyyyyyy') then
         ia2 = 0
       else
         ia2 = lastc ( keyword2 )
       endif
       fnlen = lastc ( filename )

! Check if the input file exists and open it.

       inquire ( file=filename(1:fnlen), exist=lfname )

       if (.not.lfname) then

! Maybe the file name extension was omitted?

         inquire ( file=filename(1:fnlen)//'.fits', exist=lfname )

         filename = filename(1:fnlen)//'.fits'
         fnlen = lastc ( filename )
       endif

       if (lverbose .eq. '-verb1') then
         if (ia2 .eq. 0) then
           write (iotty,'(/a)') ' READHEAD: '//action(1:iac)//' '//keyword1(1:ia1)//' '//filename(1:fnlen-5)
         else  
           write (iotty,'(/a)') ' READHEAD: '//action(1:iac)//' '//keyword1(1:ia1)//' '//keyword2(1:ia2)//' '//filename(1:fnlen-5)
         endif
       endif

       if (.not.lfname) then
         write (iotty,'(/a)') '   READHEAD: ERROR: File '''//filename(1:fnlen)//''' not found.'
         stop 1
       endif

       if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   Reading header of '''//filename(1:fnlen)//'''...'

! Read and print FITS header.

       call rphead ( iotty, filename(1:fnlen), lsave, action, keyword1(1:ia1), keyword2(1:ia2), ia1, ia2, lverbose )

       if (lverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rphead 
       
     &            ( iotty, filename, lsave, action, keyword1, keyword2, ia1, ia2, lverbose )
!__________________________________________________________________________________________________________________________________
!
!  Print out all the header keywords in all extensions of a FITS file.
!__________________________________________________________________________________________________________________________________
!
       implicit      none              
       logical       lsave
       integer       status, unit, readwrite, blocksize, nkeys, nspace, hdutype, i, ik1, ik2, ia, ia1, ia2, ic, lastc, iotty
       character*(*) filename, keyword1, keyword2, action, lverbose
       character*80 record, record2
       character*256 cline
       real*8        cdelt1, cdelt2
       external      ftgiou, ftopen, ftghsp, ftgrec, ftmrhd, ftcmsg, ftclos, ftfiou, printerr, lastc
!__________________________________________________________________________________________________________________________________
!
       status    = 0
       readwrite = 0
       record2 = ' '
       cline = ' '
       if (.not. lsave .or. action .eq. 'save+show') write (iotty,'()')

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Open the FITS file, with read-only access.

       call ftopen ( unit, filename, readwrite, blocksize, status )

100    continue

! Determine the number of keywords in the header.

       call ftghsp ( unit, nkeys, nspace, status )

       if (status .gt. 0) then
         call printerr ( status )
         return
       endif
         
       if (lsave) then
         open (11,file='.+readhead',status='unknown')     
       endif

! Read each 80-character keyword record, and print it out.

       ic = 0

       do i=1,nkeys
         
         call ftgrec ( unit, i, record, status )

         if (status .gt. 0) then
           call printerr ( status )
           return
         endif
         
         if (.not.lsave .or. action .eq. 'save+show') write (iotty,'(a81)') record
         
         if (lsave .and. index ( record, keyword1 ) .eq. 1) then

           ia = lastc ( record )

           if (index ( record, keyword1 ) .eq. 1) then 
             ik1 = index ( record, '= ' )
             ik2 = max ( index ( record(31:ia+1), '/' ), 1 ) + 30 - ik1 - 1
             cline = cline(1:ic) // record(1:ia1) // '=' // record(ik1+1:ik1+ik2) // ' '
             ic = lastc ( cline ) + 1
             if (keyword1 .eq. 'CDELT1' .or. keyword1 .eq. 'CD1_1') then 
               read (record(ik1+1:ik1+ik2),*) cdelt1
               if (abs ( cdelt1 ) .le. 1.0d-30) then 
                 if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   READHEAD: ERROR: Invalid keyword value for '
     &                                                      //keyword1//': '//record(ik1+1:ik1+ik2)
                 stop 11
               endif
             endif
           endif
         endif
       enddo

       do i=1,nkeys
         
         call ftgrec ( unit, i, record, status )

         if (lsave .and. index ( record, keyword2 ) .eq. 1) then

           ia = lastc ( record )

           if (index ( record, keyword2 ) .eq. 1 .and. keyword2 .ne. '') then 
             ik1 = index ( record, '= ' )
             ik2 = max ( index ( record(31:ia+1), '/' ), 1 ) + 30 - ik1 - 1
             cline = cline(1:ic) // record(1:ia2) // '=' // record(ik1+1:ik1+ik2) // ' '
             ic = lastc ( cline ) + 1
             record2 = record
             if (keyword2 .eq. 'CDELT2' .or. keyword2 .eq. 'CD2_2') then 
               read (record(ik1+1:ik1+ik2),*) cdelt2
               if (abs ( cdelt2 ) .le. 1.0d-30) then 
                 if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   READHEAD: ERROR: Invalid keyword value for '
     &                                                      //keyword2//': '//record(ik1+1:ik1+ik2)
                 stop 22
               endif
             endif
           endif
         endif
       enddo

       if (lsave) then                 

         if (index ( cline, keyword1 ) .eq. 1 .or. (ia2 .gt. 0 .and. index ( cline, keyword2 ) .gt. 0)) then

           if (index ( record2, keyword2 ) .eq. 0 .or. keyword2 .eq. '') then 

             if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Saving FITS keyword '//keyword1//' in file ''.+readhead''...'

           elseif (index ( record2, keyword2 ) .eq. 1 .and. keyword2 .ne. '') then 

             if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Saving FITS keywords '//keyword1//' '//keyword2
     &                                                 //' in file ''.+readhead''...'
           endif
           ic = lastc ( cline )
           write (11,'(a)') cline(1:ic)
           close (11)
         else
           if (ia1 .gt. 0 .and. index ( cline, keyword1 ) .eq. 0) then
             write (iotty,'(/a)') '   READHEAD: ERROR: Keyword '//keyword1//' not found.'
             stop 2
           endif
           if (ia2 .gt. 0 .and. index ( cline, keyword2 ) .eq. 0) then
             write (iotty,'(a)') '   READHEAD: ERROR: Keyword '//keyword2//' not found.'
             stop 3
           endif
         endif
       endif

! Print out an END record and a blank line to mark the header's end.

       if (status .eq. 0 .and. (.not. lsave .or. action .eq. 'save+show'))then
         write (iotty,'(a4)') ' END'
       endif

! Try moving to the next extension in the FITS file if it exists.

       call ftmrhd ( unit, 1, hdutype, status )

       if (status .eq. 0)then

! Success, so loop back and print out keywords in this extension.

         if (.not.lsave) write (iotty,'(a)') ' Press any key for extensions...'
         if (.not.lsave) read  (iotty,*)
         goto 100
       else if (status .eq. 107)then

! Hit end of file, so quit.

         if (.not. lsave .or. action .eq. 'save+show') write (iotty,'(a)') ' No more extensions.'
         status = 0
         call ftcmsg
       endif

! Close the file, free the unit number, and exit.

       call ftclos ( unit, status )
       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages.

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||