
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program separate
!__________________________________________________________________________________________________________________________________
!
! Extract all extensions of a FITS image in separate (numbered) files.
!
! May contain fragments from "The FITSIO Cookbook" v1.1 (April 1995) by William D Pence (HEASARC).
!
! This code is part of GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay  
!__________________________________________________________________________________________________________________________________
!                 
       implicit      none

       logical       lfname, lunix

       character*2   cnum    
       character*5   cfpptime
       character*6   cfitsversion, lverbose  
       character*7   clibname
       character*21  compda
       character*500 filename, outfile, arg1, arg2

       integer       i, status, inunit, outunit, readwrite, blocksize, morekeys, hdutype, fnlen, blank, firstb, lastc, folen, lb
     &             , ia1, ia2, iotty

       real          fitsvers
       
       parameter   ( iotty = 6 )
       
       external      firstb, lastc
!__________________________________________________________________________________________________________________________________
!
! Determine operating system

       call osystem ( lunix )

       call ftvers ( fitsvers )

       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       compda = __DATE__//' '//cfpptime

!  Get command line parameter (file name).

       call getarg ( 1, arg1 )
       call getarg ( 2, arg2 )
       
       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       fnlen = ia1

       lverbose = '-verb2'
       if (ia2 .gt. 0 .and. arg2(1:ia2-1) .eq. '-verb') lverbose = arg2(1:ia2)
       
       if (lverbose .eq. '-verb2') write (*,'(/a$)') ' SEPARATE = Extract All FITS Extensions = '//compda
       if (arg1(1:ia1) .eq. ':') stop
       if (lverbose .eq. '-verb2') write (*,'()')
       if (lverbose .eq. '-verb2') write (*,'( a)') ' Alexander Men`shchikov, SAp IRFU CEA Saclay, France.'
       if (lverbose .eq. '-verb2') write (*,'( a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                            //' by William D Pence.'
       if (fnlen .eq. 0) then
         write (*,'(/a/)') ' USAGE: separate <image> [-verb{0|1|2}]'
         stop 99
       endif

       if (lverbose .eq. '-verb1') write (*,'(/a)') ' SEPARATE: '//arg1(1:ia1)

!  Get the file name.

       blank = firstb ( arg1 )
       filename = arg1(1:blank-1)
       fnlen = lastc ( filename )

!  Check if the input file exists and open it.

       inquire ( file=filename(1:fnlen), exist=lfname )
       if (.not.lfname) then

!  Maybe the file name extension was omitted?

         inquire ( file=filename(1:fnlen)//'.fits', exist=lfname )

         if (.not.lfname) then
           write (*,'(/a)') '   SEPARATE: ERROR: File '''//filename(1:fnlen)//''' not found.'
           stop 99
         endif
         filename = arg1(1:blank-1)//'.fits'
         fnlen = lastc ( filename )
       endif

       status = 0

!  Get unused Logical Unit Numbers to use to open the FITS files.

       call ftgiou ( inunit, status )

!  Open the input FITS file, with readonly access.

       readwrite = 0

       call ftopen ( inunit, filename(1:fnlen), readwrite, blocksize, status )

       i = 1
       write (cnum,'(i2)') i
       if (cnum(1:1) .eq. ' ') cnum(1:1) = '0'

!  Create new output file name.

       outfile = filename(1:fnlen-4)//cnum//'.fits'
       folen = lastc ( outfile )

!  Create the new empty FITS file with the standard block size.

       blocksize = 1
       morekeys  = 0

       call ftgiou ( outunit, status )

       if (status .gt. 0) then
         call printerr ( status )
         write (*,*) '   SEPARATE: stopped at point a'
         stop 99
       endif

       call deletefile ( outfile, status )

       if (status .gt. 0) then
         call printerr ( status )
         write (*,*) '   SEPARATE: stopped at point b'
         stop 99
       endif

       call ftinit ( outunit, outfile, blocksize, status )

       if (status .gt. 0) then
         call printerr ( status )
         write (*,*) '   SEPARATE: stopped at point c'
         stop 99
       endif

       if (lverbose .eq. '-verb2') write (*,'(/a,i2,a)') '   Extracting FITS extension', i,'...'

       call ftcopy ( inunit, outunit, morekeys, status )

       if (status .gt. 0) then
         call printerr ( status )
         write (*,*) '   SEPARATE: stopped at point d'
         stop 99
       endif

!  Close the FITS file.

       call ftclos ( outunit, status )

       if (status .gt. 0) then
         call printerr ( status )
         write (*,*) '   SEPARATE: stopped at point e'
         stop 99
       endif

!  Free the unit number.

       call ftfiou ( outunit, status )

       if (status .gt. 0) then
         call printerr ( status )
         write (*,*) '   at point a'
         stop 99
       endif

       status = 0

!  Loop over all extensions in the input file.

       do i=2,1000

!  Skip to another extension in the input file.

         call ftmrhd ( inunit, 1, hdutype, status )

         if (status .gt. 0) then
           if (status .eq. 107) then
             status = 0   !<-- 107 is normal, it just means we hit the end of file.
             if (i .eq. 2) then
               if (lverbose .eq. '-verb2') write (*,'(/a)') '   ERROR: SINGLE IMAGE FOUND: NO EXTENSIONS TO EXTRACT.'
               call deletefile ( outfile, status )
               if (status .gt. 0) then
                 call printerr ( status )
                 write (*,'(/a)') '   ERROR: Trouble deleting file '''//outfile(1:folen)//'''...'
                 stop 99
               endif
               stop 77
             endif
           endif
           if (status .eq. 252) status = 0   !<-- on some bad files, 252 appears, but at the very end of file.
           exit
         endif

         write (cnum,'(i2)') i
         if (cnum(1:1) .eq. ' ') cnum(1:1) = '0'

!  Create new output file name.

         outfile = filename(1:fnlen-4)//cnum//'.fits'
         folen = lastc ( outfile )

!  Get unused logical unit number.

         call ftgiou ( outunit, status )

!  Initialize new output file.

         call deletefile ( outfile, status )

         call ftinit ( outunit, outfile, blocksize, status )

!  Copy the extension from the input file to the output file.

         if (lverbose .eq. '-verb2') write (*,'(a,i2,a)') '   Extracting FITS extension', i,'...'

         call ftcopy ( inunit, outunit, morekeys, status )

!  Close the FITS file.

         call ftclos ( outunit, status )

!  Free the unit number.

         call ftfiou ( outunit, status )

       enddo

!  Check for any error, and if so print out error messages.

       if (status .gt. 0) call printerr ( status )
              
       if (lverbose .eq. '-verb2') write (*,'(/a)') ' Done.'

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine deletefile ( filename, status )
!__________________________________________________________________________________________________________________________________
!
!  A simple little routine to delete a FITS file
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       status, unit, blocksize
       character*(*) filename
!__________________________________________________________________________________________________________________________________
!
!  Simply return if status is greater than zero

       if (status .gt. 0) return

!  Get an unused Logical Unit Number to use to open the FITS file

       call ftgiou ( unit, status )

!  Try to open the file, to see if it exists

       call ftopen ( unit, filename, 1, blocksize, status )

       if (status .eq. 0) then

!  File was opened;  so now delete it

         call ftdelt ( unit, status )

       else if (status .eq. 103) then

!  File doesn't exist, so just reset status to zero and clear errors

          status = 0

          call ftcmsg ()

       else

!  There was some other error opening the file; delete the file anyway

         status = 0

         call ftcmsg ()
         call ftdelt ( unit, status )
       end if

!  Free the unit number for later reuse

       call ftfiou ( unit, status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||