
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program splitcube
!__________________________________________________________________________________________________________________________________
!
! Split three-dimensional data cube along one of its dimensions, producing a set of numbered two-dimensional FITS images.
!
! This code is part of GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!                 
       implicit      none

       logical       lfnama, lunix

       character*2   ck   
       character*5   cfpptime
       character*6   cfitsversion, lverbose  
       character*7   clibname
       character*8   ctime
       character*10  cdate
       character*21  compda
       character*80  object, creator, bunit, ctype1, ctype2, history
       character*128 header, footer(2)
       character*500 filenama, filesub, filename, arg1, arg2, arg3

       integer       i, j, k, nxmx, nymx, nzmx, islen, fnlen1, ndate, iotty, firstb, lastc, lastbs, ia1, blank, ia2, ia3
     &             , ibs1, nsaxis, lb, nx, ny, nz

       parameter   ( nxmx = 6500, nymx = nxmx, nzmx = 5, iotty = 6 )

       real          fitsvers 
       real*8        dx, dy, wave, xr, yr, rp1, rp2, ra, dec, beam, rot1, rot2, epoch, equinox, funmin, funmax, bzero, bscale
     &             , datamin, datamax, cd11, cd12, cd21, cd22
       real*8        fun(nxmx,nymx,nzmx), funs(nxmx,nymx)

       external      firstb, lastc, rfits, wfits, when
!__________________________________________________________________________________________________________________________________
!
! Determine operating system

       call osystem ( lunix )

       call ftvers ( fitsvers )

       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif

       clibname = 'CFITSIO'
       cfpptime = __TIME__
       compda = __DATE__//' '//cfpptime

! Get command line parameters (file names).

       call getarg ( 1, arg1 )
       call getarg ( 2, arg2 )
       call getarg ( 3, arg3 )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )

       lverbose = '-verb2'
       if (ia3 .gt. 0 .and. arg3(1:ia3-1) .eq. '-verb') lverbose = arg3(1:ia3)
       if (ia2 .gt. 0 .and. arg2(1:ia2-1) .eq. '-verb') lverbose = arg2(1:ia2)

       if (lverbose .eq. '-verb2') write (iotty,'(/a$)') ' SPLITCUBE = Split FITS Data Cube = '//compda
       if (arg1(1:ia1) .eq. ':') stop
       if (lverbose .eq. '-verb2') write (iotty,'()')
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Alexander Men`shchikov, SAp IRFU CEA Saclay, France.'
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                                //' by William D Pence.'

       if (ia1 .eq. 0) then
         write (iotty,'(/a)') ' USAGE: splitcube <datacube> [1|2|3] [-verb{0|1|2}]'
         stop 99
       endif

       filenama = arg1(1:ia1)
       fnlen1 = lastc ( filenama )
       ibs1 = lastbs ( filenama )
       
       nsaxis = 3
       if (ia2 .gt. 0 .and. arg2(1:ia2-1) .ne. '-verb') read (arg2(1:ia2),*) nsaxis
       if (ia2 .gt. 0 .and. arg2(1:ia2-1) .eq. '-verb') arg2 = ' '

       if (lverbose .eq. '-verb1') write (iotty,'(/a)') ' SPLITCUBE: '//arg1(1:ia1)//' '//arg2(1:ia2)

! Check if the input file exists and open it.

       inquire ( file=filenama(1:fnlen1), exist=lfnama )
       if (.not.lfnama) then

! Maybe the file name extension was omitted?

         inquire ( file=filenama(1:fnlen1)//'.fits', exist=lfnama )
         if (.not.lfnama) then
           write (iotty,'(/a)') '   SPLITCUBE: ERROR: File '''//filenama(1:fnlen1)//''' not found.'
           stop 1
         endif
         filename = filenama(1:fnlen1)//'.fits'
         filenama = filename
         fnlen1 = lastc ( filenama )
       endif

       do k=1,nzmx
         do j=1,nymx
           do i=1,nxmx
             fun(i,j,k) = 0.0d0
           enddo
         enddo
       enddo

       if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   Reading image '''//filenama(1:fnlen1)//'''...'

       call rfits ( nxmx, nymx, nzmx, nx, ny, nz, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec
     &            , filenama(1:fnlen1), header, footer, iotty, 0, creator, beam, funmin, funmax, blank, rot1, rot2, cd11, cd12
     &            , cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history, lverbose )

       if (nz .gt. 0) then
         if (lverbose .eq. '-verb2') write (iotty,'(a,i5,1x,''x'',1x,i5,1x,''x'',1x,i5,a)') '   Datacube dimensions: '
     &                                     , nx, ny, nz, ' pixels'
       else 
         if (lverbose .eq. '-verb2') write (iotty,'(/a,i5,1x,''x'',1x,i5,1x,''x'',1x,i5,a)') '   SPLITCUBE: ERROR: NOT A DATACUBE: '
     &                                     , nx, ny, nz, ' pixels'
         if (nx .gt. 0 .and. ny .gt. 0) then
           if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   SPLITCUBE: ERROR: SINGLE IMAGE FOUND: NO DATACUBE TO SPLIT.'
           stop 77
         endif
         stop 88
       endif

       if (lverbose .eq. '-verb2') write (iotty,'(a,i1,a)') '   Splitting datacube in separate images along the axis #'
     &                                   , nsaxis, '...'

! Get the current date and time.

       call when ( lunix, ctime, cdate, ndate, 4 )
       
       do k=1,nz
         write (ck,'(i2)') k
         if (ck(1:1) .eq. ' '  ) ck(1:1) = '0'
         filesub = filenama(1:fnlen1-5)//'.'//ck//'.fits'
         islen = lastc ( filesub )

         do j=1,ny
           do i=1,nx
             funs(i,j) = fun(i,j,k)
           enddo
         enddo
         
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Writing  image '''//filesub(1:islen)//'''...'

         header = filesub(1:islen)
       
         call wfits ( cfitsversion, nxmx, nymx, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, funs, dx, dy, object, xr, yr
     &              , filesub(1:islen), header, footer, cdate, ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22
     &              , epoch, equinox, bzero, bscale, wave, datamin, datamax, history )
       enddo

       if (lverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( nxmx, nymx, nzmx, nx, ny, nz, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname
     &            , header, footer, iotty, iolog, creator, beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch
     &            , equinox, bzero, bscale, wave, datamn, datamx, history, lverbose )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, anynull, myfile
       character*(*) ctype1, ctype2, creator, fname, header, footer(2), bunit, object, history, lverbose
       character*25  cdatetime
       character*80  footer1, footer2, comment
       integer       i, j, k, nx, ny, nz, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, lastc, iotty, iolog, maxdim
     &             , ilc1, ilc2, rw, nfound, blank, stat01, stat02, stat03, stat04, stat05, stat06, stat07, stat08, stat09
     &             , stat10, stat11, stat12, stat13, stat14, stat15, stat16, stat17, stat18, stat19, stat20, stat21, stat22
     &             , stat23, stat24, stat25, stat26, stat27, stat28, stat29, stat30, stat31, stat32, stat0, nan, nxmx, nymx, nzmx
       parameter   ( maxdim = 3 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, cd11, cd12, cd21, cd22
       real*8        fun(nxmx,nymx,nzmx)
       real*8        as2deg, datamx, datamn
       parameter   ( as2deg = 1.0d0 / 3600.0d0 )
       external      lastc, ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat01 = 0
       stat02 = 0
       stat03 = 0
       stat04 = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') '   RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') '   RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') '   RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') '   RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
       if (stat .eq. 0 .and. (comment(1:22) .eq. 'Alexander Men`shchikov' .or. comment(1:22) .eq. 'Alexander Men''shchikov')) then
         myfile = .true.
       else
         myfile = .false.
         stat = 0
         creator = 'SPLITCUBE'
       endif

       header = 'FITS Plot Header'

       if (myfile) then
         call ftgkys ( unit, 'HEADER' , header , comment, stat01 )
         call ftgkys ( unit, 'FOOTER1', footer1, comment, stat02 )
         call ftgkys ( unit, 'FOOTER2', footer2, comment, stat03 )
         ilc1 = lastc ( footer1 )
         ilc2 = lastc ( footer2 )
         if (ilc1 .ne. 0 .and. ilc2 .ne. 0) then
           footer(1) = footer1(1:ilc1)//footer2(1:ilc2)
         else
           footer(1) = 'Footer'
         endif
       else
         call ftgkys ( unit, 'OBJECT' , header, comment, stat04 )
         if (stat04 .ne. 0 .or. header(1:10) .eq. '          ') header = fname
         footer(1) = header
       endif

! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       stat08 = 0
       stat0 = stat01 + stat02 + stat03 + stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15
     &       + stat16 + stat17 + stat18 + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28
     &       + stat29 + stat30 + stat31

       if (lverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat01 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword HEADER absent or error'
         if (stat02 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER1 absent or error'
         if (stat03 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER2 absent or error'
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat29 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EPOCH absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = 'No object specified'

! Read the FITS file into the 2D array.
                                                        
       call ftg3dd ( unit, group, dble ( blank ), nxmx, nymx, naxes(1), naxes(2), naxes(3), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       nx = naxes(1)
       ny = naxes(2)
       nz = naxes(3)
       funmin = 1.0d30
       funmax =-1.0d30
       nan = 0
       do k=1,nz
         do j=1,ny
           do i=1,nx
             if (isnan ( fun(i,j,k) )) then  
               nan = nan + 1
               if (lverbose .ne. '-verb0') then
                 if (nan .lt. 100) then
                   write (iotty,'(a,i4,a,i4,a,i4,a)') '   RFITS: Element (', i,',', j,',',k, ') contains not a number (NaN)...'
                 elseif (nan .eq. 100) then
                   write (iotty,'(a)') '   RFITS: Many NaNs, no more info on the elements.'
                 endif
               endif
             else
               funmin = min ( funmin, fun(i,j,k) )
               funmax = max ( funmax, fun(i,j,k) )
             endif  
           enddo
         enddo
       enddo
       
       if (iotty .gt. 0) write (iotty,'(a,1x,2(1pe14.7,1x))') '   RFITS: Minmax values:', funmin, funmax
       if (iolog .gt. 0) write (iolog,'(a,1x,2(1pe14.7,1x))') '   RFITS: Minmax values:', funmin, funmax

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, nxmx, nymx, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname
     &            , header, footer, cdate, ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero
     &            , bscale, wave, datamn, datamx, history )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend
       character*(*) fname, ctype1, ctype2, object, creator, header, footer(2), ctime, cdate, bunit, cfitsversion, history
       character*19  cdatetime
       integer       i, j, nx, ny, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, naxes(2), iat, iflen, lastc, blank
     &             , nxmx, nymx
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, dfun, rot1, rot2, epoch, equinox, datamn
     &             , datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, cosrota, sinrota, rotangle, pi
       real*8        fun(nxmx,nymx)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      lastc
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 2-D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 2
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       datamn = 1.0d30
       datamx =-1.0d30
       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j) )) then  
             dfun = fun(i,j)
             if (dfun .lt. datamn ) datamn = dfun
             if (dfun .gt. datamx ) datamx = dfun
           endif
         enddo
       enddo

       bitpix = -32     !<-- single precision floating-point values

       cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

       if (cfitsversion .eq. '5.030') then
         call fdelete ( fname, status )
       endif

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Create the new empty FITS file.

       if (cfitsversion .eq. '5.030') then
         call ftinit ( unit, fname, blocksze, status )
       else
         call ftinit ( unit, '!'//fname, blocksze, status )
       endif

! Write the required header keywords.

       call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

       call ftpkys ( unit, 'CREATOR', creator,'Alexander Men`shchikov, SAp IRFU CEA Saclay', status )
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'HEADER' , header, 'title', status )

       iat = 0
       iat = index ( footer(1), ' |F' )
       if (iat .eq. 0) iat = index ( footer(1), ' mm ' )
       iflen = lastc ( footer(1) )

       if (iat .gt. 0) then
         call ftpkys ( unit, 'FOOTER1', footer(1)(1:iat-1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', footer(1)(iat:iflen), 'part 2', status )
       else
         call ftpkys ( unit, 'FOOTER1', footer(1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', ' ', 'part 2', status )
       endif
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )

! Write all parameters.

       dxdeg = dx * as2deg
       dydeg = dy * as2deg
       
! Correct if the input images had cd11 or cd22 set to zero.

       if (abs ( cd11 ) .lt. almostzero .or. abs ( cd11 ) .gt. dxdeg) then
         cd11 = -dxdeg
         cd12 = 0.0d0
       endif
       if (abs ( cd22 ) .lt. almostzero .or. abs ( cd22 ) .gt. dxdeg) then
         cd21 = 0.0d0
         cd22 = dydeg
       endif

       cosrota = - cd11 / dxdeg
       sinrota = - cd12 / dxdeg
       rotangle = acos ( cosrota ) * 180.0d0 / pi
       
       if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = 180.0d0 - rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = 180.0d0 + rotangle
       elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = - rotangle
       endif
       rot2 = rot1

       call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
       call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
       call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
       if  (blank .gt. 0)
     & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
       call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
       call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
       call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
       call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CDELT1' , -dxdeg, 13 , 'coordinate increment along axis', status )
       call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
       call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
       call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
       call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
       call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
       call ftpkyd ( unit, 'EPOCH'  , epoch  , 13, ' ', status )
       call ftpkyd ( unit, 'EQUINOX', equinox, 13, ' ', status )
       call ftphis ( unit, history, status )

       if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'image was convolved with the beam (arcsec)', status )

! Write the array to the FITS file.

       call ftp2dd ( unit, group, nxmx, nx, ny, fun, status )

! Close the file.

       call ftclos ( unit, status )

! Free the unit number.

       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||