
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program modfits
!__________________________________________________________________________________________________________________________________
!
! Modify (in many ways) a single FITS image and some keywords of its header.
!
! This code is part of GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, SAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!                 
       implicit      none

       integer       inmx, iotty, nextrmax
       parameter   ( iotty = 6, nextrmax = 10000000 )

       logical       lfnama, changed, lunix, l1value, l4values, ldebug, l4conn, l8conn, lcleanskeletons
                                           
       character*1   crop, head, answer
       character*2   caddp, caddpx, caddpy
       character*5   cfpptime
       character*6   cfitsversion, lverbose          
       character*7   clibname
       character*8   ctime
       character*10  ccfact, cbsize, cbarea, cpixel, cdate
       character*21  compda
       character*80  object, creator, ctype1, ctype2, cvalue, cvalue1, cvalue2, cvalue3, cvalue4, bunit, history
       character*128 action, param, header, footer(2), cline
       character*500 arg1, arg2, arg3, arg4, arg5, arg6, arg7, filename, outname, profname

       integer       i, j, k, isp1, fnlen1, ion, ndate, nx, ny, nx1, ny1, nx2, ny2, firstb, lastc, lasts, nxo, nyo, ic, jc, nconn
     &             , ia1, ia2, ia3, ia4, ia5, ia6, ia7, nbw, iact, ival, ipar, blank, icva, im1, ip1, jm1, jp1, lb, n, i1, i2, j1
     &             , ic1, ic2, ic3, icva1, icva2, icva3, icva4, nextr, irc, nrem, naddpixmax, nymin, nymax, nextrmx, j2, ii, jj
     &             , nxmno, nxmxo, nymno, nymxo, ipn, ix1, ix2, iy1, iy2, iu1, iu2, iv1, iv2, ixlength, iylength, iulength, ncon
     &             , ivlength, l, m, lengthmin, idirection, iwidth, nskw, numpix4, numpix8, lev, npts, ik, jk

       real          fitsvers
       real*8        dx, dy, sfact, crpix1, crpix2, crval1, crval2, value1, value2, value3, value4, conversionfactor, rcore
     &             , ra, dec, beam, funmin, funmax, const, fJypixel, sq_arcsecs_per_sterad, fnewmin, fnewmax, thresh, value, levfac
     &             , sq_arcsecs_per_beam_area, pi, crota1, crota2, epoch, equinox, bzero, bscale, datamin, datamax, funfit, minpix
     &             , almostzero, deriv1a, deriv1b, deriv1c, deriv1d, deriv1e, deriv1f, deriv1g, deriv1h, deriv2a, deriv2b, deriv2c
     &             , deriv2d, wave, factMJysr2Jybeam, factMJysr2mJybeam, factJypixel2Jybeam, factJypixel2mJybeam, background
     &             , pixels_per_beam_area, factJypixel2MJysr, boupixval, newpixval, zeros, beamsize, beamarea, rad, facpix, factpix
     &             , sparsity, goodtoremove, cd11, cd12, cd21, cd22, arg(10000), prof(10000), lowintensity, elongoff, directioncode
     &             , level0, level, xi1, xi2, yj1, yj2, tlen, cosa, sina, pt1
                                                                                                              
       integer, allocatable :: n1x(:), n2x(:), ntouching(:,:), nsx(:), nxmn(:), nxmx(:), nymn(:), nymx(:)     
       real*8 , allocatable :: funa(:,:), funb(:,:), iworker(:,:), nsegm(:,:), mxco(:), myco(:), deltx(:), delty(:), nonzero(:)
     &                       , afwhm(:), bfwhm(:), atheta(:), equivrad(:), momxco(:), momyco(:), elongation(:)

       parameter   ( sq_arcsecs_per_sterad = 3282.80635d0 * 3600.0d0**2, pi = 3.14159265358979d0, almostzero = 1.0d-30 )

       parameter   ( boupixval = 0.0d0, ldebug = .false. )

       external      firstb, lastc, rfits, wfits, when, lastbs
!__________________________________________________________________________________________________________________________________
!
! Determine operating system

       call osystem ( lunix )

       call ftvers ( fitsvers )

       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       compda = __DATE__//' '//cfpptime

! Get the current date and time.

       call when ( lunix, ctime, cdate, ndate, 4 )

! Get command line parameters (file names).

       call getarg ( 1, arg1 )
       call getarg ( 2, arg2 )
       call getarg ( 3, arg3 )
       call getarg ( 4, arg4 )
       call getarg ( 5, arg5 )
       call getarg ( 6, arg6 )
       call getarg ( 7, arg7 )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )
       ia6 = lastc ( arg6 )
       ia7 = lastc ( arg7 )
       
       lverbose = '-verb2'
       if (ia7 .gt. 0 .and. arg7(1:ia7-1) .eq. '-verb') lverbose = arg7(1:ia7)
       if (ia6 .gt. 0 .and. arg6(1:ia6-1) .eq. '-verb') lverbose = arg6(1:ia6)
       if (ia5 .gt. 0 .and. arg5(1:ia5-1) .eq. '-verb') lverbose = arg5(1:ia5)
       if (ia4 .gt. 0 .and. arg4(1:ia4-1) .eq. '-verb') lverbose = arg4(1:ia4)
       if (ia3 .gt. 0 .and. arg3(1:ia3-1) .eq. '-verb') lverbose = arg3(1:ia3)
       
       if (lverbose .eq. '-verb2') write (iotty,'(/a$)') ' MODFITS = Modify FITS Image & Header = '//compda
       if (arg1(1:ia1) .eq. ':') stop
       if (lverbose .eq. '-verb2') write (iotty,'()')
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Alexander Men`shchikov, SAp IRFU CEA Saclay, France.'
       if (lverbose .eq. '-verb2') write (iotty,'( a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                                //' by William D Pence.'
       action = ' '
       value = 0.0d0
       cvalue = ' '
       param = ' '
       icva = 0
       iact = 0
       ival = 0
       ipar = 1
       l1value = .true.
       l4values = .false.
       naddpixmax = 0
       
       if (ia4 .gt. 0 .and. (arg4(1:ia4) .ne. '-o' .and. arg3(1:ia3) .ne. '-o' .and. arg4(1:ia4-1) .ne. '-verb')) then
         action = arg1(1:ia1)
         param = arg3(1:ia3)//' '
         filename = arg4(1:ia4)
         if (action .ne. 'keyword' .and. action .ne. 'key') then
! Sometimes due to mistakes on command line, the value may have a minus before a negative number, so we'll take care of that.
           if (arg2(1:2) .ne. '--') then
             read (arg2(1:ia2),*,err=90) value
           else
             read (arg2(2:ia2),*,err=90) value
           endif   
           goto 91
  90       continue         
           write (iotty,'(/a)') '   MODFITS: ERROR: Trouble reading a value from the string (1): '//arg2(1:ia2)
           stop 99
  91       continue
         endif
         cvalue = arg2(1:ia2)//' '
         icva = lastc ( cvalue ) + 1
         ipar = ia3 + 1
         ival = ia2
         iact = ia1
       elseif (ia3 .gt. 0 .and. (arg3(1:ia3) .ne. '-o' .and. arg2(1:ia2) .ne. '-o' .and. arg3(1:ia3-1) .ne. '-verb')) then
         action = arg1(1:ia1)
         filename = arg3(1:ia3)
         param='n '
! Sometimes due to mistakes on command line, the value may have a minus before a negative number, so we'll take care of that.
         if (arg2(1:2) .eq. '--') arg2(1:2) = ' -'
         ic1 = index ( arg2(1:ia2), ',' )
         if (ic1 .eq. 0) then
           read (arg2(1:ia2),*,err=1) value
           l1value = .true.
           l4values = .false.
           cvalue = arg2(1:ia2)//' '
           icva = lastc ( cvalue ) + 1
           goto 3
   1       continue
             write (iotty,'(/a)') '   MODFITS: ERROR: Trouble reading a value from the string (2): '//arg2(1:ia2)
             stop 99
         else
           l1value = .false.
           l4values = .true.
           ic2 = index ( arg2(ic1+1:ia2), ',' ) + ic1
           ic3 = index ( arg2(ic2+1:ia2), ',' ) + ic2
           cvalue1 = arg2(1:ic1-1)//' '
           cvalue2 = arg2(ic1+1:ic2-1)//' '
           cvalue3 = arg2(ic2+1:ic3-1)//' '
           cvalue4 = arg2(ic3+1:ia2)//' '
           icva1 = lastc ( cvalue1 ) + 1
           icva2 = lastc ( cvalue2 ) + 1
           icva3 = lastc ( cvalue3 ) + 1
           icva4 = lastc ( cvalue4 ) + 1
           read (arg2(1:ic1-1),*,end=2,err=2) value1
           read (arg2(ic1+1:ic2-1),*,end=2,err=2) value2
           read (arg2(ic2+1:ic3-1),*,end=2,err=2) value3
           read (arg2(ic3+1:ia2),*,end=2,err=2) value4
           goto 3
   2       continue
             write (iotty,'(/a)') ' MODFITS: '//action(1:iact)//' '//cvalue1(1:icva1)//cvalue2(1:icva2)
     &                                        //cvalue3(1:icva3)//cvalue4(1:icva4)//param(1:ipar)//filename(1:fnlen1-5)
             write (iotty,'(/a)') ' MODFITS: ERROR: Trouble reading 4 values from the string: '//arg2(1:ia2)
             stop 99
         endif
   3     continue
         ipar = 2
         ival = ia2
         iact = ia1
       elseif (ia2 .gt. 0 .and. arg2(1:ia2) .ne. '-o') then  
         filename = arg2(1:ia2)
         action = arg1(1:ia1)
         iact = ia1
         if (index ( action(1:iact), '~' ) .gt. 0) then
           cvalue = arg2(1:ia2)//' '
           icva = lastc ( cvalue ) + 1
         endif
       elseif (ia1 .gt. 0) then
         if (index ( arg1(1:ia1), '~' ) .gt. 0) then
           action = arg1(1:ia1)
           iact = ia1
         else
           filename = arg1(1:ia1)
         endif
       else
         write (iotty,'(/a)') ' USAGE: modfits [ACTION [<val>] [<par>]] <image> [-o <outname>] [-verb{0|1|2}]'
         write (iotty,'(  )')   
         write (iotty,'(a )') ' This is a versatile utility that allows various modifications of a'  
         write (iotty,'(a )') ' single FITS image; <image> and <outname> are the input and output images.'  
         write (iotty,'(a )') ' The following list briefly documents its parameters:'  
         write (iotty,'(  )')   
         write (iotty,'(a )') ' ACTION [<val>] [<par>]'
         write (iotty,'(  )')   
         write (iotty,'(a )') ' abs[olute] ................... convert image to absolute values'
         write (iotty,'(a )') ' neg[ate] ..................... negate image (multiply by -1)'
         write (iotty,'(a )') ' nan[zero] .................... convert all NaNs to zero'
         write (iotty,'(a )') ' inv[ert] ..................... invert image (non-zero pixels)'
         write (iotty,'(a )') ' squ[are] ..................... convert image to squared'
         write (iotty,'(a )') ' sqr[oot][m] .................. convert image to square root'
         write (iotty,'(a )') ' log[10][m] ................... convert image to logarithms'
         write (iotty,'(a )') ' der[iv]1 ..................... convert image to 1st derivatives'
         write (iotty,'(a )') ' der[iv]2 ..................... convert image to 2nd derivatives'
         write (iotty,'(a )') ' seg[ment]4|8 ................. convert image to 4- or 8-connected masks'
         write (iotty,'(a )') ' ske[leton] <val> ............. convert image to skeletons <val> pixels wide'
         write (iotty,'(a )') ' con[nect] <val> .............. reconnect one-pixel breaks in skeletons'
         write (iotty,'(a )') ' edg[e] ....................... remove inner pixels of non-zero structures'
         write (iotty,'(a )') ' ann[ex] ...................... annex connected pixels by segmentation masks'
         write (iotty,'(a )') ' thi[cken] .................... thicken clean structures by annexing 1 pixel'
         write (iotty,'(a )') ' mer[ge] ...................... merge clean structures separated by 1 pixel'
         write (iotty,'(a )') ' bsr[core] .................... background-subtract a model with rcore'
         write (iotty,'(a )') ' add[const] <val> ............. add to image a constant value <val>'
         write (iotty,'(a )') ' mul[tiply] <val> ............. multiply image by a factor <val>'
         write (iotty,'(a )') ' div[ide]   <val> ............. divide image by a factor <val>'
         write (iotty,'(a )') ' max[imum]  <val> [y|n] ....... set maximum to <val> and zero pixels if "y"'
         write (iotty,'(a )') ' min[imum]  <val> [y|n] ....... set minimum to <val> and zero pixels if "y"'
         write (iotty,'(a )') ' bor[der] [-]<val> ............ modify all borders uniformly by <val> pixels'
         write (iotty,'(a )') ' bor[der] [-]<vleft>,[-]<vright>,[-]<vlower>,[-]<vupper> ... more flexible' 
         write (iotty,'(a )') ' pro[file] <x1>,<y1>,<x2>,<y2>  create image profile along the straight line' 
         write (iotty,'(a )') ' exp[and]   <val> ............. expand image at all borders by <val> pixels'
         write (iotty,'(a )') ' key[word]  <val> <par> ....... set a header <val> keyword to value <par>'
         write (iotty,'(a )') ' lay[er]    <val> ............. remove a "zero" layer of <val> pixels'
         write (iotty,'(a )') ' cle[an]4|8[sk] <val> <par> ... remove 4- or 8-connected clusters of pixels'
         write (iotty,'(a )') ' rec[on]4|8 <val> <par> ....... reconstruct filaments 4- or 8-connected'
         write (iotty,'(a )') ' MJy/sr~Jy/beam    [-]<val> ... convert: beamsize <val> or beamarea -<val>'
         write (iotty,'(a )') ' MJy/sr~mJy/beam   [-]<val> ... convert: beamsize <val> or beamarea -<val>'
         write (iotty,'(a )') ' Jy/beam~MJy/sr    [-]<val> ... convert: beamsize <val> or beamarea -<val>'
         write (iotty,'(a )') ' mJy/beam~MJy/sr   [-]<val> ... convert: beamsize <val> or beamarea -<val>'
         write (iotty,'(a )') ' Jy/pixel~Jy/beam  [-]<val> ... convert: beamsize <val> or beamarea -<val>'
         write (iotty,'(a )') ' Jy/pixel~mJy/beam [-]<val> ... convert: beamsize <val> or beamarea -<val>'
         write (iotty,'(a )') ' Jy/beam~Jy/pixel  [-]<val> ... convert: beamsize <val> or beamarea -<val>'
         write (iotty,'(a )') ' mJy/beam~Jy/pixel [-]<val> ... convert: beamsize <val> or beamarea -<val>'
         write (iotty,'(a )') ' Jy/pixel~MJy/sr .............. convert the units (first to second)'
         write (iotty,'(a )') ' MJy/sr~Jy/pixel .............. convert the units (first to second)'
         write (iotty,'(  )')   
         write (iotty,'(a )') ' FITS keywords recognized: HEADER CREATOR OBJECT CTYPE1 CTYPE2 CRPIX1 CRPIX2'
         write (iotty,'(a )') ' CRVAL1 CRVAL2 CROTA1 CROTA2 CDELT1 CDELT2 CD1_1 CD1_2 CD2_1 CD2_2 RA DEC'
         write (iotty,'(a )') ' NAXIS1 NAXIS2 BZERO BSCALE BLANK DATAMIN DATAMAX EQUINOX WAVE HISTORY'
         write (iotty,'(  )')   
         write (iotty,'(a )') ' NOTE1: CDELT1 and CDELT2 must be specified in arcseconds.'  
         write (iotty,'(a )') ' NOTE2: beamsize must be specified in arcseconds.'  
         write (iotty,'(a )') ' NOTE3: beamarea must be specified in arcseconds^2.'  
         stop 99
       endif  
       if (lverbose .eq. '-verb2') write (iotty,'()')

       if (iact .ne. 0) then
         if (action(1:iact) .eq. 'keyword' .or. action(1:iact) .eq. 'key') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Change image header keyword '',a,''to '',a)') cvalue(1:icva)
     &                                                                                                 , param(1:ipar)
         endif
         if (action(1:iact) .eq. 'absolute' .or. action(1:iact) .eq. 'abs') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Take absolute value of the image'')')
         endif
         if (action(1:iact) .eq. 'negate' .or. action(1:iact) .eq. 'neg') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Negate image (multiplying by -1.0)'')')
         endif
         if (action(1:iact) .eq. 'nanzero' .or. action(1:iact) .eq. 'nan') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Convert all NaNs to zero'')')
         endif
         if (action(1:iact) .eq. 'addconst' .or. action(1:iact) .eq. 'add') then
           if (value .ge. 0.0d0) then
             if (lverbose .eq. '-verb2') write (iotty,'(''   Add a constant value of'',1pe11.4)') value
           else
             if (lverbose .eq. '-verb2') write (iotty,'(''   Subtract a constant value of'',1pe11.4)') abs ( value )
           endif
         endif
         if (action(1:iact) .eq. 'multiply' .or. action(1:iact) .eq. 'mul') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Multiply image by a factor of '',1pe11.4)') value
         endif
         if (index ( action(1:iact), '~' ) .gt. 0) then
           if (index ( action(1:iact), 'beam' ) .gt. 0) then
             if (lverbose .eq. '-verb2') write (iotty,'(''   Convert image physical units: '',a)') action(1:iact)
     &                                                                                       //' (beam='//cvalue(1:icva-1)//')'
             if (ia2 .gt. 0) read (arg2(1:ia2),*,err=46) value
             goto 47
  46         continue
             if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   MODFITS: ERROR: Trouble reading beam size from command line.'
             stop 99
  47         continue
           else
             if (lverbose .eq. '-verb2') write (iotty,'(''   Convert image physical units: '',a)') action(1:iact)
           endif
         endif
         if (action(1:iact) .eq. 'divide' .or. action(1:iact) .eq. 'div') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Divide image by a factor of '',1pe11.4)') value
         endif
         if (action(1:iact) .eq. 'square' .or. action(1:iact) .eq. 'squ') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Take square of the image'')')
         endif
         if (action(1:6) .eq. 'sqroot' .or. action(1:3) .eq. 'sqr') then
           if (action(1:iact) .eq. 'sqrootm' .or. action(1:iact) .eq. 'sqrm') then 
             if (lverbose .eq. '-verb2') write (iotty,'(''   Take square root (modified) of the image'')')
           else
             if (lverbose .eq. '-verb2') write (iotty,'(''   Take square root of the image'')')
           endif
         endif
         if (action(1:5) .eq. 'log10' .or. action(1:3) .eq. 'log') then
           if (action(1:iact) .eq. 'log10m' .or. action(1:iact) .eq. 'logm') then 
             if (lverbose .eq. '-verb2') write (iotty,'(''   Take log10 (modified) of the image'')')
           else
             if (lverbose .eq. '-verb2') write (iotty,'(''   Take log10 of the image'')')
           endif
         endif
         if (action(1:5) .eq. 'layer' .or. action(1:3) .eq. 'lay') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Remove ''''zero layer'''' of +/-'',1pe11.4)') value
         endif
         if (action(1:5) .eq. 'clean' .or. action(1:3) .eq. 'cle') then
           l4conn = .false.
           l8conn = .false.
           lcleanskeletons = .false.
           if (action(6:6) .eq. '4' .or. action(4:4) .eq. '4') l4conn = .true.
           if (action(6:6) .eq. '8' .or. action(4:4) .eq. '8') l8conn = .true.
           if (action(7:8) .eq. 'sk' .or. action(5:6) .eq. 'sk') lcleanskeletons = .true.
           if (.not.l4conn .and. .not.l8conn) then
             if (lverbose .eq. '-verb2') write (iotty,'(a)') '   MODFITS: ERROR: Invalid action: "'//action(1:iact)//'".'
             if (lverbose .eq. '-verb2') write (iotty,'(a)') '            Known actions: "clean4", "cle4", "clean8", "cle8",'
             if (lverbose .eq. '-verb2') write (iotty,'(a)') '            "clean4sk", "cle4sk", "clean8sk", "cle8sk"'
             stop 99
           endif         
           minpix = dble ( nint ( value ) )
           if (lverbose .eq. '-verb2') then
             if (l4conn) then
               write (iotty,'(''   Remove 4-connected clusters smaller than'',1pe10.3,'' pixels'')') minpix
               write (iotty,'(''   Remove 4-connected clusters with axes ratio smaller than '',a)') param(1:ipar)
             endif
             if (l8conn) then
               write (iotty,'(''   Remove 8-connected clusters smaller than'',1pe10.3,'' pixels'')') minpix
               write (iotty,'(''   Remove 8-connected clusters with axes ratio smaller than '',a)') param(1:ipar)
             endif
           endif  
           read (param(1:ipar),*,err=48) facpix   !<-- facpix is not used now
           goto 49
  48       continue
           if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   MODFITS: ERROR: Trouble reading FACPIX from command line.'
           stop 99
  49       continue
         endif
         if (action(1:5) .eq. 'recon' .or. action(1:3) .eq. 'rec') then
           l4conn = .false.
           l8conn = .false.
           if (action(6:6) .eq. '4' .or. action(4:4) .eq. '4') l4conn = .true.
           if (action(6:6) .eq. '8' .or. action(4:4) .eq. '8') l8conn = .true.
           if (.not.l4conn .and. .not.l8conn) then
             if (lverbose .eq. '-verb2') write (iotty,'(a)') '   MODFITS: ERROR: Invalid action: "'//action(1:iact)//'".'
             if (lverbose .eq. '-verb2') write (iotty,'(a)') '            Known actions: "recon4", "rec4", "recon8", "rec8".'
             stop 99
           endif         
           minpix = dble ( nint ( value ) )
           if (lverbose .eq. '-verb2') then
             if (l4conn) then
               write (iotty,'(''   Reconstruct 4-connected filaments:'',1pe10.3,'' pixels'')') minpix
             endif
             if (l8conn) then
               write (iotty,'(''   Reconstruct 8-connected filaments:'',1pe10.3,'' pixels'')') minpix
             endif
           endif  
           read (param(1:ipar),*,err=58) level0
           goto 59
  58       continue
           if (lverbose .eq. '-verb2') write (iotty,'(/a)') '   MODFITS: ERROR: Trouble reading LEVEL0 from command line.'
           stop 99
  59       continue
         endif
         if (action(1:5) .eq. 'annex' .or. action(1:3) .eq. 'ann') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Annex connected pixels by segmentation masks'')')
         endif
         if (action(1:7) .eq. 'thicken' .or. action(1:3) .eq. 'thi') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Thicken clean structures by annexing 1 pixel'')')
         endif
         if (action(1:5) .eq. 'merge' .or. action(1:3) .eq. 'mer') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Merge clean structures separated by 1 pixel'')')
         endif
         if (action(1:7) .eq. 'segment' .or. action(1:3) .eq. 'seg') then
           l4conn = .false.
           l8conn = .false.
           if (action(8:8) .eq. '4' .or. action(4:4) .eq. '4') l4conn = .true.
           if (action(8:8) .eq. '8' .or. action(4:4) .eq. '8') l8conn = .true.
           if (.not.l4conn .and. .not.l8conn) then
             if (lverbose .eq. '-verb2') write (iotty,'(a)') '   MODFITS: ERROR: Invalid action: "'//action(1:iact)//'".'
             if (lverbose .eq. '-verb2') write (iotty,'(a)') '            Known actions: "segment4", "seg4", "segment8", "seg8".'
             stop 99
           endif         
           if (lverbose .eq. '-verb2') then
             if (l4conn) write (iotty,'(''   Convert image to 4-connected segmentation masks'')')
             if (l8conn) write (iotty,'(''   Convert image to 8-connected segmentation masks'')')
           endif  
           if (lverbose .eq. '-verb2') write (iotty,'(''   Compute image of segmentation masks'')')
         endif
         if (action(1:6) .eq. 'deriv1' .or. action(1:4) .eq. 'der1') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Compute 1st derivative of the image'')')
         endif
         if (action(1:6) .eq. 'deriv2' .or. action(1:4) .eq. 'der2') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Compute 2nd derivative of the image'')')
         endif
         if (action(1:4) .eq. 'edge' .or. action(1:3) .eq. 'edg') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Remove inner pixels of all structures'')')
         endif
         if (action(1:8) .eq. 'skeleton' .or. action(1:3) .eq. 'ske') then
           nskw = max ( nint ( value ), 1 )
           if (mod ( nskw, 2 ) .eq. 0) nskw = nskw + 1
           if (nskw .eq. 1) then
             if (lverbose .eq. '-verb2') write (iotty,'(''   Convert image into skeletons '',i1,'' pixel wide'')') nskw
           else
             if (lverbose .eq. '-verb2') write (iotty,'(''   Convert image into skeletons '',i1,'' pixels wide'')') nskw
           endif
         endif
         if (action(1:7) .eq. 'connect' .or. action(1:3) .eq. 'con') then
           ncon = max ( nint ( value ), 1 )
           if (lverbose .eq. '-verb2') write (iotty,'(''   Reconnect skeletons separated by one-pixel breaks'')')
         endif
         if (action(1:iact) .eq. 'invert' .or. action(1:iact) .eq. 'inv') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Invert image'')')
         endif
         if (action(1:iact) .eq. 'bsrcore' .or. action(1:iact) .eq. 'bsrc') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Background-subtract model object'')')
         endif
         if (action(1:iact) .eq. 'minimum' .or. action(1:iact) .eq. 'min' .or. 
     &       action(1:iact) .eq. 'maximum' .or. action(1:iact) .eq. 'max') then
           if (lverbose .eq. '-verb2') write (iotty,'(''   Set new minimum and/or maximum values (param='',a1,'')'')') param
         endif
         if (action(1:iact) .eq. 'border' .or. action(1:iact) .eq. 'bor') then
           if (l1value) then
             if (nint(abs(value)) .lt. 10) then
               caddp='i1'
             else if (nint(abs(value)) .lt. 100) then
               caddp='i2'
             else if (nint(abs(value)) .lt. 1000) then
               caddp='i3'
             else if (nint(abs(value)) .lt. 10000) then
               caddp='i4'
             else
               caddp='i5'
             endif
             if (nint(value) .ge. 0.0d0) then
               if (lverbose .eq. '-verb2') write (iotty,'(''   Add border of '','//caddp//','' pixels around the image'')') 
     &                                           nint ( value )
             else
               if (lverbose .eq. '-verb2') write (iotty,'(''   Remove border of '','//caddp//','' pixels around the image'')') 
     &                                           nint ( abs ( value ) )
             endif
             naddpixmax = nint ( abs ( value ) )
           endif
           if (l4values) then
             if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Change image borders by '//arg2(1:ia2)//' pixels'
             naddpixmax = max ( nint ( value1 ) + nint ( value2 ), nint ( value3 ) + nint ( value4 ) )
           endif
         endif
         if (action(1:iact) .eq. 'profile' .or. action(1:iact) .eq. 'pro') then
           if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Create profile along the line (x1,y1,x2,y2): '//arg2(1:ia2)
         endif
         if (action(1:iact) .eq. 'expand' .or. action(1:iact) .eq. 'exp') then
           if (nint(value) .lt. 10) then
             caddp='i1'
           else if (nint(value) .lt. 100) then
             caddp='i2'
           else if (nint(value) .lt. 1000) then
             caddp='i3'
           else if (nint(value) .lt. 10000) then
             caddp='i4'
           else
             caddp='i5'
           endif
           if (lverbose .eq. '-verb2') write (iotty,'(''   Expand image boundary by '','//caddp//','' pixels'')') nint ( value )
           naddpixmax = nint ( value )
         endif
       endif
       
       fnlen1 = lastc ( filename )

! Check if the input file exists and open it.

       inquire ( file=filename(1:fnlen1), exist=lfnama )
       if (.not.lfnama) then

! Maybe the file name extension was omitted?

         filename = filename(1:fnlen1)//'.fits'
         fnlen1 = lastc ( filename )
         inquire ( file=filename(1:fnlen1), exist=lfnama )
       endif

       isp1 = lasts ( filename ) + 1

       outname = filename(isp1:fnlen1-5)//'.m.fits'
       if (arg2(1:2) .eq. '-o') outname = arg3
       if (arg3(1:2) .eq. '-o') outname = arg4
       if (arg4(1:2) .eq. '-o') outname = arg5
       if (arg5(1:2) .eq. '-o') outname = arg6
       ion = lastc ( outname )
       if (outname(max(1,ion-4):ion) .ne. '.fits') then
         outname = outname(1:ion)//'.fits'
         ion = lastc ( outname )
       endif

       if (l1value) then
         if (lverbose .eq. '-verb1') write (iotty,'(/a)') ' MODFITS: '//action(1:iact)//' '//cvalue(1:icva)//param(1:ipar)
     &                                              //filename(1:fnlen1-5)
       endif
       if (l4values) then
         if (lverbose .eq. '-verb1') write (iotty,'(/a)') ' MODFITS: '//action(1:iact)//' '//cvalue1(1:icva1)//cvalue2(1:icva2)
     &                                              //cvalue3(1:icva3)//cvalue4(1:icva4)//param(1:ipar)//filename(1:fnlen1-5)
       endif

       if (.not.lfnama .and. index ( action(1:iact), '~' ) .eq. 0) then
         write (iotty,'(/a)') '   MODFITS: ERROR: File '''//filename(1:fnlen1-5)//''' not found.'
         stop 1
       endif
       
       dx = 1.0d-10
       dy = 1.0d-10
       rcore = 0.0d0
                                   
! Determine numbers of pixels in the 1st FITS image.

       if (lfnama) then

         call getnumpix ( filename(1:fnlen1), nx, ny )

         inmx = max ( nx, ny ) + max ( 2 * naddpixmax, 0 )
       
         allocate ( funa(inmx,inmx), funb(inmx,inmx), stat=irc )
         
         if (irc .ne. 0) then
           write (iotty,'(/a)') '   MODFITS: ERROR: Trouble allocating memory (11).'
           stop 11
         endif

         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Reading image '''//filename(1:fnlen1)//'''...'

         call rfits ( inmx, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, funa, dx, dy, object, ra, dec
     &              , filename(1:fnlen1), header, footer, iotty, 0, creator, beam, funmin, funmax, blank, crota1, crota2
     &              , cd11, cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history, lverbose, rcore )
         
         if (lverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax
         changed = .false.
         nx1 = 0
         nx2 = 0
         ny1 = 0
         ny2 = 0
         sfact = 1.0d0
         const = 0.0d0
         do j=1,ny
           do i=1,nx
             funb(i,j) = funa(i,j)
           enddo
         enddo
       else
         if (index ( action(1:iact), '~' ) .gt. 0 .and. index ( action(1:iact), 'beam' ) .gt. 0 .and. ia2 .eq. 0) then
           write (iotty,'(/''   Enter  beam size (arcsec): '',$)')
           read (iotty,*) value
         else
           if (index ( action(1:iact), '~' ) .gt. 0 .and. index ( action(1:iact), 'pixel' ) .gt. 0) write (iotty,'()')
         endif
         if (index ( action(1:iact), '~' ) .gt. 0 .and. index ( action(1:iact), 'pixel' ) .gt. 0) then
           write (iotty,'( ''   Enter pixel size (arcsec): '',$)')
           read (iotty,*) dx
           dy = dx
         endif
       endif
       
       fJypixel = 1.0d6 * dx * dy / sq_arcsecs_per_sterad

       if (iact .eq. 0 .or. action(1:iact) .eq. 'keyword' .or. action(1:iact) .eq. 'key') then
  
         if (iact .eq. 0) then
           write (iotty,'(/''   Change image header keywords (y/N)? '',$)')
           read (iotty,'(a)') head
         endif
  
         if (iact .gt. 0 .or. head .eq. 'Y' .or. head .eq. 'y') then
           if (iact .eq. 0) then
             write (iotty,*)
             write (iotty,'(''      Note: To keep old values, just hit Enter.'')')
           endif
           
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  HEADER value: '',$)')
             write (iotty,'(a)') header(1:lastc(header))
             write (iotty,'( ''      New  HEADER value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               header = cline
               changed = .true.
             else
               header = ' '
             endif
           else
             if (cvalue .eq. 'HEADER') then
               header = param
               changed = .true.
               goto 77
             endif
           endif
           
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old CREATOR value: '',$)')
             write (iotty,'(a)') creator(1:lastc(creator))
             write (iotty,'( ''      New CREATOR value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               creator = cline
               changed = .true.
             else
               creator = ' '
             endif
           else
             if (cvalue .eq. 'CREATOR') then
               creator = param
               changed = .true.
               goto 77
             endif
           endif
           
           if (iact .eq. 0) then
             do i=1,80
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  OBJECT value: '',$)')
             write (iotty,'(a)') object(1:lastc(object))
             write (iotty,'( ''      New  OBJECT value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               object = cline
               changed = .true.
             else
               object = ' '
             endif
           else
             if (cvalue .eq. 'OBJECT') then
               object = param
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CTYPE1 value: '',$)')
             write (iotty,'(a)') ctype1(1:lastc(ctype1))
             write (iotty,'( ''      New  CTYPE1 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               ctype1 = cline
               changed = .true.
             else
               ctype1 = ' '
             endif
           else
             if (cvalue .eq. 'CTYPE1') then
               ctype1 = param
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CTYPE2 value: '',$)')
             write (iotty,'(a)') ctype2(1:lastc(ctype2))
             write (iotty,'( ''      New  CTYPE2 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               ctype2 = cline
               changed = .true.
             else
               ctype2 = ' '
             endif
           else
             if (cvalue .eq. 'CTYPE2') then
               ctype2 = param
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  BUNIT value: '',$)')
             write (iotty,'(a)') bunit(1:lastc(bunit))
             write (iotty,'( ''      New  BUNIT value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               bunit = cline
               changed = .true.
             else
               bunit = ' '
             endif
           else
             if (cvalue .eq. 'BUNIT') then
               bunit = param
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CRPIX1 value: '',f16.7)') crpix1
             write (iotty,'( ''      New  CRPIX1 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) crpix1
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CRPIX1') then
               read (param,*) crpix1
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CRPIX2 value: '',f16.7)') crpix2
             write (iotty,'( ''      New  CRPIX2 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) crpix2
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CRPIX2') then
               read (param,*) crpix2
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CRVAL1 value: '',f16.7,'' arcsec'')') crval1
             write (iotty,'( ''      New  CRVAL1 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) crval1
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CRVAL1') then
               read (param,*) crval1
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CRVAL2 value: '',f16.7,'' arcsec'')') crval2
             write (iotty,'( ''      New  CRVAL2 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) crval2
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CRVAL2') then
               read (param,*) crval2
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CROTA1 value: '',f16.7,'' degrees'')') crota1
             write (iotty,'( ''      New  CROTA1 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) crota1
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CROTA1') then
               read (param,*) crota1
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CROTA2 value: '',f16.7,'' degrees'')') crota2
             write (iotty,'( ''      New  CROTA2 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) crota2
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CROTA2') then
               read (param,*) crota2
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  RA  value: '',f16.7,'' degrees'')') ra
             write (iotty,'( ''      New  RA  value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) ra
               changed = .true.
             endif
           else
             if (cvalue .eq. 'RA') then
               read (param,*) ra
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  DEC value: '',f16.7,'' degrees'')') dec
             write (iotty,'( ''      New  DEC value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) dec
               changed = .true.
             endif
           else
             if (cvalue .eq. 'DEC') then
               read (param,*) dec
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  NAXIS1 value: '',i5,'' arcsec'')') nx
             write (iotty,'( ''      New  NAXIS1 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) nx
               changed = .true.
             endif
           else
             if (cvalue .eq. 'NAXIS1') then
               read (param,*) nx
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  NAXIS2 value: '',i5,'' arcsec'')') ny
             write (iotty,'( ''      New  NAXIS2 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) ny
               changed = .true.
             endif
           else
             if (cvalue .eq. 'NAXIS2') then
               read (param,*) ny
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CDELT1 value: '',f16.7,'' arcsec'')') dx
             write (iotty,'( ''      New  CDELT1 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) dx
               dx = abs ( dx )
               if (dx .lt. 0.03d0) dx = dx * 3600.0d0
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CDELT1') then
               read (param,*) dx
               dx = abs ( dx )
               if (dx .lt. 0.03d0) dx = dx * 3600.0d0
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CDELT2 value: '',f16.7,'' arcsec'')') dy
             write (iotty,'( ''      New  CDELT2 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) dy
               if (abs ( dy ) .lt. 0.03d0) dy = dy * 3600.0d0
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CDELT2') then
               read (param,*) dy
               if (abs ( dy ) .lt. 0.03d0) dy = dy * 3600.0d0
               changed = .true.
               goto 77
             endif
           endif

           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CD1_1 value: '',f16.7)') cd11
             write (iotty,'( ''      New  CD1_1 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) cd11
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CD1_1') then
               read (param,*) cd11
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CD1_2 value: '',f16.7)') cd12
             write (iotty,'( ''      New  CD1_2 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) cd12
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CD1_2') then
               read (param,*) cd12
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CD2_1 value: '',f16.7)') cd21
             write (iotty,'( ''      New  CD2_1 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) cd21
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CD2_1') then
               read (param,*) cd21
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  CD2_2 value: '',f16.7)') cd22
             write (iotty,'( ''      New  CD2_2 value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) cd22
               changed = .true.
             endif
           else
             if (cvalue .eq. 'CD2_2') then
               read (param,*) cd22
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  DATAMIN value: '',f16.7)') datamin
             write (iotty,'( ''      New  DATAMIN value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) datamin
               changed = .true.
             endif
           else
             if (cvalue .eq. 'DATAMIN') then
               read (param,*) datamin
               changed = .true.
               goto 77
             endif
           endif

           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  DATAMAX value: '',f16.7)') datamax
             write (iotty,'( ''      New  DATAMAX value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) datamax
               changed = .true.
             endif
           else
             if (cvalue .eq. 'DATAMAX') then
               read (param,*) datamax
               changed = .true.
               goto 77
             endif
           endif

           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  BZERO value: '',f16.7)') bzero
             write (iotty,'( ''      New  BZERO value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) bzero
               changed = .true.
             endif
           else
             if (cvalue .eq. 'BZERO') then
               read (param,*) bzero
               changed = .true.
               goto 77
             endif
           endif

           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  BSCALE value: '',f16.7)') bscale
             write (iotty,'( ''      New  BSCALE value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) bscale
               changed = .true.
             endif
           else
             if (cvalue .eq. 'BSCALE') then
               read (param,*) bscale
               changed = .true.
               goto 77
             endif
           endif

           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  BLANK value: '',f16.7)') blank
             write (iotty,'( ''      New  BLANK value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) blank
               changed = .true.
             endif
           else
             if (cvalue .eq. 'BLANK') then
               read (param,*) blank
               changed = .true.
               goto 77
             endif
           endif

           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  WAVE value: '',f16.7)') wave
             write (iotty,'( ''      New  WAVE value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) wave
               changed = .true.
             endif
           else
             if (cvalue .eq. 'WAVE') then
               read (param,*) wave
               changed = .true.
               goto 77
             endif
           endif

           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  EQUINOX value: '',f16.7)') equinox
             write (iotty,'( ''      New  EQUINOX value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) equinox
               changed = .true.
             endif
           else
             if (cvalue .eq. 'EQUINOX') then
               read (param,*) equinox
               changed = .true.
               goto 77
             endif
           endif
    
           if (iact .eq. 0) then
             do i=1,128
               cline(i:i) = ' '
             enddo
             write (iotty,'(/''      Old  HISTORY value: '',$)')
             write (iotty,'(a)') history(1:lastc(history))
             write (iotty,'( ''      New  HISTORY value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) then
               read (cline,*) history
               changed = .true.
             endif
           else
             if (cvalue .eq. 'HISTORY') then
               history = param
               changed = .true.
               goto 77
             endif
           endif

 77        continue
         endif
       endif

       if (iact .eq. 0 .or. action(1:iact) .eq. 'absolute' .or. action(1:iact) .eq. 'abs') then

         if (iact .eq. 0) then
           write (iotty,'(''   Take absolute value of the image (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
         
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           if (iact .eq. 0) then
             thresh = 0.0d0
             write (iotty,'(/''      Apply negative threshold (y/N)? '',$)')
             read (iotty,'(a)') answer
           
             if (answer .eq. 'Y' .or. answer .eq. 'y') then
               write (iotty,'(/''      Threshold value (< 0): '',$)')
               read (iotty,'(a)') cline
               if (lastc ( cline ) .gt. 0) read (cline,*) thresh
             endif
           else
             thresh = value
           endif
           
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .lt. thresh) funb(i,j) = - funa(i,j)
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

       if (iact .eq. 0 .or. action(1:iact) .eq. 'negate' .or. action(1:iact) .eq. 'neg') then

         if (iact .eq. 0) then
           write (iotty,'(''   Negate the image (multiply by -1) (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
        
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
       
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = - funa(i,j)
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif
       
       if (iact .eq. 0 .or. action(1:iact) .eq. 'nanzero' .or. action(1:iact) .eq. 'nan') then

         if (iact .eq. 0) then
           write (iotty,'(''   Convert all NaNs to zero (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
        
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (.not.isnan ( funa(i,j) )) then  
                 if (funa(i,j) .gt. funmax) funmax = funa(i,j)
                 if (funa(i,j) .lt. funmin) funmin = funa(i,j)
               endif  
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
       
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (isnan ( funa(i,j) )) then  
                 funb(i,j) = 0.0d0
               endif
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif
       
       if (iact .eq. 0 .or. action(1:iact) .eq. 'addconst' .or. action(1:iact) .eq. 'add') then
    
         if (iact .eq. 0) then
           write (iotty,'(''   Add (subtract) a constant (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
             
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
       
           if (iact .eq. 0) then
             write (iotty,'(/''      Constant to add (> 0) or subtract (< 0): '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) const
           else
             const = value
           endif
           
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j) + const
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif
       
       if (iact .eq. 0 .or. action(1:iact) .eq. 'multiply' .or. action(1:iact) .eq. 'mul') then

         if (iact .eq. 0) then
           write (iotty,'(''   Multiply the image by a factor (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
         
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           if (iact .eq. 0) then
             write (iotty,'(/''     Note: factor to rescale from MJy/sr to Jy/pixel:'',1pe11.4)') fJypixel
             write (iotty,'(/''     Scaling factor: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) sfact
           else
             sfact = value
           endif

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j) * sfact
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           changed = .true.
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

       if (iact .eq. 0 .or. index ( action(1:iact), '~' ) .gt. 0) then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Convert image physical units: '',a,$)') action(1:iact)
           read (iotty,'(a)') answer
         endif
         
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           if (lfnama) then
             funmin =  1.0d+30
             funmax = -1.0d+30
             do j=1,ny
               do i=1,nx
                 if (funa(i,j) .gt. funmax) funmax = funa(i,j)
                 if (funa(i,j) .lt. funmin) funmin = funa(i,j)
               enddo
             enddo

             if (iact .eq. 0) write (iotty,'()')
             if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
             if (lverbose .eq. '-verb2') write (iotty,*) funmax
             if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
             if (lverbose .eq. '-verb2') write (iotty,*) funmin
           else
             write (iotty,'()')
           endif

           beamsize = 0.0d0
           beamarea = 0.0d0
           if (iact .eq. 0) then
             write (iotty,'(/''     Beam size (FWHM, arcsec): '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) beamsize
           else
             if (value .gt. 0.0d0) then
               beamsize = value
             else
               beamarea = abs ( value )
             endif
           endif

           if (beamsize .gt. 1.0d-30) then
             sq_arcsecs_per_beam_area = pi * (beamsize / 2.0)**2 / log ( 2.0 )
             pixels_per_beam_area = pi * (beamsize / 2.0)**2 / (dx * dy) / log ( 2.0 )
           else
             sq_arcsecs_per_beam_area = beamarea
             pixels_per_beam_area = beamarea
           endif

           factMJysr2Jybeam = 1.0d6 * sq_arcsecs_per_beam_area / sq_arcsecs_per_sterad

           if (action(1:iact) .eq. 'MJy/sr~Jy/beam') conversionfactor = factMJysr2Jybeam
           if (action(1:iact) .eq. 'Jy/beam~MJy/sr') conversionfactor = 1.0d0 / factMJysr2Jybeam

           factMJysr2mJybeam = 1.0d9 * sq_arcsecs_per_beam_area / sq_arcsecs_per_sterad

           if (action(1:iact) .eq. 'MJy/sr~mJy/beam') conversionfactor = factMJysr2mJybeam
           if (action(1:iact) .eq. 'mJy/beam~MJy/sr') conversionfactor = 1.0d0 / factMJysr2mJybeam

           factJypixel2Jybeam = pixels_per_beam_area

           if (action(1:iact) .eq. 'Jy/pixel~Jy/beam') conversionfactor = factJypixel2Jybeam
           if (action(1:iact) .eq. 'Jy/beam~Jy/pixel') conversionfactor = 1.0d0 / factJypixel2Jybeam

           factJypixel2mJybeam = 1.0d3 * pixels_per_beam_area

           if (action(1:iact) .eq. 'Jy/pixel~mJy/beam') conversionfactor = factJypixel2mJybeam
           if (action(1:iact) .eq. 'mJy/beam~Jy/pixel') conversionfactor = 1.0d0 / factJypixel2mJybeam
                                                      
           factJypixel2MJysr = sq_arcsecs_per_sterad / (dx * dy) / 1.0d6

           if (action(1:iact) .eq. 'Jy/pixel~MJy/sr') conversionfactor = factJypixel2MJysr
           if (action(1:iact) .eq. 'MJy/sr~Jy/pixel') conversionfactor = 1.0d0 / factJypixel2MJysr
                                                      
           if (lverbose .eq. '-verb1') write (iotty,*)
           if (lverbose .ne. '-verb0') write (iotty,*) '  conversion factor:', conversionfactor

           write (ccfact,'(1pe10.3)') conversionfactor

           if (beamsize .gt. 1.0d-30) then
             write (cbsize,'(1pe10.3)') beamsize
             history = action(1:iact)//': beam size:'//cbsize//'; conversion factor:'//ccfact
           elseif (beamarea .gt. 1.0d-30) then
             write (cbarea,'(1pe10.3)') beamarea
             history = action(1:iact)//': beam area:'//cbarea//'; conversion factor:'//ccfact
           else
             write (cpixel,'(1pe10.3)') dx
             history = action(1:iact)//': pixel size:'//cpixel//'; conversion factor:'//ccfact
           endif
           
           if (.not.lfnama .and. index ( action(1:iact), '~' ) .gt. 0) then
             write (iotty,'()')
             stop
           endif

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j) * conversionfactor
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           changed = .true.
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'divide' .or. action(1:iact) .eq. 'div') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Divide the image by a factor (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
         
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           if (iact .eq. 0) then
             write (iotty,'(/''     Scaling factor: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) sfact
           else
             sfact = value
             if (sfact .eq. 0.0d0) then
               write (iotty,'(/a)') '   MODFITS: ERROR: Division by zero not allowed.'
               stop 99
             endif
           endif

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j) / sfact
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'square' .or. action(1:iact) .eq. 'squ') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Take square of the image (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
         
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j)**2
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'thicken' .or. action(1:iact) .eq. 'thi') then
             
         if (iact .eq. 0) then
           write (iotty,'(''   Thicken clean structures in the image (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
     
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
               funb(i,j) = funa(i,j)
             enddo
           enddo
             
           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           do j=1,ny
             jm1 = max (j - 1, 1)
             jp1 = min (j + 1, ny)
             do i=1,nx
               im1 = max (i - 1, 1)
               ip1 = min (i + 1, nx)
               if (funa(i,j) .lt. almostzero) then
                 if (funa(im1,j  ) .gt. almostzero) funb(i,j) = funa(im1,j  )
                 if (funa(ip1,j  ) .gt. almostzero) funb(i,j) = funa(ip1,j  )
                 if (funa(i  ,jm1) .gt. almostzero) funb(i,j) = funa(i  ,jm1)
                 if (funa(i  ,jp1) .gt. almostzero) funb(i,j) = funa(i  ,jp1)
                 if (funa(im1,jm1) .gt. almostzero) funb(i,j) = funa(im1,jm1)
                 if (funa(ip1,jp1) .gt. almostzero) funb(i,j) = funa(ip1,jp1)
                 if (funa(im1,jp1) .gt. almostzero) funb(i,j) = funa(im1,jp1)
                 if (funa(ip1,jp1) .gt. almostzero) funb(i,j) = funa(ip1,jp1)
               endif
             enddo
           enddo
             
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'merge' .or. action(1:iact) .eq. 'mer') then
       
         if (iact .eq. 0) then
           write (iotty,'(''   Merge structures separated by one pixel (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
     
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
               funb(i,j) = funa(i,j)
             enddo
           enddo
       
           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           do j=1,ny
             jm1 = max (j - 1, 1)
             jp1 = min (j + 1, ny)
             do i=1,nx
               im1 = max (i - 1, 1)
               ip1 = min (i + 1, nx)
               if (funa(i,j) .lt. almostzero) then
                 if (funa(im1,j) .gt. almostzero .and.
     &               funa(ip1,j) .gt. almostzero) funb(i,j) = (funa(im1,j) + funa(ip1,j)) / 2.0d0
                 if (funa(i,jm1) .gt. almostzero .and. 
     &               funa(i,jp1) .gt. almostzero) funb(i,j) = (funa(i,jm1) + funa(i,jp1)) / 2.0d0
               endif
             enddo
           enddo
       
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:6) .eq. 'sqroot' .or. action(1:3) .eq. 'sqr') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Take square root of the image (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
         
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           if (action(1:7) .eq. 'sqrootm' .or. action(1:4) .eq. 'sqrm') then 
                    
! The critical value of 0.32768 is found by requiring that the first derivative of the power function y=x**0.8 
! is equal to the first derivative of y=x, i.e. that the resulting first derivative is smooth.
! This is easily calculated as x = (1/0.8)**(1/(0.8-1)) = 0.8**5 = 0.32768. In order to have the function
! itself smooth, one needs to shift the power function down by 0.08192 = 0.32768**0.8 - 0.32768 = 0.8**4 - 0.8**5
!
! For function y=x**0.5 it's calculated as x = (1/0.5)**(1/(0.5-1)) = 0.5**2 = 0.25. In order to have the function
! itself smooth, one needs to shift the power function down by delta = 0.25**0.5 - 0.5**2
                    
             funfit = 0.5d0**2    !< = 0.25
             
             do j=1,ny
               do i=1,nx
                 if (funa(i,j) .ge. funfit) then
                   funb(i,j) = funa(i,j)**(0.5d0) - (0.25d0**(0.5d0) - funfit)
                 else
                   if (funa(i,j) .le. -funfit) then
                     funb(i,j) = -abs ( funa(i,j) )**(0.5d0) + (0.25d0**(0.5d0) - abs ( funfit ))
                   else
                     funb(i,j) = funa(i,j)
                   endif
                 endif
               enddo
             enddo
           else
             do j=1,ny
               do i=1,nx
                 if (funa(i,j) .ge. 0.0d0) then
                   funb(i,j) = sqrt ( funa(i,j) )
                 else
                   funb(i,j) = 0.0d0
                 endif
               enddo
             enddo
           endif
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:5) .eq. 'log10' .or. action(1:3) .eq. 'log') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Take log10 of the image (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
         
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           if (action(1:iact) .eq. 'log10m' .or. action(1:iact) .eq. 'logm') then 

             do j=1,ny
               do i=1,nx
                 if (funa(i,j) .ge. 1.0d0) then
                   funb(i,j) = log10 ( funa(i,j) ) + 1.0d0
                 else
                   if (funa(i,j) .le. -1.0d0) then
                     funb(i,j) = -log10 ( -funa(i,j) ) - 1.0d0
                   else
                     funb(i,j) = funa(i,j)
                   endif
                 endif
               enddo
             enddo
           else
             if (funmin .le. 1.0d-10) then 
               do j=1,ny
                 do i=1,nx
                   if (funa(i,j) .ge. 1.0d0) then
                     funb(i,j) = max ( log10 ( funa(i,j) ), 0.0d0 )
                   else
                     if (funa(i,j) .ge. 0.0d0) then
                       funb(i,j) = 0.0d0
                     endif
                   endif
                 enddo
               enddo
               do j=1,ny
                 do i=1,nx
                   if (funa(i,j) .le. -1.0d0) then
                     funb(i,j) = min ( -log10 ( -funa(i,j) ), 0.0d0 )
                   else
                     if (funa(i,j) .le. 0.0d0) then
                       funb(i,j) = 0.0d0
                     endif
                   endif
                 enddo
               enddo
             else
               do j=1,ny
                 do i=1,nx
                   funb(i,j) = log10 ( funa(i,j) )
                 enddo
               enddo
             endif
           endif
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamax = funmax
           datamin = funmin
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:5) .eq. 'layer' .or. action(1:3) .eq. 'lay') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Remove ''''zero layer'''' out of the image (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif

         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           if (iact .eq. 0) then
             write (iotty,'(/''     Limiting value: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) sfact
           else
             sfact = value
           endif
         
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           do j=1,ny
             do i=1,nx
               if (funa(i,j) .ge. sfact .or. funa(i,j) .le. -sfact) then 
                 funb(i,j) = funa(i,j)
               else
                 funb(i,j) = 0.0d0
               endif
             enddo
           enddo

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamax = funmax
           datamin = funmin
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'skeleton' .or. action(1:iact) .eq. 'ske') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Convert image to skeletons (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
         
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           if (iact .eq. 0) then
             write (iotty,'(/''     Width of the skeleton: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) nskw
           else
             nskw = max ( nint ( value ), 1 )
           endif

           nskw = min ( nskw, min ( nx, ny ) / 10 )

           if (mod ( nskw, 2 ) .eq. 0) nskw = nskw + 1

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
               funb(i,j) = 0.0d0
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           
           do j=1,ny
             jm1 = max ( j - 1, 1 )
             jp1 = min ( j + 1, ny )
             do i=1,nx
               if (funa(i,j) .gt. almostzero) then
                 im1 = max ( i - 1, 1 )
                 ip1 = min ( i + 1, nx )
                 numpix4 = 0
                 numpix8 = 0
                 if (funa(im1,j  ) .gt. almostzero) numpix4 = numpix4 + 1
                 if (funa(ip1,j  ) .gt. almostzero) numpix4 = numpix4 + 1
                 if (funa(i  ,jm1) .gt. almostzero) numpix4 = numpix4 + 1
                 if (funa(i  ,jp1) .gt. almostzero) numpix4 = numpix4 + 1
                 if (funa(im1,jm1) .gt. almostzero) numpix8 = numpix8 + 1
                 if (funa(ip1,jp1) .gt. almostzero) numpix8 = numpix8 + 1
                 if (funa(im1,jp1) .gt. almostzero) numpix8 = numpix8 + 1
                 if (funa(ip1,jm1) .gt. almostzero) numpix8 = numpix8 + 1

                 if (numpix4 .eq. 4 .or. numpix8 .ge. 3) then

                   lowintensity = 0.5d0 * funa(i,j)
                   
                   ix1 = 1
                   do ii=i,1,-1
                     if (funa(ii,j) .lt. lowintensity) then
                       ix1 = ii + 1
                       exit
                     endif
                   enddo
                   ix2 = nx
                   do ii=i,nx
                     if (funa(ii,j) .lt. lowintensity) then
                       ix2 = ii - 1
                       exit
                     endif
                   enddo
                   ixlength = ix2 - ix1 + 1

                   iy1 = 1
                   do jj=j,1,-1
                     if (funa(i,jj) .lt. lowintensity) then
                       iy1 = jj + 1
                       exit
                     endif
                   enddo
                   iy2 = ny
                   do jj=j,ny
                     if (funa(i,jj) .lt. lowintensity) then
                       iy2 = jj - 1
                       exit
                     endif
                   enddo
                   iylength = iy2 - iy1 + 1

                   iu1 = 1
                   do m=1,min(i-1,j-1)
                     if (funa(i-m,j-m) .lt. lowintensity) then
                       iu1 = m
                       exit
                     endif
                   enddo
                   iu2 = 1
                   do n=1,min(nx-i,ny-j)
                     if (funa(i+n,j+n) .lt. lowintensity) then
                       iu2 = n
                       exit
                     endif
                   enddo
                   iulength = iu2 + iu1 + 1

                   iv1 = 1
                   do m=1,min(i-1,ny-j)
                     if (funa(i-m,j+m) .lt. lowintensity) then
                       iv1 = m
                       exit
                     endif
                   enddo
                   iv2 = 1
                   do n=1,min(nx-i,j-1)
                     if (funa(i+n,j-n) .lt. lowintensity) then
                       iv2 = n
                       exit
                     endif
                   enddo
                   ivlength = iv2 + iv1 + 1

! Sort directions of the filament profiles in a sequence of increasing widths.

                   lengthmin = min ( ixlength, iylength, iulength, ivlength )
                   
                   if (lengthmin .eq. 0) then
                     write (*,*) '   MODFITS: SKELETON: ERROR: lengthmin = 0'
                     stop 99
                   endif
                   
                   if (lengthmin .eq. ixlength) then
                     idirection = 1
                     iwidth = ixlength
                   else if (lengthmin .eq. iylength) then
                     idirection = 2
                     iwidth = iylength
                   else if (lengthmin .eq. iulength) then
                     idirection = 3
                     iwidth = iulength
                   else if (lengthmin .eq. ivlength) then
                     idirection = 4
                     iwidth = ivlength
                   endif

! The non-zero values of the skeleton image will contain the direction, where the corresponding filament has smallest width. 
! The coded information is in the form (1.0d0 + directioncode), where directioncode = direction / 1000.0d0, where 
! direction = {1|2|3|4} is one of the four main directions. The coded direction value is transferred to the skeleton image,
! which is expected to always be normalized to 1.0 (plus the direction code). **NOT USED ANYMORE**

                   newpixval = 1.0d0
                   directioncode = 0.0d0
                     
                   if (idirection .eq. 1) then
                     if (funa(im1,j) .lt. funa(i,j) .and. funa(ip1,j) .lt. funa(i,j)) then
                       do l=max(j-nskw/2,1),min(j+nskw/2,ny)
                         do k=max(i-nskw/2,1),min(i+nskw/2,nx)
                           if (funb(k,l) .lt. almostzero) funb(k,l) = newpixval + directioncode
                         enddo
                       enddo
                       goto 10
                     endif
                   endif
                   if (idirection .eq. 2) then
                     if (funa(i,jm1) .lt. funa(i,j) .and. funa(i,jp1) .lt. funa(i,j)) then
                       do l=max(j-nskw/2,1),min(j+nskw/2,ny)
                         do k=max(i-nskw/2,1),min(i+nskw/2,nx)
                           if (funb(k,l) .lt. almostzero) funb(k,l) = newpixval + directioncode
                         enddo
                       enddo
                       goto 10
                     endif
                   endif
                   if (idirection .eq. 3) then
                     if (funa(im1,jm1) .lt. funa(i,j) .and. funa(ip1,jp1) .lt. funa(i,j) .and. 
     &                    funa(im1,jm1) .gt. almostzero .and. funa(ip1,jp1) .gt. almostzero) then
                       do l=max(j-nskw/2,1),min(j+nskw/2,ny)
                         do k=max(i-nskw/2,1),min(i+nskw/2,nx)
                           if (funb(k,l) .lt. almostzero) funb(k,l) = newpixval + directioncode
                         enddo
                       enddo
                       goto 10
                     endif
                   endif
                   if (idirection .eq. 4) then
                     if (funa(im1,jp1) .lt. funa(i,j) .and. funa(ip1,jm1) .lt. funa(i,j) .and.
     &                    funa(im1,jp1) .gt. almostzero .and. funa(ip1,jm1) .gt. almostzero) then
                       do l=max(j-nskw/2,1),min(j+nskw/2,ny)
                         do k=max(i-nskw/2,1),min(i+nskw/2,nx)
                           if (funb(k,l) .lt. almostzero) funb(k,l) = newpixval + directioncode
                         enddo
                       enddo
                       goto 10
                     endif
                   endif
  10               continue
                 endif
               endif
             enddo
           enddo

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax

           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:5) .eq. 'clean' .or. action(1:3) .eq. 'cle') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Remove small clusters of pixels (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif

         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then

           allocate ( iworker(nx,ny), nsegm(nx,ny), n1x(ny), n2x(ny), mxco(1), myco(1), deltx(1), delty(1), ntouching(1,1)
     &              , nsx(1), nxmn(nextrmax), nxmx(nextrmax), nymn(nextrmax), nymx(nextrmax), stat=irc )

           if (irc .ne. 0) then
             write (iotty,'(/a)') '   MODFITS: ERROR: Trouble allocating memory (12).'
             stop 12
           endif

           if (iact .eq. 0) then
             write (iotty,'(/''     Smallest number of pixels: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) minpix
             write (iotty,'(/''     Control parameter: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) facpix
           endif
         
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             n1x(j) = 1
             n2x(j) = nx
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo
           
           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           if (funmin .eq. funmax .and. funmax .lt. almostzero) then
             changed = .false.
             goto 5555
           endif
           
           nymin = 1
           nymax = ny
           do j=1,ny
             do i=1,nx
               iworker(i,j) = funa(i,j) / funmax / 1.1d0
               funb(i,j) = iworker(i,j)
               nsegm(i,j) = 0.0d0
             enddo
           enddo

           if (l8conn) then
             do j=1,ny
               jm1 = max ( j - 1, 1 )
               jp1 = min ( j + 1, ny )
               do i=1,nx
                 if (iworker(i,j) .gt. almostzero) then
                   im1 = max ( i - 1, 1 )
                   ip1 = min ( i + 1, nx )
                   if (iworker(im1,jm1) .gt. almostzero .and.
     &                 iworker(im1,j) .lt. almostzero .and. iworker(i,jm1) .lt. almostzero ) then
                     iworker(i,jm1) = iworker(i,j)
                     funb(i,jm1) = iworker(i,j)
                     funa(i,jm1) = funa(i,j)
                   endif
                   if (iworker(im1,jp1) .gt. almostzero .and.
     &                 iworker(im1,j) .lt. almostzero .and. iworker(i,jp1) .lt. almostzero ) then
                     iworker(im1,j) = iworker(i,j)
                     funb(im1,j) = iworker(i,j)
                     funa(im1,j) = funa(i,j)
                   endif
                   if (iworker(ip1,jp1) .gt. almostzero .and.
     &                 iworker(ip1,j) .lt. almostzero .and. iworker(i,jp1) .lt. almostzero ) then
                     iworker(i,jp1) = iworker(i,j)
                     funb(i,jp1) = iworker(i,j)
                     funa(i,jp1) = funa(i,j)
                   endif
                   if (iworker(ip1,jm1) .gt. almostzero .and.
     &                 iworker(ip1,j) .lt. almostzero .and. iworker(i,jm1) .lt. almostzero ) then
                     iworker(ip1,j) = iworker(i,j)
                     funb(ip1,j) = iworker(i,j)
                     funa(ip1,j) = funa(i,j)
                   endif
                 endif
               enddo
             enddo
           endif

! Remove some pixels in skeleton "thick" areas connecting too many pixels around.
! Fixed on 2014/01/06: moved this IF block here. Previously, it was placed above the previous IF block, 
! which "thickens" clusters and seems to interfere with the removal of pixels from "thick" areas.

           if (lcleanskeletons) then
             do j=1,ny
               jm1 = max ( j - 1, 1 )
               jp1 = min ( j + 1, ny )
               do i=1,nx
                 if (iworker(i,j) .gt. almostzero) then
                   im1 = max ( i - 1, 1 )
                   ip1 = min ( i + 1, nx )
                   numpix4 = 0
                   numpix8 = 0
                   if (iworker(im1,j  ) .gt. almostzero) numpix4 = numpix4 + 1
                   if (iworker(ip1,j  ) .gt. almostzero) numpix4 = numpix4 + 1
                   if (iworker(i  ,jm1) .gt. almostzero) numpix4 = numpix4 + 1
                   if (iworker(i  ,jp1) .gt. almostzero) numpix4 = numpix4 + 1
                   if (iworker(im1,jm1) .gt. almostzero) numpix8 = numpix8 + 1
                   if (iworker(ip1,jp1) .gt. almostzero) numpix8 = numpix8 + 1
                   if (iworker(im1,jp1) .gt. almostzero) numpix8 = numpix8 + 1
                   if (iworker(ip1,jm1) .gt. almostzero) numpix8 = numpix8 + 1
  
                   if (numpix4 .gt. 2 .or. numpix8 .gt. 2) then
                     iworker(i,j) = 0.0d0
                     funb(i,j) = 0.0d0
                     funa(i,j) = 0.0d0
                   endif
                 endif
               enddo
             enddo
           endif

           nextrmx = 1
           zeros = 1.0d-20
           mxco(1) = 1.0d0
           myco(1) = 1.0d0
           deltx(1) = 0.0d0
           delty(1) = 0.0d0
           ntouching(1,1) = 0
           nsx(1) = 0
           nextr = 0
           do j=1,ny
             do i=1,nx
               if (iworker(i,j) .gt. almostzero .and. nsegm(i,j) .le. almostzero) then
                 nextr = nextr + 1
                 if (nextr .gt. nextrmax) then
                   write (iotty,'(/a,i8)') '   MODFITS: ERROR: Number of isolated clusters reached the maximum of ', nextrmax
                   write (iotty,'(a    )') '            Increase the value and recompile MODFITS if you wish to continue.'
                   stop 89
                 endif
                 newpixval = dble ( nextr )
           
                 call fillarea ( .false., i, j, nextrmx, nextr, newpixval, 'positive', zeros, boupixval, nx, ny, n1x, n2x, nymin
     &                         , nymax, nxmn(nextr), nxmx(nextr), nymn(nextr), nymx(nextr), iworker, nsegm, nsegm, mxco, myco
     &                         , deltx, delty, ntouching, nsx )
               endif
             enddo
           enddo                   

           allocate ( momxco(nextr), momyco(nextr), afwhm(nextr), bfwhm(nextr), atheta(nextr), equivrad(nextr), elongation(nextr)
     &              , nonzero(nextr), stat=irc )
     
           if (irc .ne. 0) then
             write (iotty,'(/a)') '   MODFITS: ERROR: Trouble allocating memory (13).'
             stop 13
           endif

           do k=1,nextr
             nonzero(k) = 0.0d0
             do j=nymn(k),nymx(k)
               do i=nxmn(k),nxmx(k)
                 if (nint ( nsegm(i,j) ) .eq. k) then
                   nonzero(k) = nonzero(k) + 1.0d0
                 endif
               enddo
             enddo                   
           enddo

! Next reassignment of iworker from funb is necessary, as dimensions of funb is (inmx,inmx), not (nx,ny).

           do j=1,ny
             do i=1,nx
               iworker(i,j) = funb(i,j)
             enddo
           enddo
           
           call sizemeasure ( iotty, nextr, nxmn, nxmx, nymn, nymx, nx, ny, iworker, nsegm, dx, dy, momxco, momyco, afwhm, bfwhm
     &                      , atheta, equivrad, elongation, nonzero, almostzero )

           if (ldebug) then
             open ( 15, file=filename(1:fnlen1)//'.cleaning', status='unknown' )
             write (15,'(a)') '#'
             write (15,'(a)') '# Results of cleaning of connected clusters of pixels by MODFITS.'
             write (15,'(a)') '# The value GOOD = 1.00 has a meaning of GOOD FOR REMOVAL.'
             write (15,'(a)') '# The values of AFHM01 and BFWHM01 were multiplied by 2.'
             write (15,'(a)') '#'
             write (15,'(a,2(1pe11.3))') '# Input parameters (minpix, facpix):', minpix, facpix  !<-- facpix is not used now
             write (15,'(a)') '#'
             write (15,'(a)') '#      N    XCO_P    YCO_P  GOOD  AFWH01x2   BFWH01x2   THEP01     ELONGATIO  SPARSITY   FACTOR'
     &                      //'     NONZERO    FACMINPIX  FXP_BEST01 FXP_ERRO01 FXT_BEST01 FXT_ERRO01 SIG_MONO01 SN_RATIO01'
             write (15,'(a)') '#'
           endif
           
! If ellipse area >> pixel area (sparsity >> 1), this probably means that the filament is spiral-like or porous otherwise.

           nrem = 0
           do k=1,nextr
             if (facpix .gt. almostzero) then
               sparsity = afwhm(k) * bfwhm(k) / equivrad(k)**2
               elongoff = exp ( 20.0d0 * (1.2d0 - elongation(k)) ) + 1.0d0
               factpix = elongoff / max ( elongation(k), 1.0d0 ) / max ( sparsity, 1.0d0 )
             else
               factpix = 1.0d0
             endif

             goodtoremove = 0.0d0

             if (nonzero(k) .lt. factpix * minpix) then
               nrem = nrem + 1
               goodtoremove = 1.0d0
               do j=nymn(k),nymx(k)
                 do i=nxmn(k),nxmx(k)
                   if (nint ( nsegm(i,j) ) .eq. k) funa(i,j) = 0.0d0
                 enddo
               enddo                   
             endif

             if (ldebug) then
               write (15,'(i8,2f9.1,f6.2,14(1pe11.3))') 
     &                   k, momxco(k), momyco(k), goodtoremove, afwhm(k)*2.0, bfwhm(k)*2.0, atheta(k), elongation(k), sparsity
     &                 , factpix, nonzero(k), factpix * minpix, 1.0d0, 0.0d0, 1.0d0, 0.0d0, 100.0d0, 100.0d0
             endif
           enddo
           
           if (ldebug) then
             close ( 15 )
           endif

           deallocate ( momxco, momyco, afwhm, bfwhm, atheta, equivrad, elongation, nonzero )

           deallocate ( iworker, nsegm, n1x, n2x, mxco, myco, deltx, delty, ntouching, nsx, nxmn, nxmx, nymn, nymx )

! When cleaning skeletons, additionally remove one pixel from groups of four 4-connected pixels
! to avoid problems in FMEASURE when it searches for the first pixel of a skeleton.

           if (lcleanskeletons) then
             do j=1,ny
               jm1 = max ( j - 1, 1 )
               jp1 = min ( j + 1, ny )
               do i=1,nx
                 if (funa(i,j) .gt. almostzero) then
                   im1 = max ( i - 1, 1 )
                   ip1 = min ( i + 1, nx )
                   if ((funa(i  ,jp1) .gt. almostzero .and. funa(ip1,j  ) .gt. almostzero .and.
     &                  funa(ip1,jp1) .gt. almostzero .and. funa(im1,j  ) .lt. almostzero .and.
     &                  funa(i  ,jm1) .lt. almostzero .and. funa(im1,jm1) .lt. almostzero .and.
     &                  funa(im1,jp1) .lt. almostzero .and. funa(ip1,jm1) .lt. almostzero) .or.
     &                 (funa(i  ,jm1) .gt. almostzero .and. funa(im1,j  ) .gt. almostzero .and.
     &                  funa(im1,jm1) .gt. almostzero .and. funa(ip1,j  ) .lt. almostzero .and.
     &                  funa(i  ,jp1) .lt. almostzero .and. funa(ip1,jp1) .lt. almostzero .and.
     &                  funa(im1,jp1) .lt. almostzero .and. funa(ip1,jm1) .lt. almostzero) .or.
     &                 (funa(i  ,jp1) .gt. almostzero .and. funa(im1,j  ) .gt. almostzero .and.
     &                  funa(im1,jp1) .gt. almostzero .and. funa(ip1,j  ) .lt. almostzero .and.
     &                  funa(i  ,jm1) .lt. almostzero .and. funa(ip1,jm1) .lt. almostzero .and.
     &                  funa(im1,jm1) .lt. almostzero .and. funa(ip1,jp1) .lt. almostzero) .or.
     &                 (funa(i  ,jm1) .gt. almostzero .and. funa(ip1,j  ) .gt. almostzero .and.
     &                  funa(im1,jm1) .gt. almostzero .and. funa(im1,j  ) .lt. almostzero .and.
     &                  funa(i  ,jp1) .lt. almostzero .and. funa(im1,jp1) .lt. almostzero .and.
     &                  funa(im1,jm1) .lt. almostzero .and. funa(ip1,jp1) .lt. almostzero)) then
                     funa(i,j) = 0.0d0
                   endif
                 endif
               enddo
             enddo
           endif

           if (lverbose .eq. '-verb2') then
             if (nrem .ge. 1000000) then
               write (iotty,'(''   Removed '',i7,'' connected clusters'')') nrem
             elseif (nrem .ge. 100000) then
               write (iotty,'(''   Removed '',i6,'' connected clusters'')') nrem
             elseif (nrem .ge. 10000) then
               write (iotty,'(''   Removed '',i5,'' connected clusters'')') nrem
             elseif (nrem .ge. 1000) then                           
               write (iotty,'(''   Removed '',i4,'' connected clusters'')') nrem
             elseif (nrem .ge. 100) then                            
               write (iotty,'(''   Removed '',i3,'' connected clusters'')') nrem
             elseif (nrem .ge. 10) then                             
               write (iotty,'(''   Removed '',i2,'' connected clusters'')') nrem
             else                                                     
               write (iotty,'(''   Removed '',i1,'' connected clusters'')') nrem
             endif
           endif

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j)
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamax = funmax
           datamin = funmin
                
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamin
           if (iact .eq. 0) write (iotty,'()')
 5555      continue           
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:7) .eq. 'segment' .or. action(1:3) .eq. 'seg') then

         if (iact .eq. 0) then
           write (iotty,'(''   Convert image into segmentation (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif

         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo
           
           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
    
           if (abs ( funmin ) .lt. almostzero .and. abs ( funmax ) .lt. almostzero) then
             changed = .false.
             goto 6666
           endif

           allocate ( iworker(nx,ny), nsegm(nx,ny), n1x(ny), n2x(ny), mxco(1), myco(1), deltx(1), delty(1), ntouching(1,1), nsx(1)
     &              , stat=irc )

           if (irc .ne. 0) then
             write (iotty,'(/a)') '   MODFITS: ERROR: Trouble allocating memory (12).'
             stop 12
           endif

           nymin = 1
           nymax = ny
           do j=1,ny
             n1x(j) = 1
             n2x(j) = nx
             do i=1,nx
               iworker(i,j) = funa(i,j) / funmax / 1.1d0
               nsegm(i,j) = 0.0d0
             enddo
           enddo
           
           if (l8conn) then
             do j=1,ny
               jm1 = max ( j - 1, 1 )
               jp1 = min ( j + 1, ny )
               do i=1,nx
                 if (iworker(i,j) .gt. almostzero) then
                   im1 = max ( i - 1, 1 )
                   ip1 = min ( i + 1, nx )
                   if (iworker(im1,jm1) .gt. almostzero .and.
     &                 iworker(im1,j) .lt. almostzero .and. iworker(i,jm1) .lt. almostzero) then
                     iworker(i,jm1) = iworker(i,j)
                     funa(i,jm1) = funa(i,j)
                   endif
                   if (iworker(im1,jp1) .gt. almostzero .and.
     &                 iworker(im1,j) .lt. almostzero .and. iworker(i,jp1) .lt. almostzero) then
                     iworker(im1,j) = iworker(i,j)
                     funa(im1,j) = funa(i,j)
                   endif
                   if (iworker(ip1,jp1) .gt. almostzero .and.
     &                 iworker(ip1,j) .lt. almostzero .and. iworker(i,jp1) .lt. almostzero) then
                     iworker(i,jp1) = iworker(i,j)
                     funa(i,jp1) = funa(i,j)
                   endif
                   if (iworker(ip1,jm1) .gt. almostzero .and.
     &                 iworker(ip1,j) .lt. almostzero .and. iworker(i,jm1) .lt. almostzero) then
                     iworker(ip1,j) = iworker(i,j)
                     funa(ip1,j) = funa(i,j)
                   endif
                 endif
               enddo
             enddo
           endif
    
           nextrmx = 1
           zeros = 1.0d-20
           mxco(1) = 1.0d0
           myco(1) = 1.0d0
           deltx(1) = 0.0d0
           delty(1) = 0.0d0
           ntouching(1,1) = 0
           nsx(1) = 0
           nextr = 0
           do j=1,ny
             do i=1,nx
               if (iworker(i,j) .gt. almostzero .and. nsegm(i,j) .le. almostzero) then
                 nextr = nextr + 1
                 newpixval = dble ( nextr )
    
                 call fillarea ( .false., i, j, nextrmx, nextr, newpixval, 'positive', zeros, boupixval, nx, ny, n1x, n2x, nymin
     &                         , nymax, nxmno, nxmxo, nymno, nymxo, iworker, nsegm, nsegm, mxco, myco, deltx, delty, ntouching
     &                         , nsx )
               endif
             enddo
           enddo                   

! The non-zero values of the segmentation image will contain the direction, where a filament has smallest width. 
! The coded information will be taken from the input skeleton value, which is expected to be (1.0d0 + directioncode), where
! directioncode = direction / 1000.0d0, where direction = {1|2|3|4} is one of the four main directions. The coded direction value
! is transferred without any change to the output segmentation image. If the input image is not a skeleton normalized to 1.0, 
! then no such direction coding will be done.

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = nsegm(i,j)
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamax = funmax
           datamin = funmin
           
           deallocate ( iworker, nsegm, n1x, n2x, mxco, myco, deltx, delty, ntouching, nsx )

           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamin
           if (iact .eq. 0) write (iotty,'()')
 6666      continue
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:5) .eq. 'erase' .or. action(1:3) .eq. 'era') then
       
         if (iact .eq. 0) then
           write (iotty,'(''   Erase pixels off skeletons (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
       
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
       
           if (iact .eq. 0) then
             write (iotty,'(/''     Number: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) ncon  !<-- not used...
           endif
           nconn = 0
  
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j)
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           do j=1,ny
             jm1 = max ( j - 1, 1 )
             jp1 = min ( j + 1, ny )
             do i=1,nx
               im1 = max ( i - 1, 1 )
               ip1 = min ( i + 1, nx )
 
               if (funa(i,j) .gt. almostzero) then
                 if (funa(im1,j) .gt. almostzero .and. funa(im1,jm1) .gt. almostzero .and. funa(im1,jp1) .gt. almostzero) then
                   nconn = nconn + 1
                   funa(i,j) = 0.0d0
                 else if (funa(ip1,j) .gt. almostzero .and. funa(ip1,jm1) .gt. almostzero .and. funa(ip1,jp1) .gt. almostzero) then
                   nconn = nconn + 1
                   funa(i,j) = 0.0d0
                 else if (funa(i,jm1) .gt. almostzero .and. funa(ip1,jm1) .gt. almostzero .and. funa(im1,jm1) .gt. almostzero) then
                   nconn = nconn + 1
                   funa(i,j) = 0.0d0
                 else if (funa(i,jp1) .gt. almostzero .and. funa(ip1,jp1) .gt. almostzero .and. funa(im1,jp1) .gt. almostzero) then
                   nconn = nconn + 1
                   funa(i,j) = 0.0d0
                 endif
               endif
             enddo
           enddo

           do j=1,ny
             jm1 = max ( j - 1, 1 )
             jp1 = min ( j + 1, ny )
             do i=1,nx
               im1 = max ( i - 1, 1 )
               ip1 = min ( i + 1, nx )
 
               if (funa(i,j) .gt. almostzero) then
                 if (funa(im1,jm1) .gt. almostzero .and. funa(i,jp1) .gt. almostzero .and. funa(ip1,j) .gt. almostzero) then
                   nconn = nconn + 1
                   funa(i,j) = 0.0d0
                 else if (funa(im1,jp1) .gt. almostzero .and. funa(i,jm1) .gt. almostzero .and. funa(ip1,j) .gt. almostzero) then
                   nconn = nconn + 1
                   funa(i,j) = 0.0d0
                 else if (funa(ip1,jp1) .gt. almostzero .and. funa(im1,j) .gt. almostzero .and. funa(i,jm1) .gt. almostzero) then
                   nconn = nconn + 1
                   funa(i,j) = 0.0d0
                 else if (funa(ip1,jm1) .gt. almostzero .and. funa(im1,j) .gt. almostzero .and. funa(i,jp1) .gt. almostzero) then
                   nconn = nconn + 1
                   funa(i,j) = 0.0d0
                 endif
               endif
             enddo
           enddo

           if (lverbose .eq. '-verb2') then
             if (nconn .ge. 10000) then
               write (iotty,'(''   Erased '',i5,'' pixels off skeletons'')') nconn
             elseif (nconn .ge. 1000) then                 
               write (iotty,'(''   Erased '',i4,'' pixels off skeletons'')') nconn
             elseif (nconn .ge. 100) then                  
               write (iotty,'(''   Erased '',i3,'' pixels off skeletons'')') nconn
             elseif (nconn .ge. 10) then                   
               write (iotty,'(''   Erased '',i2,'' pixels off skeletons'')') nconn
             else                                          
               write (iotty,'(''   Erased '',i1,'' pixels off skeletons'')') nconn
             endif
           endif

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j)
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamax = funmax
           datamin = funmin
         
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:7) .eq. 'connect' .or. action(1:3) .eq. 'con') then
       
         if (iact .eq. 0) then
           write (iotty,'(''   Reconnect skeletons separated by one-pixel breaks (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
       
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
       
           if (iact .eq. 0) then
             write (iotty,'(/''     Number of surrounding pixels to reconnect: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) ncon  !<-- not used...
           endif
           nconn = 0

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j)
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo
           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

! Reconnect some breaks in skeleton, if the breaks are just one pixel wide.

           do j=1,ny
             jm1 = max ( j - 1, 1 )
             jp1 = min ( j + 1, ny )
             do i=1,nx
               if (funb(i,j) .lt. almostzero) then
                 im1 = max ( i - 1, 1 )
                 ip1 = min ( i + 1, nx )
                 numpix4 = 0
                 numpix8 = 0
                 if (funb(im1,j  ) .gt. almostzero) numpix4 = numpix4 + 1
                 if (funb(ip1,j  ) .gt. almostzero) numpix4 = numpix4 + 1
                 if (funb(i  ,jm1) .gt. almostzero) numpix4 = numpix4 + 1
                 if (funb(i  ,jp1) .gt. almostzero) numpix4 = numpix4 + 1
                 if (funb(im1,jm1) .gt. almostzero) numpix8 = numpix8 + 1
                 if (funb(ip1,jp1) .gt. almostzero) numpix8 = numpix8 + 1
                 if (funb(im1,jp1) .gt. almostzero) numpix8 = numpix8 + 1
                 if (funb(ip1,jm1) .gt. almostzero) numpix8 = numpix8 + 1
                 if (numpix4 + numpix8 .eq. 2 .or.
     &               (numpix4 .eq. 2 .and. numpix8 .eq. 1) .or. (numpix4 .eq. 1 .and. numpix8 .eq. 2) .or.
     &               (numpix4 .eq. 2 .and. numpix8 .eq. 2) .or. (numpix4 .eq. 2 .and. numpix8 .eq. 2)) then
            
                   if (funb(im1,j) .gt. almostzero .and. 
     &                 ((funb(ip1,j) .gt. almostzero .and. nint ( funb(ip1,j) ) .ne. nint ( funb(im1,j) )) .or. 
     &                  (funb(ip1,jp1) .gt. almostzero .and. nint ( funb(ip1,jp1) ) .ne. nint ( funb(im1,j) )) .or.
     &                  (funb(ip1,jm1) .gt. almostzero .and. nint ( funb(ip1,jm1) ) .ne. nint ( funb(im1,j) )))) then 
                     nconn = nconn + 1
                     funa(i,j) = dble ( nint ( funb(im1,j) ) ) + dble ( 2 ) / 1000.0d0  !<-- define and code pixel direction
                   else if (funb(ip1,j) .gt. almostzero .and. 
     &                      ((funb(im1,j) .gt. almostzero .and. nint ( funb(im1,j) ) .ne. nint ( funb(ip1,j) )) .or. 
     &                       (funb(im1,jp1) .gt. almostzero .and. nint ( funb(im1,jp1) ) .ne. nint ( funb(ip1,j) )) .or. 
     &                       (funb(im1,jm1) .gt. almostzero .and. nint ( funb(im1,jm1) ) .ne. nint ( funb(ip1,j) )))) then
                     nconn = nconn + 1
                     funa(i,j) = dble ( nint ( funb(ip1,j) ) ) + dble ( 2 ) / 1000.0d0  !<-- define and code pixel direction
                   else if (funb(i,jm1) .gt. almostzero .and. 
     &                      ((funb(i,jp1) .gt. almostzero .and. nint ( funb(i,jp1) ) .ne. nint ( funb(i,jm1) )) .or. 
     &                       (funb(im1,jp1) .gt. almostzero .and. nint ( funb(im1,jp1) ) .ne. nint ( funb(i,jm1) )) .or. 
     &                       (funb(ip1,jp1) .gt. almostzero .and. nint ( funb(ip1,jp1) ) .ne. nint ( funb(i,jm1) )))) then
                     nconn = nconn + 1
                     funa(i,j) = dble ( nint ( funb(i,jm1) ) ) + dble ( 1 ) / 1000.0d0  !<-- define and code pixel direction
                   else if (funb(i,jp1) .gt. almostzero .and. 
     &                      ((funb(i,jm1) .gt. almostzero .and. nint ( funb(i,jm1) ) .ne. nint ( funb(i,jp1) )) .or. 
     &                       (funb(im1,jm1) .gt. almostzero .and. nint ( funb(im1,jm1) ) .ne. nint ( funb(i,jp1) )) .or. 
     &                       (funb(ip1,jm1) .gt. almostzero .and. nint ( funb(ip1,jm1) ) .ne. nint ( funb(i,jp1) )))) then
                     nconn = nconn + 1
                     funa(i,j) = dble ( nint ( funb(i,jp1) ) ) + dble ( 1 ) / 1000.0d0  !<-- define and code pixel direction
                   else if (funb(im1,jm1) .gt. almostzero .and. 
     &                      ((funb(ip1,jp1) .gt. almostzero .and. nint ( funb(ip1,jp1) ) .ne. nint ( funb(im1,jm1) )) .or. 
     &                       (funb(i,jp1) .gt. almostzero .and. nint ( funb(i,jp1) ) .ne. nint ( funb(im1,jm1) )) .or. 
     &                       (funb(ip1,j) .gt. almostzero .and. nint ( funb(ip1,j) ) .ne. nint ( funb(im1,jm1) )))) then
                     nconn = nconn + 1
                     funa(i,j) = dble ( nint ( funb(im1,jm1) ) ) + dble ( 4 ) / 1000.0d0  !<-- define and code pixel direction
                   else if (funb(ip1,jp1) .gt. almostzero .and. 
     &                      ((funb(im1,jm1) .gt. almostzero .and. nint ( funb(im1,jm1) ) .ne. nint ( funb(ip1,jp1) )) .or. 
     &                       (funb(i,jm1) .gt. almostzero .and. nint ( funb(i,jm1) ) .ne. nint ( funb(ip1,jp1) )) .or. 
     &                       (funb(im1,j) .gt. almostzero .and. nint ( funb(im1,j) ) .ne. nint ( funb(ip1,jp1) )))) then 
                     nconn = nconn + 1
                     funa(i,j) = dble ( nint ( funb(ip1,jp1) ) ) + dble ( 4 ) / 1000.0d0  !<-- define and code pixel direction
                   else if (funb(im1,jp1) .gt. almostzero .and. 
     &                      ((funb(ip1,jm1) .gt. almostzero .and. nint ( funb(ip1,jm1) ) .ne. nint ( funb(im1,jp1) )) .or. 
     &                       (funb(ip1,j) .gt. almostzero .and. nint ( funb(ip1,j) ) .ne. nint ( funb(im1,jp1) )) .or. 
     &                       (funb(i,jm1) .gt. almostzero .and. nint ( funb(i,jm1) ) .ne. nint ( funb(im1,jp1) )))) then
                     nconn = nconn + 1
                     funa(i,j) = dble ( nint ( funb(im1,jp1) ) ) + dble ( 3 ) / 1000.0d0  !<-- define and code pixel direction
                   else if (funb(ip1,jm1) .gt. almostzero .and. 
     &                      ((funb(im1,jp1) .gt. almostzero .and. nint ( funb(im1,jp1) ) .ne. nint ( funb(ip1,jm1) )) .or. 
     &                       (funb(im1,j) .gt. almostzero .and. nint ( funb(im1,j) ) .ne. nint ( funb(ip1,jm1) )) .or. 
     &                       (funb(i,jp1) .gt. almostzero .and. nint ( funb(i,jp1) ) .ne. nint ( funb(ip1,jm1) )))) then
                     nconn = nconn + 1
                     funa(i,j) = dble ( nint ( funb(ip1,jm1) ) ) + dble ( 3 ) / 1000.0d0  !<-- define and code pixel direction
                   endif
                 endif
               endif
             enddo
           enddo
           if (lverbose .eq. '-verb2') then
             if (nconn .ge. 10000) then
               write (iotty,'(''   Inserted '',i5,'' pixels reconnecting skeletons'')') nconn
             elseif (nconn .ge. 1000) then                           
               write (iotty,'(''   Inserted '',i4,'' pixels reconnecting skeletons'')') nconn
             elseif (nconn .ge. 100) then                            
               write (iotty,'(''   Inserted '',i3,'' pixels reconnecting skeletons'')') nconn
             elseif (nconn .ge. 10) then                             
               write (iotty,'(''   Inserted '',i2,'' pixels reconnecting skeletons'')') nconn
             else                                                     
               write (iotty,'(''   Inserted '',i1,'' pixels reconnecting skeletons'')') nconn
             endif
           endif
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. almostzero) then
                 funb(i,j) = funa(i,j) - dble ( nint ( funa(i,j) ) ) + 1.0d0
               else
                 funb(i,j) = 0.0d0
               endif
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamax = funmax
           datamin = funmin
  
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:5) .eq. 'recon' .or. action(1:3) .eq. 'rec') then
         if (iact .eq. 0) then
           write (iotty,'(''   Reconstruct filaments (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif

         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
         
           allocate ( iworker(nx,ny), nsegm(nx,ny), n1x(ny), n2x(ny), mxco(1), myco(1), deltx(1), delty(1), ntouching(1,1), nsx(1)
     &              , nxmn(nextrmax), nxmx(nextrmax), nymn(nextrmax), nymx(nextrmax), nonzero(nextrmax), stat=irc )
     
           if (irc .ne. 0) then
             write (iotty,'(/a)') '   MODFITS: ERROR: Trouble allocating memory (12a).'
             stop 12
           endif
           
           if (iact .eq. 0) then
             write (iotty,'(/''     Smallest number of pixels: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) minpix
             write (iotty,'(/''     Control parameter: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) level0
           endif
  
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             n1x(j) = 1
             n2x(j) = nx
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo
           nymin = 1
           nymax = ny
           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           nextrmx = 1
           zeros = 1.0d-20
           mxco(1) = 1.0d0
           myco(1) = 1.0d0
           deltx(1) = 0.0d0
           delty(1) = 0.0d0
           ntouching(1,1) = 0
           nsx(1) = 0

           levfac = 1.05d0
           level = abs ( level0 ) / levfac

           do lev=1,9999
           
             level = level * levfac

             do j=1,ny
               do i=1,nx
                 if (funa(i,j) .ge. level) then
                   iworker(i,j) = funa(i,j) / funmax / 1.1d0
                 else
                   iworker(i,j) = 0.0d0
                 endif
                 funb(i,j) = iworker(i,j)
                 nsegm(i,j) = 0.0d0
               enddo
             enddo
             if (l8conn) then
               do j=1,ny
                 jm1 = max ( j - 1, 1 )
                 jp1 = min ( j + 1, ny )
                 do i=1,nx
                   im1 = max ( i - 1, 1 )
                   ip1 = min ( i + 1, nx )
                   if (iworker(i,j) .gt. almostzero .and. iworker(im1,jm1) .gt. almostzero .and.
     &                 iworker(im1,j) .lt. almostzero .and. iworker(i,jm1) .lt. almostzero ) then
                     iworker(i,jm1) = iworker(i,j)
                     funb(i,jm1) = iworker(i,j)
                   endif
                   if (iworker(i,j) .gt. almostzero .and. iworker(im1,jp1) .gt. almostzero .and.
     &                 iworker(im1,j) .lt. almostzero .and. iworker(i,jp1) .lt. almostzero ) then
                     iworker(im1,j) = iworker(i,j)
                     funb(im1,j) = iworker(i,j)
                   endif
                   if (iworker(i,j) .gt. almostzero .and. iworker(ip1,jp1) .gt. almostzero .and.
     &                 iworker(ip1,j) .lt. almostzero .and. iworker(i,jp1) .lt. almostzero ) then
                     iworker(i,jp1) = iworker(i,j)
                     funb(i,jp1) = iworker(i,j)
                   endif
                   if (iworker(i,j) .gt. almostzero .and. iworker(ip1,jm1) .gt. almostzero .and.
     &                 iworker(ip1,j) .lt. almostzero .and. iworker(i,jm1) .lt. almostzero ) then
                     iworker(ip1,j) = iworker(i,j)
                     funb(ip1,j) = iworker(i,j)
                   endif
                 enddo
               enddo
             endif
             
             nextr = 0

             do j=1,ny
               do i=1,nx
                 if (iworker(i,j) .gt. almostzero .and. nsegm(i,j) .le. almostzero) then
                   nextr = nextr + 1
                   if (nextr .gt. nextrmax) then
                     write (iotty,'(/a,i8)') '   MODFITS: ERROR: Number of isolated clusters reached the maximum of ', nextrmax
                     write (iotty,'(a    )') '            Increase the value and recompile MODFITS if you wish to continue.'
                     stop 89
                   endif
                   newpixval = dble ( nextr )
    
                   call fillarea ( .false., i, j, nextrmx, nextr, newpixval, 'positive', zeros, boupixval, nx, ny, n1x, n2x, nymin
     &                           , nymax, nxmn(nextr), nxmx(nextr), nymn(nextr), nymx(nextr), iworker, nsegm, nsegm, mxco, myco
     &                           , deltx, delty, ntouching, nsx )
                 endif
               enddo
             enddo
             
             if (nextr .eq. 0) exit
           
             allocate ( momxco(nextr), momyco(nextr), afwhm(nextr), bfwhm(nextr), atheta(nextr), equivrad(nextr), elongation(nextr)
     &                , stat=irc )
     
             if (irc .ne. 0) then
               write (iotty,'(/a)') '   MODFITS: ERROR: Trouble allocating memory (13).'
               stop 13
             endif
    
             do k=1,nextr
               nonzero(k) = 0.0d0
               do j=nymn(k),nymx(k)
                 do i=nxmn(k),nxmx(k)
                   if (nint ( nsegm(i,j) ) .eq. k) then
                     nonzero(k) = nonzero(k) + 1.0d0
                   endif
                 enddo
               enddo                   
             enddo

! Next reassignment of iworker from funb is necessary, as dimensions of funb is (inmx,inmx), not (nx,ny).

             do j=1,ny
               do i=1,nx
                 iworker(i,j) = funb(i,j)
               enddo
             enddo

             call sizemeasure ( iotty, nextr, nxmn, nxmx, nymn, nymx, nx, ny, iworker, nsegm, dx, dy, momxco, momyco, afwhm, bfwhm
     &                        , atheta, equivrad, elongation, nonzero, almostzero )
     
! If ellipse area >> pixel area (sparsity >> 1), this probably means that the filament is spiral-like or porous otherwise.

             do k=1,nextr

               if (level0 .gt. almostzero) then
                 sparsity = afwhm(k) * bfwhm(k) / equivrad(k)**2
                 elongoff = exp ( 20.0d0 * (1.2d0 - elongation(k)) ) + 1.0d0
                 factpix = elongoff / max ( elongation(k), 1.0d0 ) / max ( sparsity, 1.0d0 )
               else
                 factpix = 1.0d0
               endif

               if (nonzero(k) .le. factpix * minpix) then
                 do j=nymn(k),nymx(k)
                   do i=nxmn(k),nxmx(k)
                     if (nint ( nsegm(i,j) ) .eq. k) funa(i,j) = level
                   enddo
                 enddo                   
               endif
             enddo
    
             deallocate ( momxco, momyco, afwhm, bfwhm, atheta, equivrad, elongation )

             do j=1,ny
               do i=1,nx
                 if (funa(i,j) .gt. level) goto 444
               enddo
             enddo
             exit
  444        continue             
           enddo
           
           deallocate ( iworker, nsegm, n1x, n2x, mxco, myco, deltx, delty, ntouching, nsx, nxmn, nxmx, nymn, nymx, nonzero )

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j)
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamax = funmax
           datamin = funmin
         
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:5) .eq. 'annex' .or. action(1:3) .eq. 'ann') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Annex connected pixels by segmentation masks (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif

         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo
           nextr = nint ( funmax )

           allocate ( iworker(nx,ny), nsegm(nx,ny), n1x(ny), n2x(ny), mxco(nextr), myco(nextr), deltx(nextr), delty(nextr)
     &              , ntouching(nextr,nextr), nsx(nextr), nxmn(nextr), nxmx(nextr), nymn(nextr), nymx(nextr), stat=irc )

           if (irc .ne. 0) then
             write (iotty,'(/a)') '   MODFITS: ERROR: Trouble allocating memory (12).'
             stop 12
           endif

           do k=1,nextr
             do j=1,ny
               do i=1,nx
                 if (abs ( funa(i,j) - dble ( k ) ) .lt. almostzero) then
                   mxco(k) = dble ( i )
                   myco(k) = dble ( j )
                   goto 111
                 endif
               enddo
             enddo
 111         continue
           enddo
           do k=1,nextr
             deltx(k) = 0.0d0
             delty(k) = 0.0d0
             nsx(k) = 0
             do n=1,nextr
               ntouching(k,n) = 0
             enddo
           enddo
           do j=1,ny
             n1x(j) = 1
             n2x(j) = nx
             do i=1,nx
               funb(i,j) = 0.0d0
               iworker(i,j) = min ( funa(i,j), 0.999d0 )
               nsegm(i,j) = 0.0d0
             enddo
           enddo
           nymin = 1
           nymax = ny
           zeros = 1.0d-20

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           do k=1,nextr
             newpixval = dble ( k )
             i = nint ( mxco(k) )
             j = nint ( myco(k) )
           
             call fillarea ( .false., i, j, nextr, nextr, newpixval, 'positive', zeros, boupixval, nx, ny, n1x, n2x, nymin, nymax
     &                     , nxmn(k), nxmx(k), nymn(k), nymx(k), iworker, nsegm, nsegm, mxco, myco, deltx, delty, ntouching, nsx )
           enddo
           
           do j=1,ny
             do i=1,nx
               funb(i,j) = nsegm(i,j)
             enddo
           enddo

           deallocate ( iworker, nsegm, n1x, n2x, mxco, myco, deltx, delty, ntouching, nsx, nxmn, nxmx, nymn, nymx )

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamax = funmax
           datamin = funmin
                
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:5) .eq. 'deriv' .or. action(1:3) .eq. 'der') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Compute derivative of the image (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif

         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
               funb(i,j) = 0.0d0
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           deriv1a = 0.0d0
           deriv1b = 0.0d0
           deriv1c = 0.0d0
           deriv1d = 0.0d0
           deriv1e = 0.0d0
           deriv1f = 0.0d0
           deriv1g = 0.0d0
           deriv1h = 0.0d0

           do j=1,ny
             jm1 = max (  1, j - 1 )
             jp1 = min ( ny, j + 1 )
             do i=1,nx
               im1 = max (  1, i - 1 )
               ip1 = min ( nx, i + 1 )
               deriv1a = (funa(i,j) - funa(im1,j)) / dx
               deriv1b = (funa(ip1,j) - funa(i,j)) / dx
               deriv1c = (funa(i,j) - funa(i,jm1)) / dy
               deriv1d = (funa(i,jp1) - funa(i,j)) / dy
               deriv1e = (funa(i,j) - funa(im1,jm1)) / sqrt ( dx**2 + dy**2 )
               deriv1f = (funa(ip1,jp1) - funa(i,j)) / sqrt ( dx**2 + dy**2 )
               deriv1g = (funa(i,j) - funa(im1,jp1)) / sqrt ( dx**2 + dy**2 )
               deriv1h = (funa(ip1,jm1) - funa(i,j)) / sqrt ( dx**2 + dy**2 )

               if (action(1:6) .eq. 'deriv1' .or. action(1:4) .eq. 'der1') then
 
                 funb(i,j) = (abs ( deriv1a ) + abs ( deriv1b ) + abs ( deriv1c ) + abs ( deriv1d )
     &                      + abs ( deriv1e ) + abs ( deriv1f ) + abs ( deriv1g ) + abs ( deriv1h )) / 8.0d0
               endif

               if (action(1:6) .eq. 'deriv2' .or. action(1:4) .eq. 'der2') then

                 deriv2a = (deriv1b - deriv1a) / dx**2
                 deriv2b = (deriv1d - deriv1c) / dy**2
                 deriv2c = (deriv1f - deriv1e) / (dx**2 + dy**2)
                 deriv2d = (deriv1h - deriv1g) / (dx**2 + dy**2)

                 funb(i,j) = -(deriv2a + deriv2b + deriv2c + deriv2d) / 4.0d0
               endif
             enddo
           enddo

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamax = funmax
           datamin = funmin
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) datamin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'invert' .or. action(1:iact) .eq. 'inv') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Invert the image (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
         
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (abs ( funa(i,j) ) .gt. 1.0d-30) then
                 funb(i,j) = 1.0d0 / funa(i,j)
                 if (funb(i,j) .gt. funmax) funmax = funb(i,j)
                 if (funb(i,j) .lt. funmin) funmin = funb(i,j)
               endif
             enddo
           enddo
           do j=1,ny
             do i=1,nx
               if (abs ( funa(i,j) ) .le. 1.0d-30) then
                 funb(i,j) = funmax
               endif
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'edge' .or. action(1:iact) .eq. 'edg') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Remove inner pixels of all structures (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
         
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j)
               if (funa(i,j) .gt. 1.0d-30) then
                 im1 = max ( i - 1, 1 )
                 jm1 = max ( j - 1, 1 )
                 ip1 = min ( i + 1, nx )
                 jp1 = min ( j + 1, ny )
                 if (funa(im1,j  ) .gt. 1.0d-30 .and. funa(ip1,j  ) .gt. 1.0d-30 .and.
     &               funa(i  ,jm1) .gt. 1.0d-30 .and. funa(i  ,jp1) .gt. 1.0d-30 .and.
     &               funa(im1,jm1) .gt. 1.0d-30 .and. funa(im1,jp1) .gt. 1.0d-30 .and.
     &               funa(ip1,jm1) .gt. 1.0d-30 .and. funa(ip1,jp1) .gt. 1.0d-30) then
                   funb(i,j) = 0.0d0
                 endif
                 if (funb(i,j) .gt. funmax) funmax = funb(i,j)
                 if (funb(i,j) .lt. funmin) funmin = funb(i,j)
               endif
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif
       
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'bsrcore' .or. action(1:iact) .eq. 'bsrc') then
 
         if (iact .eq. 0) then
           write (iotty,'(''   Background-subtract model outside rcore (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif
         
         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo

           if (rcore .lt. almostzero) then
             write (iotty,'(/a)') '   MODFITS: ERROR: Core radius is zero: check the FITS header.'
             stop 99
           endif

           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           ic = nx / 2 + 1
           jc = ny / 2 + 1
           do i=ic,nx
             rad = dble ( i - ic ) * dx
             if (rad .gt. rcore) then
               background = funa(i,jc)
               goto 50
             endif
           enddo
 50        continue
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               rad = sqrt ( (dble ( i - ic ) * dx)**2 + (dble ( j - jc ) * dy)**2 )
               if (rad .le. rcore) then
                 funb(i,j) = funa(i,j) - background
               else
                 funb(i,j) = 0.0d0
               endif
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           if (lverbose .eq. '-verb2') write (iotty,'(''     New maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     New minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'minimum' .or. action(1:iact) .eq. 'min' .or. 
     &                      action(1:iact) .eq. 'maximum' .or. action(1:iact) .eq. 'max') then

         if (iact .eq. 0) then
           write (iotty,'(''   New minimum and/or maximum values (y/N)? '',$)')
           read (iotty,'(a)') answer
         endif

         if (iact .gt. 0 .or. answer .eq. 'Y' .or. answer .eq. 'y') then
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funa(i,j) .gt. funmax) funmax = funa(i,j)
               if (funa(i,j) .lt. funmin) funmin = funa(i,j)
             enddo
           enddo
           fnewmax = funmax
           fnewmin = funmin
           
           if (iact .eq. 0) write (iotty,'()')
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old maximum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmax
           if (lverbose .eq. '-verb2') write (iotty,'(''     Old minimum: '',$)')
           if (lverbose .eq. '-verb2') write (iotty,*) funmin

           if (iact .eq. 0) then
             write (iotty,'( ''     New maximum: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) fnewmax
             write (iotty,'( ''     New minimum: '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) fnewmin
  
             if (fnewmin .ne. funmin) then
               funmin =  1.0d+30
               funmax = -1.0d+30
               do j=1,ny
                 do i=1,nx
                   funb(i,j) = max ( funa(i,j), fnewmin )
                   if ((answer .eq. 'Y' .or. answer .eq. 'y') .and. funb(i,j) .le. fnewmin) funb(i,j) = 0.0d0
                   if (funb(i,j) .gt. funmax) funmax = funb(i,j)
                   if (funb(i,j) .lt. funmin) funmin = funb(i,j)
                 enddo
               enddo
               changed = .true.
             endif
             if (fnewmax .ne. funmax) then
               funmin =  1.0d+30
               funmax = -1.0d+30
               do j=1,ny
                 do i=1,nx
                   funb(i,j) = min ( funa(i,j), fnewmax )
                   if (funb(i,j) .gt. funmax) funmax = funb(i,j)
                   if (funb(i,j) .lt. funmin) funmin = funb(i,j)
                 enddo
               enddo
               datamin = funmin
               datamax = funmax
               changed = .true.
             endif
                    
           elseif (action(1:iact) .eq. 'minimum' .or. action(1:iact) .eq. 'min') then

             fnewmin = value           
             do i=1,80
               cline(i:i) = ' '
             enddo
             answer = 'Y'
             if (iact .eq. 0) then
               write (iotty,'(''       Zero values below the new minimum (Y/n)? '',$)')
               read (iotty,'(a)') cline
               if (lastc ( cline ) .gt. 0) read (cline,'(a)') answer
             else
               read (param(1:ipar),*) answer
             endif
             funmin =  1.0d+30
             funmax = -1.0d+30
             do j=1,ny
               do i=1,nx
                 if (funa(i,j) .gt. funmax) funmax = funa(i,j)
                 if (funa(i,j) .lt. funmin) funmin = funa(i,j)
               enddo
             enddo
             if (fnewmin .lt. funmin .and. abs ( fnewmin - funmin ) .lt. 1.0d-5 * abs ( funmin )) fnewmin = funmin * 1.00001
             funmin =  1.0d+30
             funmax = -1.0d+30
             do j=1,ny
               do i=1,nx
                 funb(i,j) = max ( funa(i,j), fnewmin )
                 if ((answer .eq. 'Y' .or. answer .eq. 'y') .and. funb(i,j) .le. fnewmin) funb(i,j) = 0.0d0
                 if (funb(i,j) .gt. funmax) funmax = funb(i,j)
                 if (funb(i,j) .lt. funmin) funmin = funb(i,j)
               enddo
             enddo
             datamin = funmin
             datamax = funmax
             if (lverbose .eq. '-verb2') write (iotty,'( ''     New maximum: '',$)')
             if (lverbose .eq. '-verb2') write (iotty,*) funmax
             if (lverbose .eq. '-verb2') write (iotty,'( ''     New minimum: '',$)')
             if (lverbose .eq. '-verb2') write (iotty,*) funmin
             changed = .true.

           elseif (action(1:iact) .eq. 'maximum' .or. action(1:iact) .eq. 'max') then

             fnewmax = value
             do i=1,80
               cline(i:i) = ' '
             enddo
             answer = 'Y'
             if (iact .eq. 0) then
               write (iotty,'(''       Zero values above the new maximum (Y/n)? '',$)')
               read (iotty,'(a)') cline
               if (lastc ( cline ) .gt. 0) read (cline,'(a)') answer
             else
               read (param(1:ipar),*) answer
             endif
             funmin =  1.0d+30
             funmax = -1.0d+30
             do j=1,ny
               do i=1,nx
                 funb(i,j) = min ( funa(i,j), fnewmax )
                 if ((answer .eq. 'Y' .or. answer .eq. 'y') .and. funb(i,j) .ge. fnewmax) funb(i,j) = 0.0d0
                 if (funb(i,j) .gt. funmax) funmax = funb(i,j)
                 if (funb(i,j) .lt. funmin) funmin = funb(i,j)
               enddo
             enddo
             datamin = funmin
             datamax = funmax
             if (lverbose .eq. '-verb2') write (iotty,'( ''     New maximum: '',$)')
             if (lverbose .eq. '-verb2') write (iotty,*) funmax
             if (lverbose .eq. '-verb2') write (iotty,'( ''     New minimum: '',$)')
             if (lverbose .eq. '-verb2') write (iotty,*) funmin
             changed = .true.
           endif
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'border' .or. action(1:iact) .eq. 'bor') then

         if (iact .eq. 0) then
           write (iotty,'(''   Add (remove) border around the image (y/N)? '',$)')
           read (iotty,'(a)') crop
         endif
         
         if (iact .gt. 0 .or. crop .eq. 'Y' .or. crop .eq. 'y') then
           if (iact .eq. 0) then
             write (iotty,'()')
             write (iotty,'(''     Note: Positive values to add, negative to remove.'')')
             write (iotty,'(''           For non-uniform border, just hit Enter.'')')
             do i=1,80
               cline(i:i) = ' '
             enddo
           endif
           
           if (iact .eq. 0) then
             write (iotty,'(/''     Uniform border width (pixels): '',$)')
             read (iotty,'(a)') cline
           endif
           nxo = nx
           nyo = ny
           if (iact .gt. 0 .or. lastc ( cline ) .gt. 0) then 
             if (iact .eq. 0) then 
               read (cline,*) nx1
               nx2 = nx1
               ny1 = nx1
               ny2 = nx1
             else
               if (l1value) then
                 nx1 = nint ( value )
                 nx2 = nx1
                 ny1 = nx1
                 ny2 = nx1
               endif
               if (l4values) then
                 nx1 = nint ( value1 )
                 nx2 = nint ( value2 )
                 ny1 = nint ( value3 )
                 ny2 = nint ( value4 )
               endif
             endif
           else
             nx1 = 0
             write (iotty,'(/''     Left  border width (pixels): '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) nx1
             nx2 = 0
             write (iotty,'( ''     Right border width (pixels): '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) nx2
             ny1 = 0
             write (iotty,'( ''     Lower border width (pixels): '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) ny1
             ny2 = 0
             write (iotty,'( ''     Upper border width (pixels): '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) ny2
           endif
           
           do j=1,inmx
             do i=1,inmx
               funa(i,j) = 0.0d0
             enddo
           enddo

           nx = nx + nx1 + nx2
           ny = ny + ny1 + ny2
           
           if (nx1 .ge. 0 .and. ny1 .ge. 0) then
             do j=1,nyo
               do i=1,nxo
                 funa(i+nx1,j+ny1) = funb(i,j)
               enddo
             enddo
           endif
           if (nx1 .lt. 0 .and. ny1 .ge. 0) then
             do j=1,nyo
               do i=1,nxo+nx1
                 funa(i,j+ny1) = funb(i-nx1,j)
               enddo
             enddo
           endif
           if (nx1 .ge. 0 .and. ny1 .lt. 0) then
             do j=1,nyo+ny1
               do i=1,nxo
                 funa(i+nx1,j) = funb(i,j-ny1)
               enddo
             enddo
           endif
           if (nx1 .lt. 0 .and. ny1 .lt. 0) then
             do j=1,nyo+ny1
               do i=1,nxo+nx1
                 funa(i,j) = funb(i-nx1,j-ny1)
               enddo
             enddo
           endif
           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               funb(i,j) = funa(i,j)
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           crpix1 = crpix1 + dble ( nx1 )
           crpix2 = crpix2 + dble ( ny1 )
           changed = .true.
           
           if (nxo .lt. 10) then
             caddpx='i1'
           else if (nxo .lt. 100) then
             caddpx='i2'
           else if (nxo .lt. 1000) then
             caddpx='i3'
           else if (nxo .lt. 10000) then
             caddpx='i4'
           else
             caddpx='i5'
           endif
           if (nyo .lt. 10) then
             caddpy='i1'
           else if (nyo .lt. 100) then
             caddpy='i2'
           else if (nyo .lt. 1000) then
             caddpy='i3'
           else if (nyo .lt. 10000) then
             caddpy='i4'
           else
             caddpy='i5'
           endif
           if (lverbose .eq. '-verb2') then
             write (iotty,'(''   Old image dimensions: '','//caddpx//','' '','//caddpy//','' pixels'')') nxo, nyo
           endif
           if (nx .lt. 10) then
             caddpx='i1'
           else if (nx .lt. 100) then
             caddpx='i2'
           else if (nx .lt. 1000) then
             caddpx='i3'
           else if (nx .lt. 10000) then
             caddpx='i4'
           else
             caddpx='i5'
           endif
           if (ny .lt. 10) then
             caddpy='i1'
           else if (ny .lt. 100) then
             caddpy='i2'
           else if (ny .lt. 1000) then
             caddpy='i3'
           else if (ny .lt. 10000) then
             caddpy='i4'
           else
             caddpy='i5'
           endif
           if (lverbose .eq. '-verb2') then
             write (iotty,'(''   New image dimensions: '','//caddpx//','' '','//caddpy//','' pixels'')') nx, ny
             write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax
           endif
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'profile' .or. action(1:iact) .eq. 'pro') then

         if (iact .eq. 0) then
           write (iotty,'(''   Create image profile along straight line (y/N)? '',$)')
           read (iotty,'(a)') crop
         endif
         
         if (iact .gt. 0 .or. crop .eq. 'Y' .or. crop .eq. 'y') then
           if (iact .gt. 0) then 
             if (l4values) then
               xi1 = value1
               yj1 = value2
               xi2 = value3
               yj2 = value4
             endif
           else
             write (iotty,'(/''     Start coordinate (x1, pixels): '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) xi1
             write (iotty,'( ''     Start coordinate (y1, pixels): '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) yj1
             write (iotty,'( ''     End coordinate (x2, pixels): '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) xi2
             write (iotty,'( ''     End coordinate (y2, pixels): '',$)')
             read (iotty,'(a)') cline
             if (lastc ( cline ) .gt. 0) read (cline,*) yj2
           endif
           i1 = nint ( xi1 )
           j1 = nint ( yj1 )
           i2 = nint ( xi2 )
           j2 = nint ( yj2 )
           if (i1 .lt. 1 .or. i1 .gt. nx .or. j1 .lt. 1 .or. j1 .gt. ny .or. 
     &         i2 .lt. 1 .or. i2 .gt. nx .or. j2 .lt. 1 .or. j2 .gt. ny) then
             write (iotty,'(/a)') '   MODFITS: ERROR: Line coordinates outside of image dimensions.'
             stop 99
           endif

           tlen = sqrt ( (xi2 - xi1)**2 + (yj2 - yj1)**2 )
           cosa = (xi2 - xi1) / tlen
           sina = (yj2 - yj1) / tlen
           npts = nint ( tlen ) + 1
           pt1 = 0.0d0

           do k=1,npts
             ik = nint ( xi1 + cosa * dble ( k - 1 ) )
             jk = nint ( yj1 + sina * dble ( k - 1 ) )
             arg(k) = pt1 + dx * dble ( k - 1 )
             prof(k) = funa(ik,jk)
             funb(ik,jk) = funmax * 1.2d0
           enddo

           funmin =  1.0d+30
           funmax = -1.0d+30
           do j=1,ny
             do i=1,nx
               if (funb(i,j) .gt. funmax) funmax = funb(i,j)
               if (funb(i,j) .lt. funmin) funmin = funb(i,j)
             enddo
           enddo
           datamin = funmin
           datamax = funmax
           changed = .true.
           
           if (lverbose .eq. '-verb2') then
             profname = outname(1:ion)//'.prof.dat'
             ipn = lastc ( profname )
             write (iotty,'(a)') '   Saving profile in '''//profname(1:ipn)//'''...'
             open ( 15, file=profname(1:ipn), status='unknown' )
             write (15,'(a)') '# MODFITS = Modify FITS Image & Header = '//compda
             write (15,'(a)') '# Alexander Men`shchikov, SAp IRFU CEA Saclay, France.'
             write (15,'(a)') '# Using '//clibname(lb:7)//' library version'//cfitsversion//' by William D Pence.'
             write (15,'(a)') '#'
             write (15,'(a)') '# PROFILE EXTRACTED FROM THE IMAGE: '//filename(1:fnlen1)
             write (15,'(a)') '# ALONG A STRAIGHT LINE (X1,Y1,X2,Y2): '//arg2(1:ia2)//' (PIXELS)'
             write (15,'(a)') '#'
             write (15,'(a)') '# TABULATED QUANTITIES:'
             write (15,'(a)') '#'
             write (15,'(a)') '#   N ........ Distance in pixels'
             write (15,'(a)') '#   ARCSEC ... Distance in acseconds'
             write (15,'(a)') '#   VALUE .... Image value at the distance'
             write (15,'(a)') '#'
             write (15,'(a)') '#   N  ARCSEC    VALUE'
             write (15,'(a)') '#'
             do k=1,npts
               write (15,'(1x,i4,1x,1pe9.2,1x,1pe14.7)') k, arg(k), prof(k)
             enddo
             close ( 15 )
             write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax
           endif
           if (iact .eq. 0) write (iotty,'()')
         endif
       endif

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       if (iact .eq. 0 .or. action(1:iact) .eq. 'expand' .or. action(1:iact) .eq. 'exp') then
        
         if (iact .eq. 0) then
           write (iotty,'(''   Expand image around its boundary (y/N)? '',$)')
           read (iotty,'(a)') crop
         endif
         
         if (iact .gt. 0 .or. crop .eq. 'Y' .or. crop .eq. 'y') then
           do i=1,80
             cline(i:i) = ' '
           enddo
           if (iact .eq. 0) then
             write (iotty,'(/''     Uniform border width (pixels): '',$)')
             read (iotty,'(a)') cline
           endif
           if (iact .gt. 0 .or. lastc ( cline ) .gt. 0) then 
             if (iact .eq. 0) then
               read (cline,*) nbw
             else 
               nbw = nint ( value )
             endif                 
             
             if (nbw .gt. 0) then
               nx1 = nx + 2 * nbw
               ny1 = ny + 2 * nbw
               
               call expandit ( 'outflows', nx, ny, nx1, ny1, nbw, inmx, inmx, funb, funa )
           
               nx = nx1
               ny = ny1
               crpix1 = crpix1 + dble ( nbw )
               crpix2 = crpix2 + dble ( nbw )
             endif
             do j=1,ny
               do i=1,nx
                 funb(i,j) = funa(i,j)
               enddo
             enddo
             changed = .true.
           endif
         endif
       endif

       if (changed) then
         if (iact .eq. 0) write (iotty,'()')
         if (lverbose .eq. '-verb2') write (iotty,'(a)') '   Writing output image '''//outname(1:ion)//'''...'

         call wfits ( cfitsversion, inmx, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, funb, dx, dy, object
     &              , crval1, crval2, outname(1:ion), header, footer, cdate, ctime, creator, beam, blank, crota1, crota2, cd11
     &              , cd12, cd21, cd22, epoch, equinox, bzero, bscale, wave, datamin, datamax, history, action(1:iact) )
                                                                    
         deallocate ( funa, funb )
         
         if (lverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'
       else
         if (lverbose .eq. '-verb2') write (iotty,'(/a)') ' No changes, nothing to do.'
       endif

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine sizemeasure 
       
     &            ( iotty, nextr, nxmin, nxmax, nymin, nymax, nx, ny, image, segmn, dx, dy, momxco, momyco, afwhm, bfwhm, atheta
     &            , equivrad, elongation, nonzero, almostzero )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
!!!       character*(*) lverbose

       integer       n, nx, ny, l, m, nextr, irc, iotty
       integer       nxmin(*), nxmax(*), nymin(*), nymax(*)
 
       real*8        dx, dy, pi, sqpixel, sourcearea, dblm, dbll, convtodegrees, sqreightlogtwo, pixel
     &             , intsxx, intsyy, intsxy, xcorsc, ycorsc, momp2, momm2, sqrmo, asig, bsig, intsd, ints0m, ints0
     &             , almostzero, intx, inty, int0, int0m, intd, asize, bsize
     
       real*8        image(nx,ny), segmn(nx,ny), momxco(*), momyco(*), afwhm(*), bfwhm(*), atheta(*), equivrad(*), elongation(*)
     &             , nonzero(*)

       real*8 , allocatable :: momxx(:), momyy(:), momxy(:)

       parameter   ( pi = 3.14159265358979d0, convtodegrees = 180.0d0 / pi )
!__________________________________________________________________________________________________________________________________
!                  
       sqreightlogtwo = sqrt ( 8.0d0 * log ( 2.0d0 ) )
       pixel = dx
       sqpixel = dx * dy

       allocate ( momxx(nextr), momyy(nextr), momxy(nextr), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   MEASURE: ERROR: Trouble allocating memory (76).'
         stop 76
       endif

       do n=1,nextr
         intx = 0.0d0                        
         inty = 0.0d0                        
         int0 = 0.0d0                        
         do m=nymin(n),nymax(n)              
           dblm = dble ( m )                 
           do l=nxmin(n),nxmax(n)            
             dbll = dble ( l )               
             if (n .eq. nint ( segmn(l,m) )) then
               if (image(l,m) .gt. almostzero) then
                 intd = image(l,m) * sqpixel
                 int0 = int0 + intd
                 intx = intx + dbll * intd
                 inty = inty + dblm * intd
               endif
             endif
           enddo
         enddo
         int0m = int0 + almostzero
         momxco(n) = intx / int0m
         momyco(n) = inty / int0m
       enddo

       do n=1,nextr
         ints0 = 0.0d0
         intsxx = 0.0d0
         intsyy = 0.0d0
         intsxy = 0.0d0           
         do m=nymin(n),nymax(n)
           dblm = dble ( m )
           do l=nxmin(n),nxmax(n)
             dbll = dble ( l )       
             if (n .eq. nint ( segmn(l,m) )) then
               intsd = segmn(l,m) * sqpixel
               ints0 = ints0 + intsd
               xcorsc = dbll - momxco(n)
               ycorsc = dblm - momyco(n)
               intsxx = intsxx + xcorsc * xcorsc * intsd
               intsyy = intsyy + ycorsc * ycorsc * intsd
               intsxy = intsxy + xcorsc * ycorsc * intsd
             endif
           enddo
         enddo
         ints0m = ints0 + 1.0d-30
         momxx(n) = intsxx / ints0m
         momyy(n) = intsyy / ints0m
         momxy(n) = intsxy / ints0m
       enddo
       
       do n=1,nextr
         momp2 = (momxx(n) + momyy(n)) / 2.0d0
         momm2 = (momxx(n) - momyy(n)) / 2.0d0
         
         if (abs ( momxy(n) ) .lt. almostzero) then
           if (momxx(n) .gt. momyy(n)) then
             momp2 = momxx(n)
             momm2 = momyy(n)
             atheta(n) = 90.0d0
           else
             momp2 = momyy(n)
             momm2 = momxx(n)
             atheta(n) = 0.0d0
           endif
           sqrmo = sqrt ( momm2**2 + momxy(n)**2 )
           asig  = sqrt ( momp2 + sqrmo ) * pixel
           if (momp2 .gt. sqrmo) then 
             bsig = sqrt ( momp2 - sqrmo ) * pixel
           else
             asig = 0.0d0
             bsig = 0.0d0
           endif
         else       
           sqrmo = sqrt ( momm2**2 + momxy(n)**2 )
           if (momp2 + sqrmo .ge. 0.0d0) then 
             asig  = sqrt ( momp2 + sqrmo ) * pixel
           endif
           if (momp2 - sqrmo .ge. 0.0d0) then 
             bsig  = sqrt ( momp2 - sqrmo ) * pixel
           endif
           if (momp2 + sqrmo .lt. 0.0d0 .or. momp2 - sqrmo .lt. 0.0d0) then 
             asig = 0.0d0
             bsig = 0.0d0
           endif
           atheta(n) = 180.0d0 - atan2 ( sqrmo + momm2, momxy(n) ) * convtodegrees
         endif                                        
         
         asize = sqreightlogtwo * max ( asig, pixel )
         bsize = sqreightlogtwo * max ( bsig, pixel )
         
         afwhm(n) = asize
         bfwhm(n) = bsize
         
         elongation(n) = asize / bsize
         
         if (asig .lt. pixel .or. bsig .lt. pixel) then
           elongation(n) = 1.0d0
         endif
         
         sourcearea = nonzero(n) * sqpixel
         equivrad(n) = sqrt ( sourcearea / pi )
       enddo
       
       deallocate ( momxx, momyy, momxy )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( inmx, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, header, footer
     &            , iotty, iolog, creator, beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero
     &            , bscale, wave, datamn, datamx, history, lverbose, rcore )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, anynull, myfile
       character*(*) ctype1, ctype2, creator, fname, header, footer(2), bunit, object, history, lverbose
       character*25  cdatetime
       character*80  footer1, footer2, comment
       integer       inmx, i, j, nx, ny, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, lastc, iotty, iolog, maxdim
     &             , ilc1, ilc2, rw, nfound, blank, stat01, stat02, stat03, stat04, stat05, stat06, stat07, stat08, stat09
     &             , stat10, stat11, stat12, stat13, stat14, stat15, stat16, stat17, stat18, stat19, stat20, stat21, stat22
     &             , stat23, stat24, stat25, stat26, stat27, stat28, stat29, stat30, stat31, stat32, stat0, nan
       parameter   ( maxdim = 2 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, rcore, cd11, cd12, cd21, cd22
       real*8        fun(inmx,inmx)
       real*8        as2deg, datamx, datamn
       parameter   ( as2deg = 1.0d0 / 3600.0d0 )
       external      lastc, ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat01 = 0
       stat02 = 0
       stat03 = 0
       stat04 = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
       if (stat .eq. 0 .and. (comment(1:22) .eq. 'Alexander Men`shchikov' .or. comment(1:22) .eq. 'Alexander Men''shchikov')) then
         myfile = .true.
       else
         myfile = .false.
         stat = 0
         creator = 'MODFITS'
       endif

       header = 'FITS Plot Header'

       if (myfile) then
         call ftgkys ( unit, 'HEADER' , header , comment, stat01 )
         call ftgkys ( unit, 'FOOTER1', footer1, comment, stat02 )
         call ftgkys ( unit, 'FOOTER2', footer2, comment, stat03 )
         ilc1 = lastc ( footer1 )
         ilc2 = lastc ( footer2 )
         if (ilc1 .ne. 0 .and. ilc2 .ne. 0) then
           footer(1) = footer1(1:ilc1)//footer2(1:ilc2)
         else
           footer(1) = 'Footer'
         endif
       else
         call ftgkys ( unit, 'OBJECT' , header, comment, stat04 )
         if (stat04 .ne. 0 .or. header(1:10) .eq. '          ') header = fname
         footer(1) = header
       endif

! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       call ftgkyd ( unit, 'RCORE' , rcore , comment, stat )

       if (stat .ne. 0) then
         rcore = 0.0d0
         stat = 0
       endif

       stat08 = 0
       stat0 = stat01 + stat02 + stat03 + stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15
     &       + stat16 + stat17 + stat18 + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28
     &       + stat29 + stat30 + stat31

       if (lverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat01 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword HEADER absent or error'
         if (stat02 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER1 absent or error'
         if (stat03 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword FOOTER2 absent or error'
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat29 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EPOCH absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = 'No object specified'

! Read the FITS file into the 2D array.

       call ftg2dd ( unit, group, dble ( blank ), inmx, naxes(1), naxes(2), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       nx = naxes(1)
       ny = naxes(2)
       funmin = 1.0d30
       funmax =-1.0d30
       nan = 0
       do j=1,ny
         do i=1,nx
           if (isnan ( fun(i,j) )) then  
             nan = nan + 1
             if (lverbose .ne. '-verb0') then
               if (nan .lt. 100) then
                 write (iotty,'(a,i4,a,i4,a)') '   RFITS: Element (', i,',', j, ') contains not a number (NaN)...'
               elseif (nan .eq. 100) then
                 write (iotty,'(a)') '   RFITS: Many NaNs, no more info on the elements.'
               endif
             endif
           else
             funmin = min ( funmin, fun(i,j) )
             funmax = max ( funmax, fun(i,j) )
           endif  
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, inmx, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname
     &            , header, footer, cdate, ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, epoch, equinox, bzero
     &            , bscale, wave, datamn, datamx, history, action )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend
       character*(*) fname, ctype1, ctype2, object, creator, header, footer(2), ctime, cdate, bunit, cfitsversion, history, action
       character*19  cdatetime
       integer       inmx, i, j, nx, ny, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, naxes(2), iat, iflen
     &             , lastc, blank, ica, itilde
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, dfun, rot1, rot2, epoch, equinox, datamn
     &             , cosrota, sinrota, rotangle, datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, pi
       real*8        fun(inmx,inmx)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      lastc
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 2-D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 2
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       datamn = 1.0d30
       datamx =-1.0d30
       do j=1,ny
         do i=1,nx
           dfun = fun(i,j)
           if (dfun .lt. datamn ) datamn = dfun
           if (dfun .gt. datamx ) datamx = dfun
         enddo
       enddo

       bitpix = -32     !<-- single precision floating-point values

       cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

       if (cfitsversion .eq. '5.030') then
         call fdelete ( fname, status )
       endif

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Create the new empty FITS file.

       if (cfitsversion .eq. '5.030') then
         call ftinit ( unit, fname, blocksze, status )
       else
         call ftinit ( unit, '!'//fname, blocksze, status )
       endif

! Write the required header keywords.

       call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

       call ftpkys ( unit, 'CREATOR', creator,'Alexander Men`shchikov, SAp IRFU CEA Saclay', status )
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'HEADER' , header, 'title', status )

       iat = 0
       iat = index ( footer(1), ' |F' )
       if (iat .eq. 0) iat = index ( footer(1), ' mm ' )
       iflen = lastc ( footer(1) )

       if (iat .gt. 0) then
         call ftpkys ( unit, 'FOOTER1', footer(1)(1:iat-1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', footer(1)(iat:iflen), 'part 2', status )
       else
         call ftpkys ( unit, 'FOOTER1', footer(1), 'part 1', status )
         call ftpkys ( unit, 'FOOTER2', ' ', 'part 2', status )
       endif
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )

! Write all parameters.

       dxdeg = dx * as2deg
       dydeg = dy * as2deg

! Correct if the input images had cd11 or cd22 set to zero.
       
       if (abs ( cd11 ) .lt. almostzero .or. abs ( cd11 ) .gt. dxdeg) then
         cd11 = -dxdeg
         cd12 = 0.0d0
       endif
       if (abs ( cd22 ) .lt. almostzero .or. abs ( cd22 ) .gt. dxdeg) then
         cd21 = 0.0d0
         cd22 = dydeg
       endif
       
       cosrota = - cd11 / dxdeg
       sinrota = - cd12 / dxdeg
       rotangle = acos ( cosrota ) * 180.0d0 / pi
       
       if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
         rot1 = 180.0d0 - rotangle
       elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = 180.0d0 + rotangle
       elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
         rot1 = - rotangle
       endif
       rot2 = rot1

       ica = 1
       itilde = index ( action, '~' )
       if (itilde .gt. 0) then
         bunit = action(itilde+1:)
       endif

       call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
       call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
       call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
       if  (blank .gt. 0)
     & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
       call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
       call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
       call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
       call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CDELT1' , -dxdeg, 13 , 'coordinate increment along axis', status )
       call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
       call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
       call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
       call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
       call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
       call ftpkyd ( unit, 'EPOCH'  , epoch  , 13, ' ', status )
       call ftpkyd ( unit, 'EQUINOX', equinox, 13, ' ', status )
       call ftphis ( unit, history, status )

       if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'image was convolved with the beam (arcsec)', status )

! Write the array to the FITS file.

       call ftp2dd ( unit, group, inmx, nx, ny, fun, status )

! Close the file.

       call ftclos ( unit, status )

! Free the unit number.

       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine expandit 
       
     &            ( ctype, ie, je, ieb, jeb, nbw, inmxb, inmx, image, imagewide )
!__________________________________________________________________________________________________________________________________
!
! Expand image into borders of width 'nbw'.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*(*) ctype
       integer       i, j, ie, je, ieb, jeb, nbw, inmxb, inmx, k, ni, nj
       real*8        image(inmx,inmx), imagewide(inmxb,inmxb), value
!__________________________________________________________________________________________________________________________________
!
! Initialise a larger image for no border effects in the area we are interested in.

       do j=1,jeb
         do i=1,ieb
           imagewide(i,j) = 0.0d0
         enddo
       enddo

! Shift the image to the center of the new image.

       do j=1,je
         do i=1,ie
           imagewide(i+nbw,j+nbw) = image(i,j)
         enddo
       enddo

       if (ctype .eq. 'periodic') then
              
! Assume periodicity to extend the image in x-direction.

         do i=1,nbw
           do j=1+nbw,je+nbw
             imagewide(i,j) = imagewide(i+ie,j)
           enddo
         enddo

         do i=1+nbw+ie,ieb
           do j=1+nbw,je+nbw
             imagewide(i,j) = imagewide(i-ie,j)
           enddo
         enddo

! Assume periodicity to extend the image in y-direction.

         do j=1,nbw
           do i=1+nbw,ie+nbw
             imagewide(i,j) = imagewide(i,j+je)
           enddo
         enddo

         do j=1+nbw+je,jeb
           do i=1+nbw,ie+nbw
             imagewide(i,j) = imagewide(i,j-je)
           enddo
         enddo

! In a similar fashion, take care of the corners.

         do j=1,nbw
           do i=1,nbw
             imagewide(i,j) = imagewide(i+ie,j+je)
           enddo
         enddo

         do j=1+nbw+je,jeb
           do i=1+nbw+ie,ieb
             imagewide(i,j) = imagewide(i-ie,j-je)
           enddo
         enddo

         do i=1,nbw
           do j=1+nbw+je,jeb
             imagewide(i,j) = imagewide(i,j-je)
           enddo
         enddo

         do j=1,nbw
           do i=1+nbw+ie,ieb
             imagewide(i,j) = imagewide(i-ie,j)
           enddo
         enddo

       else

! Assume "outflowing" boundary to extend the image in the x-direction.

         do i=1,nbw
           do j=1+nbw,je+nbw
             imagewide(i,j) = imagewide(nbw+1,j)
           enddo
         enddo

         do i=1+nbw+ie,ieb
           do j=1+nbw,je+nbw
             imagewide(i,j) = imagewide(nbw+ie,j)
           enddo
         enddo

! Assume "outflowing" boundary to extend the image in the y-direction.

         do j=1,nbw
           do i=1+nbw,ie+nbw
             imagewide(i,j) = imagewide(i,nbw+1)
           enddo
         enddo

         do j=1+nbw+je,jeb
           do i=1+nbw,ie+nbw
             imagewide(i,j) = imagewide(i,nbw+je)
           enddo
         enddo

! In a similar fashion, take care of the corners. 
                                 
         ni = 10
         nj = 10
         value = 0.0d0
         k = 0
         do j=1+nbw,1+nbw+nj
           do i=1+nbw,1+nbw+ni
             k = k + 1
             value = value + imagewide(i,j)
           enddo
         enddo
         value = value / dble ( k )
         do j=1,nbw
           do i=1,nbw
             imagewide(i,j) = value
           enddo
         enddo

         value = 0.0d0
         k = 0
         do j=1+nbw+je-nj,1+nbw+je
           do i=1+nbw+ie-ni,1+nbw+ie
             k = k + 1
             value = value + imagewide(i,j)
           enddo
         enddo
         value = value / dble ( k )
         do j=1+nbw+je,jeb
           do i=1+nbw+ie,ieb
             imagewide(i,j) = value
           enddo
         enddo

         value = 0.0d0
         k = 0
         do i=1+nbw,1+nbw+ni
           do j=1+nbw+je-nj,1+nbw+je
             k = k + 1
             value = value + imagewide(i,j)
           enddo
         enddo
         value = value / dble ( k )
         do i=1,nbw
           do j=1+nbw+je,jeb
             imagewide(i,j) = value
           enddo
         enddo

         value = 0.0d0
         k = 0
         do j=1+nbw,1+nbw+nj
           do i=1+nbw+ie-ni,1+nbw+ie
             k = k + 1
             value = value + imagewide(i,j)
           enddo
         enddo
         value = value / dble ( k )
         do j=1,nbw
           do i=1+nbw+ie,ieb
             imagewide(i,j) = value
           enddo
         enddo                                                                                                                            
       endif

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||